/*****************************************************************************/
/*
                                  mondesi.c

                        Monitor de Sistema (espaol)

MONDESI (pronounced "mon-deh-sih") is a browser-based VMS system monitor with a
sufficiently small resource footprint as not to impact system performance
(minimal CPU consumption and only one or two buffered IO per client update). 
No, the name's etiology has little significance at all (a small homage at
most).  I was just looking for a short (and cute) name.

This utility is intended as a general tool for system observation.  It is not a
troubleshooting tool as such, especially when a system is beginning to suffer
from resource exhaustion.  Not only does the data collection script need to
continue processing effectively (see 'SUFFICIENT PRIORITY' below) there is also
a Web server in the data transfer path.

MONDESI should work with all relatively modern browsers.  Version 4.n was
developed against Chrome 25, Firefox 19, MSIE 9 & 10, Opera 10, and Safari 6,
on OS X and Windows 7 platforms.  The intense graphing is CPU and memory
hungry so I'd not recommend Mozilla on an DEC 3000 (or even a PWS 500 :-)  Of
course YMMV.

MONDESI should also work with HTML5 WebSocket capable browsers.

Uses the system service $GETRMI() available for VMS V7.3 and later.  $GETRMI
seems to have been intended as the 'modern' and documented version of the
$GETSPI functionality.  Unfortunately, after dealing with it for a while, it's
difficult to shake the impression that someone lost interest halfway through
the project!  Note that V7.3-1 (and earlier) do not support some data items. 
These are shown as question marks.

Displayed values have been validated using the interactive MONITOR utility.

As much of the functionality is pushed off to the browser MONDESI relies
heavily on JavaScript contained in MONDESI.JS and WZ_JSGRAPHICS.JS (see below).

As MONDESI provides a readily accessable insight into a system (albeit,
fairly basic resource consumption information only) perhaps it might be best
available only under authorization.

Refresh rate defaults to two seconds.  To change this (less than two is not
recommended) define a logical name containing the value in seconds.

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_REFRESH 5

If MonDeSi stops updating the browser for any reason (e.g. server shutdown,
network connection interrupted) for a period of ninety seconds or more, the
node notes show a bold "STALLED".  MonDeSi periodically attempts to reestablish
lost network connections automatically.  The "STALLED" will disappear once
updates recommence.

NOTE: Orignally MONDESI was a vehicle for experimentation rather than an
application designed and built to address a specific need.  Over the years it
has received some use in anger and undergone a couple of major cycles of
enhancement.  It is now - for what it tries to provide - a capable and useful
tool to snapshot resource utilisation on a VMS platform.


MONITOR DISPLAY
---------------
A node is displayed as a text panel on the left and nine bar graphs on the
right.  These display the instantaneous (current) values for the various
resources and are updated dynamically at the interval when the data collector
script detects a change in values.  Each bar graph section contains a number
representing the current value and a number in the far right the maximum value
reached while monitoring.  All values are average per-second where applicable
(e.g. BIO, DIO, lock, etc.)

  +--------------------------------------------------------------------------+
  | NODENAME                     [_CPU_____________________:___|_______100_] |
  | 04-JAN-2013 05:20            [_MEM_____________|__________________1.5G_] |
  | Digital Personal Workstn     [_FLTS_________________________________10_] |
  | 1 CPU 1.5 GB Alpha V8.3      [_BIO_____:_|__________________________16_] |
  | Up: 44 14:29 Processes:35    [_DIO______________|____________________5_] |
  | interactive:4 batch:0        [_LOCK___|_____________________________39_] |
  | network:12 (v5.0.0/WASD)     [_NI________________________|_______15.6k_] |
  | [hi][cl][di][pr][pa]         [_DISK_:_________|_____________________12_] |
  | [sc][re][ze][cl][15]         [_PRC__|_______________________________35_] |
  +--------------------------------------------------------------------------+

CPU: shows total CPU utilisation in green, non-user-mode CPU
(kernel+interrupt+etc.) in brown (usually a subset of total) and average CPU
utilisation in yellow.  Where a system has single CPU the graph range is 100%,
with two CPUs the range is 200%, three is 300%, etc.

MEM: displays system physical memory (RAM) in Mbytes.  The value in the far
right is the total.  The current used is show in green.  Used memory is
calculate by subtracting the current free list plus modified list sizes from
the total physical memory.

FLTS: provides total page faults per second in green with hard page faults as a
subset in brown.  Average page faulting for the monitor period is shown in
yellow.

BIO: displays buffered I/Os per second.  Average is shown in yellow.

DIO: displays direct I/Os per second.  Average is shown in yellow.

LOCK: provides locking activity as a total of ENQs plus DEQs plus waits. 
Locks waits are shown in brown.  Average activity is shown in yellow.

NI: network interface(s) bytes received and transmitted per second; total in
green, received in brown, transmitted data being the difference.  This includes
all data moving through the physical network interface(s) including SCS, MSCP,
etc., not just protocols such as TCP/IP. 

DISK: displays total disk I/Os in green with a subset of disk writes in
brown (disk reads being the difference).  Average activity shown in yellow.

Control Buttons:

[history] button reveals/hides the history line graphs (see below).

[cluster] enabled if a clustered node, button reveals/hides cluster data

[disk] enabled if disk monitoring configured, reveals/hides disk data

[proc] enable if process monitoring is enabled, reveals/hides process display

[pause][resume] stops display updates while continuing to collect data
                release recommences display updates with paused data
                pause may be used to briefly examine something of interest
                if paused five mins MonDeSi stops itself and shows [START]

[scale] adjusts the y axis to the current maximum data value

[zero] button can be used to reset the collected maximum and average data

[print] suspends updates and invokes the browser print dialog
        button shouldn't need to be hit again after print dialogue concluded

[close] closes a standalone MonDeSi window

[5/15][15/5] toggles back and forth between 5 and 15 minute collection and
             display modes

When cluster button reveals cluster data the following additional bar graphs
and associated history graphs become visible.

  +--------------------------------------------------------------------------+
  :                                                                          :
  |                              [_SCSMSG_________________________|______1_] |
  | [hi][cl][di][pr][pa]         [_SCSKB__:_________|____________________5_] |
  | [pa][re][ze][cl][15]         [_MSCP____________|_____________________4_] |
  +--------------------------------------------------------------------------+

SCSMG: count of SCS messages exchanged.  Displays total in green and received
in brown.  (MONITOR> SCS /ITEM=(M_SEND,M_RECEIVE) analogue.)  Average is
displayed in yellow. 

SCSKB: kilobyte map rate.  Displays total in green.  (MONITOR> SCS
/ITEM=(KB_MAP) analogue.)  Average is displayed in yellow. 

MSCP: displays total requests from remote systems in green and a subset showing
write requests in brown.  Average is displayed in yellow.

When the disk button reveals disk data (configurable - see below) it displays
the per-disk operations per second.

  +--------------------------------------------------------------------------+
  :                                                                          :
  |                              [_$1$DKA0:_ALPHASYS_____:________|_____30_] |
  | [hi][cl][di][pr][pa]         [_$1$DKA100:_BACKUP_____________________0_] |
  | [sc][re][ze][cl][15]         [_$1$MDA0:_RAMDISK_:__|________________12_] |
  +--------------------------------------------------------------------------+

For configured disks the bar graph (and history) shows the device name and
volume name if mounted.  If not mounted only the device name.


NETWORK STATISTICS
------------------
MONDESI will scan the system for network devices (commonly ethernet) and use
any found for NI statistics.  If the system admin wishes to tailor the source
of the statistics the multi-valued logical name MONDESI_NI can be used to
specify one or more network interfaces.  Define this logical name system-wide
(or in a script wrapper procedure) as required.

For example:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_NI EWA0:,EWB0:

This logical name must exist as MONDESI initialises.  If undefined or the
specified device(s) cannot be accessed the ethernet data are greyed-out.


DISK STATISTICS
---------------
MONDESI will scan selected disk devices displaying the operations per second
for mounted devices.  To enable this element the system admin needs to define a
multi-valued logical name containing the devices to be monitored.  Those
selected for display must be defined using the device names as shown by a $
SHOW DEVICE command.

For example:

  $ SHOW DEVICE DKA

  Device                Device        Error    Volume         Free  Trans Mnt
   Name                 Status        Count     Label         Space Count Cnt
  $1$DKA0:    (KLAATU)  Mounted           0  ALPHASYS       90.22GB   597   1
  $1$DKA100:  (KLAATU)  Mounted           0  BACKUP        208.86GB    10   1

For example:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_DISK $1$DKA0:,$1$DKA100:,$1$MDA0:

This logical name must exist as MONDESI initialises.


PROCESS DISPLAY
---------------
MONDESI can be configured to display the top five processes for each of the
categories; CPU time, super+exec+kernel time, page faults, buffered and direct
I/O.  It is displayed as a table immediately below the bar graph area and above
the history area.  Generating and then transmitting the data is RELATIVELY
EXPENSIVE compared to other MonDeSi data. The collection code automatically
switches between newer and older SPI data and data structure (coincident with
VMS V8.3?).  The S+E+K data item is unavailable with the older.

For example, to enable this for all clients:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_ACCESS *,*,*

This logical name must exist as MONDESI initialises.

To enable top-process only for web server authenticated clients:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_ACCESS *,$*

In addition, MonDeSi can be configured to display per-process data, by clicking
on a displayed top-process.  This requires the MONDESI.EXE image to be
installed with WORLD privilege (see below).  This is relatively QUITE EXPENSIVE
in terms of data quanitiy transferred (though still noes not register in
MONITOR SYSTEM).  This displayed data should be relatively self-explanatory for
the experienced VMS users.

Per process quotas are displayed in red when they reach or exceed a minimu7m
availability, by default 15%.  The following logical name can be defined to
change this threshold.  For example:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_PROC_ALERT 25


ACCESS CONTROL
--------------
As the displayed data can be considered sensitive in most circumstances access
control on the MonDeSi script should be applied.  The access control allows any
web server authentication scheme and does not have to represent or correspond
to SYSUAF authentiation.

Access control may be effectively exercised using Web authorisation.

Important note for WASD WebSocket users:

  Currently (early 2013) not all browsers fully support all HTTP semantics
  prior to WebSocket protocol upgrade.  For example, of Chrome, Firefox,
  MSIE, Opera and Safari, only Firefox supports the standard HTTP request
  and proxy authentication mechanisms, which can be an issue if negotiating
  WebSockets through a proxy server requiring credentials, and for the
  authenticated access described in this section.

  At the time of writing (early 2013) not all browsers (only Firefox in fact)
  implement HTTP authorisation over WebSocket. WASD users employing WebSocket
  and authorisation are required to use FireFox!

Disable WebSocket Authorisation for all but selected user-agents. As noted
above, not all browsers support HTTP authorisation with WebSocket. MonDeSi
allows WebSocket access (only available with WASD 10.1 and later) to be
automatically disabled for all but selected browsers. Just define a
multi-valued logical name with zero followed by values containing unique
strings found in the cabable user-agent strings. For example, to disable
WebSocket for all but Firefox:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_WEBSOCKET 0,"Firefox"

Disabling WebSocket adds a little latency to some actions, such as per-process
display, but does not affect overall functionality.

Additional access control can be exercised in various combinations using a
multi-valued logical name.  The first value controls access to the general
monitor display, the second access to the top-process display, the third to the 
per-process display.  Access to top-process is a necessary pre-requisite for
access to per-process display.  An underscore can be used to disable access.

For example, to enable top-process and per-process for all clients:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_ACCESS *,*,*

To enable top-process for all clients and per-process for all authenticated
clients:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_ACCESS *,*,$*

Top-process for all clients and per-process for specified authenticated
clients:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_ACCESS *,*,"$<user1>,$<user2>"

Top-process for authenticated clients and per-process for specified
authenticated clients:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_ACCESS *,$*,"$=<user1>,<user2>"

Top-process and per-process for all authenticated clients:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_ACCESS *,"$*","$*"

Top-process and per-process only for specified authenticated clients:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_ACCESS "$<user1>,$<user2>","$*"

Alternatively provide one or more comma-separated, dotted-decimal IP address to
specify one or more hosts allowed to use the script, and/or one or more
comman-separated IP addresses and CIDR subnet mask to specify a range of hosts. 
IPv4 only!

For example, the specified host(s) can access top-proceess and per-process:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_ACCESS *,"#192.168.1.2",*
  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_ACCESS *,"#192.168.1.2,192.168.1.3",*

And, the specified host range (using CIDR notation):

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_ACCESS "#192.168.1.0/24",*,*


HISTORY DISPLAY
---------------
Recent data are stored within the browser and used to generate line graphs
showing values against time.  These follow the same colouring schema as the
instantaneous bar graphs.   For example, the CPU consumption line is in green
and the non-user-mode in brown.

The [history] button reveals and hides the history graphs.  The complexity of
rendering these graphics puts a noticable load on some browser systems (though
desktops are generally much more powerful than when MonDeSi was first
developed, and browser JavaScript implementations much more efficient).


PRINTING
--------
Printing a 'page' that is currently/constantly building itself (as these
displays are due to the hidden IFRAME / WebSocket asynchronously receiving the
system data updates) is problematic on all browsers tested.   As a workaround
the display updates are suspended thus stopping the page building.  The
JavaScript print is then requested.  After printing the display update is
reenabled.  Data continues to be collected during the display update
suspension.  If the [print] button remains highlighted after the print dialogue
closes click it again to resume the display updates.


CONFIGURATION
-------------
The configuration page provides a text entry area into which URLs for the nodes
to monitor are entered, one per line.  The 'nodes across window' selector allow
monitor displays to be tiled (if there is enough screen real estate).  Using
this approach widely 'separated' hosts can be monitored on a single page.  The
URL generated by the configuration dialog can be saved as a bookmark, entered
into an HTML page link, etc., for ease of subsequent access.


SUFFICIENT PRIORITY
-------------------
For a monitor such as this to be effective on a busy system it needs to have a
bit of an edge!  If INSTALLed with ALTPRI it will bump it's process priority up
to 6 when starting and reduce it back to previous when exiting (just in case
it's WASD and a persistent scripting process :-) 

  $ INSTALL REPLACE <location>MONDESI.EXE /PRIV=(ALTPRI)

Don't forget to reINSTALL any subsequently built or deployed executable.


CONTROLLING ACCESS
------------------
MONDESI provides an insight into the system and therefore is a security
consideration.  The information is fairly benign but for this and other reasons
it is often desirable to control who may use it.  Of course this can be
exercised using the mapping and/or authorisation functionality of the
underlying web server.  This is a WASD example that merely makes MONDESI
unknown outside of the local subnet.

  # WASD_CONFIG_MAP (HTTPD$MAP)
  if (!remote-addr:192.168.1.0/24) pass /cgi-bin/mondesi* "404 Not Found"


WEBSOCKET
---------
MONDESI v2.0ff optionally uses HTML5 WebSocket

  http://en.wikipedia.org/wiki/Websockets

to perform the data transfer from server to client.  The WASD server v10.1 and
later supports scripting using WebSockets as the IPC.  A single CGIplus
scripting process can support multiple WebSocket clients making it particularly
efficient.  The code must support multiple asynchonous threads of processing. 
MONDESI provides this using ASTs.  The WsLib..() functions from the WASD wsLIB
library handle the concurrent WebSocket processing.  See the WASD documentation
for detailed information on WebSocket support.

For multi-client WebSocket usage MonDeSi needs to be activated as a CGIplus
script.  This activation will happily support non-WebSocket clients as well. 
CGIplus activation can be ensured using a mapping rule providing redirection,
for example

  redirect /cgi-bin/mondesi ///cgiplus-bin/mondesi
  pass /mondesi/-/* /www_root/mondesi/*

MONDESI detects the underlying request type and adjusts its behaviour between
standard HTTP clients (using the "Web Push" techology described below) and
WebSocket clients.  In fact, under WASD, a single (CGIplus) process with
multiple WebSocket-connected clients can then be seconded to service a
CGI(plus) (i.e. non-WebSocket) request.  When the CGI(plus) client disconnects
WASD will run-down the process disconnecting the WebSocket clients in the
process (so-to-speak).  Not much can be done about this without having a
WebSocket specific script (perhaps wrapper).  Just an interesting wrinkle!


WEB "PUSH"
----------
MONDESI uses JavaScript and a Web 'push' technology (such as it is) to return
data (not images!) from a Web server scripting process to the client monitor
page where it is displayed as text or rendered into (crude but effective) bar
graphs.  Perhaps it would be better described as "HTTP streaming" or as "long
polling" but it *definitely* does not need to poll for each update.  The "push"
will last as long as the Web server and intermediate network connectivity
allows.

A parent page provides the JavaScript and HTML infrastructure to support
multiple monitored nodes each contained in an indepdendent, visible IFRAME,
which in turn used a dynamically created (i.e. JavaScript), hidden IFRAME to
execute the data collection URL (script).  This executes continuously,
periodically collecting data and outputting changed items into the response
stream.  Each response record contains a JavaScript tag element which contains
calls to the required data processing and display functions in the visible
IFRAME to process (only) each datum that needs updating.

If a data collection process (script) should exit for whatever reason (timeout,
network interruption, etc.) MONDESI should restart data collection
automatically.  If the monitored system become unavailable for any reason
(restart, network interruption, etc.) the monitor highlights the node
information with red after ninety seconds.  Elementary but suitable to the
desired outcome.


INTERFRAME COMMUNICATION (IFC)
------------------------------
The node(s) displayed in the parent page are not necessarily the same as the
parent page (indeed, multiple nodes could not all be) and of course these have
differing host names and/or IP addresses.  This incurs browser cross-domain
restrictions with communication between the node display HTML (i)frame and the
primary, top-level page, specifically - when needing to open/close the history
graph portion of the display associated with each node.

Two approaches for providing this IFC are used. 
Both are well documented on the Web.

1) The more elegant approach uses the modern, architected, HTML5 postMessage()
communication mechanism.  MONDESI checks for and uses this if available in the
browser.  It sends a message to the top-level page which invokes a function to
adjust the node display height.  Currently (at initial MONDESI development)
Firefox 3 (specifically Gecko 1.9), Opera 9.5 and MSIE 8 (at least) support it.

2) If postMessage() is unavailable MONDESI falls back to using an IFC *kludge*
involving the dynamic creation of an IFRAME within the node monitor IFRAME and
using this to load a URL for the served file BCHAN.HTML which contains a call
to a JavaScript function in the top-level page (to adjust the height).  The
data are contained in the hash portion of the URL and never appears outside the
browser.  This should be usable with most older, mainstream browsers.


GRAPHING
--------
This application uses a third-party JavaScript graphics toolkit to generate
both the bar and line graphs.

  Copyright (c) 2002-2009 Walter Zorn. All rights reserved.

  Performance optimizations for Internet Explorer
  by Thomas Frank and John Holdsworth.
  fillPolygon method implemented by Matthieu Haller.

  LICENSE: LGPL

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License (LGPL) as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

Used to be documented at

  http://www.walterzorn.com/jsgraphics/jsgraphics_e.htm

Now at a site maintained on behalf of the late Walter Zorn

  http://www.walterzorn.de/en/index.htm

Also accessable via WayBackMachine (split URL)

  http://replay.waybackmachine.org/
  20090302111911/http://www.walterzorn.com/jsgraphics/jsgraphics_e.htm

(and now archived in the [.JS] directory).

Many thanks to Walter for making this useful tool freely available. 
This toolkit works extraordinarily well considering it uses no graphics
capability in the browser!!  This is a non-SVG, non-VML, DOM kludge (not
meaning this as a pejorative).


BUILD DETAILS
-------------
See BUILD_MONDESI.COM procedure.

  $ @BUILD_MONDESI [LINK]


INSTALLED IMAGE
---------------
Per-process display requires the image to be installed with WORLD privilege.
System startup requires something like

  $  INSTALL ADD <location>MONDESI.EXE /PRIV=(WORLD)

possibly (see above)

  $  INSTALL ADD <location>MONDESI.EXE /PRIV=(WORLD,ALTRI)

and on executable image update

  $  INSTALL REPLACE <location>MONDESI.EXE


APACHE INSTALLATION
-------------------
Copy the executable to the APACHE$ROOT:[CGI-BIN] directory.

Map the JavaScript resources into web-space using (assumes the application is
located under APACHE$COMMON:[000000]):

  Alias /mondesi/-/ "/apache$common/mondesi/"


OSU ACCOMODATIONS
-----------------
OSU requires the use of the MONDESI.COM DCL wrapper to invoke the OSU CGI
environment utility before executing MONDESI.  The wrapper and the MONDESI.EXE
executable both should be placed in WWW_ROOT:[BIN].  Rather than introduce the
bulk of WASD CGILIB this deliberately slimline monitor utility checks for an
indication of the OSU environment and overtly handles the indiosyncracies.  
MONDESI does not reopen <stdout> in binary mode.  OSU requires two CGI response
headers (MONDESI supplies "Status:" and "Content-Type:").  OSU also dislikes a
response header with <CR><LF> carriage-control, so MONDESI uses only an <LF>.

Map the JavaScript resources into web-space using (assumes the application is
located under WWW_ROOT:[000000]):

  pass /mondesi/-/* /www_root/mondesi/*


WASD INSTALLATION
-----------------
Copy the executable to the CGI_EXE: directory.

Map the JavaScript resources into web-space using (assumes the application is
located in the source tree):

  # HTTPD$MAP
  pass /mondesi/-/* /wasd_root/src/mondesi/*


COPYRIGHT
---------
Copyright (C) 2008-2013 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY (update SOFTWAREVN as well!)
---------------
24-NOV-2013  MGD  v5.0.0, migrate functionality from executable to JavaScript
                          add process state count to top process display
                          CollectDiskDVI() collects disk size/used
                          CollectProcSPI1/2() process count >= max plus +10
                          CollectRMI() see 16-NOV-2013 quirk
                          VAX-specifics removed from code after failing to
                            get it working (simple things like $GETRMI FAULTS
                            and BUFIO just refused to return data)
31-MAR-2013  MGD  v4.0.0, if rapid major version jumps are good enough
                            for Chrome and the like :-)
                          per-process monitoring
                          compile Alpha and Itanium using same IEEE float
                          logical name MONDESI_PROC obsoleted
22-JAN-2013  MGD  v3.0.1, JavaScript update; finally discovered what garbles
                            printable WZ graphics on MSIE (or at least how to
                            prevent it) - reset "jg_fast".
                          few minor HTML and CSS tweaks
                          Happy birthday Emily!
06-JAN-2013  MGD  v3.0.0, significant rework
                          top process monitoring
                          individual disk monitoring
                          redesign data delivery for consistency between
                            concurrent WebSocket clients
                          note the WebSocket required REDIRECT described above!
                          Happy birthday Gnomie!
08-APR-2012  MGD  v2.1.1, nudge MSIE 10.0 (Windows 8 preview) to do its best!
                          <meta http-equiv="X-UA-Compatible" ...>
28-JAN-2012  MGD  v2.1.0, migrate to WASD WebSocket wsLIB environment
                          (after significant changes to WebSocket drafts)
23-OCT-2010  MGD  v2.0.0, happy birthday Raupu
                          add network interface (NI) data
                          add processes
                          add lock waits to LOCK display
                          add SCS messages
                          PGFL data is now shown by hovering over MEM
                          data are now presented as units, kilo, mega, giga
                          support WASD WebSocket environment
                          bugfix; nodes across
03-JUN-2009  MGD  v1.0.2, bumped the dot-point merely to identify
                          the 3.05 version of wz_jsgraphics.js
13-DEC-2008  MGD  v1.0.1, CollectRMI() allow for 7.3-1 and earlier (sigh!)
                          thanks Graham Burley for suggesting sys$gl_*jobcnt
29-NOV-2008  MGD  v1.0.0, initial
*/
/*****************************************************************************/

#define SOFTWAREVN "5.0.0"
#define SOFTWARENM "MONDESI"
#define SOFTWARECR "Copyright (C) 2008-2013 Mark G.Daniel"
#ifdef __ALPHA
   char SoftwareID [] = SOFTWARENM " AXP-" SOFTWAREVN;
#endif
#ifdef __ia64
   char SoftwareID [] = SOFTWARENM " IA64-" SOFTWAREVN;
#endif
#ifdef __VAX
#error *** VAX PLATFORM IS TOO HIT-AND-MISS FOR THIS APPLICATION ***
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <inet.h>
#include <math.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unixlib.h>

/* VMS-related header files */
#include <cvtdef.h>
#include <cvt$routines.h>
#include <descrip.h>
#include <dvidef.h>
#include <dvsdef.h>
#include <iodef.h>
#include <jpidef.h>
#include <libdtdef.h>
#include <lib$routines.h>
#include <lnmdef.h>
#include <psldef.h>
#include <rmidef.h>
#include <ssdef.h>
#include <starlet.h>
#include <stsdef.h>
#include <syidef.h>

#include "wslib.h"

/* required (to compile) for VMS V7.3-1 and earlier */
#ifndef RMI$_CPUCOMPAT
#define RMI$_CPUCOMPAT 21193
#define RMI$_CPUINTSTK 21194
#define RMI$_CPUMPSYNCH 21195
#define RMI$_CPUKERNEL 21196
#define RMI$_CPUEXEC 21197
#define RMI$_CPUSUPER 21198
#define RMI$_CPUUSER 21199
#define RMI$_CPUIDLE 21200
#define RMI$_PROCBATCNT 21202
#define RMI$_PROCINTCNT 21203
#define RMI$_PROCNETCNT 21204
#define RMI$_MSCP_EVERYTHING 4265
#endif

/* required for VMS V7.3-1 and earlier */
globalref sys$gl_ijobcnt;
globalref sys$gl_bjobcnt;
globalref sys$gl_njobcnt;

/* by default all structures on non-VAX are not member aligned */
#pragma __nomember_alignment

#define BOOL int
#define TRUE 1
#define FALSE 0
 
#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))
 
#define FI_LI "MONDESI", __LINE__
#define EXIT_FI_LI(status) { printf ("[%s:%d]", FI_LI); exit(status); }

#define AMBIT_PRIORITY 6
#define PER_PROC_ALERT 15  /* percent remaining */
#define REFRESH_INTERVAL 2

/* exit after fifteen minutes without a WebSocket connection */
#define EXIT_AFTER_SECONDS 15 * 60;

/* maximum number of network devices */
#define NI_DEVICE_MAX 8

/* maximum number of disks that can be monitored */
#define DISK_DEVICE_MAX 32
#define DISK_NAME_SIZE 16
#define DISK_USAGE_SIZE 31
#define DISK_VOLNAM_SIZE 12
/* maximim number of disks readable from system */
#define DISK_OFFSETS_MAX 100

/* maximum number of processes that can be reported */
#define PROC_MAX 5
#define PROC_NAME_SIZE 15

/* maximum number of processes readable from system */
#define PROC_CLASS_MAX 1000
#define PROC_CLASS_INC  100

/*
Proportional fonts derived from:
http://www.codestyle.org/css/font-family/sampler-SansSerif.shtml
*/
#define FONT_PROPORTIONAL "\"Lucida Sans\",Arial,Helvetica,Tahoma,sans-serif"
#define FONT_PROP_SIZE 12

/*
Monospace fonts derived from:
http://www.codestyle.org/css/font-family/sampler-Monospace.shtml
*/
#define FONT_MONOSPACE "\"Lucida Sans Typewriter\",\"Lucida Console\",\
\"Andale Mono\",\"Consolas\",\"Monaco\",\"DejaVu Sans Mono\",Courier,\
\"Courier New\",monospace,serif" 
#define FONT_MONO_PROC_SIZE 7

/* font size in points for control buttons */
#define FONT_BTN_SIZE 7

/* font size in points for notes */
#define FONT_NOTES_SIZE 9

#define ACCESS_NOT_PERMITTED \
"Status: 403\n\
Content-Type: text/plain\n\
Content-Length: 29\n\
\n\
MonDeSi Access Not Permitted\n"

/*******************/
/* data structures */
/*******************/

typedef struct STRUCT_IOSBLK
{
   unsigned short  iosb$w_status;
   unsigned short  iosb$w_bcnt;
   unsigned int    iosb$l_reserved;
}
IOSBLK;

/* from SPIDEF.H REMIDEF.H */
/* these members are aligned on non-VAX */
#  pragma __member_alignment __save 
#  pragma __member_alignment 
#  ifdef __INITIAL_POINTER_SIZE
      /* and 32 bit pointers */
#     pragma __required_pointer_size __save
#     pragma __required_pointer_size __short
#  endif

typedef struct STRUCT_SYSTEM_DATA
{
   unsigned long  ActiveCpuCount,
                  AvailCpuCount,
                  BufferedIO,
                  ClusterMsgSent,
                  ClusterMsgRcvd,
                  ClusterKbyteMap,
                  ClusterKbyteSent,
                  ClusterKbyteRcvd,
                  ClusterMember,
                  ClusterEVotes,
                  ClusterNodes,
                  ClusterQuorum,
                  ClusterVotes,
                  CpuCount,
                  DirectIO,
                  FcpRead,
                  FcpWrite,
                  FreeList,
                  HardPageFaults,
                  LockActivity,
                  LockWait,
                  MemInUse,
                  MemInUseMax,
                  MemSize,
                  ModList,
                  MscpRead,
                  MscpWrite,
                  PageFaults,
                  PageFaultsHard,
                  PageFaultsSoft,
                  PageFileFree,
                  PageFilePage,
                  PageFilePercent,
                  PageReadIO,
                  PageSize,
                  PageWriteIO,
                  PrevBufferedIO,
                  PrevBusyCpu,
                  PrevClusterMsgSent,
                  PrevClusterMsgRcvd,
                  PrevClusterKbyteMap,
                  PrevClusterKbyteSent,
                  PrevClusterKbyteRcvd,
                  PrevClusterNodes,
                  PrevWebSockConnectedCount,
                  PrevDirectIO,
                  PrevFcpRead,
                  PrevFcpWrite,
                  PrevHardPageFaults,
                  PrevLockActivity,
                  PrevLockWait,
                  PrevMemInUse,
                  PrevMscpRead,
                  PrevMscpWrite,
                  PrevPageFaults,
                  PrevPageFilePercent,
                  PrevProcessCount,
                  PrevProcessIntCount,
                  PrevUserCpu,
                  ProcessCount,
                  ProcessBatCount,
                  ProcessIntCount,
                  ProcessNetCount,
                  Processes;
 
   unsigned __int64  NetIntCount,
                     NetIntBlocksRx,
                     NetIntBlocksTx,
                     NetIntBytesRx,
                     NetIntBytesTx,
                     NetIntErrorsHard,
                     NetIntErrorsSoft,
                     PrevNetIntBlocksRx,
                     PrevNetIntBlocksTx,
                     PrevNetIntBytesRx,
                     PrevNetIntBytesTx;

   char  ArchName [15+1],
         HwName [60+1],
         NodeName [15+1],
         OutputBuffer [2048+(64*DISK_DEVICE_MAX)],
         ProcBuffer [256+(48*6*PROC_MAX)],
         VmsVersion [8+1];

   __int64  BootBinTime,
            CpuTicksBusy,
            CpuTicksUser;

   int  CpuBusyPercent,
        CpuUserPercent,
        MonValLength,
        NodeInfoLength,
        NodeInfoTimeStamp,
        PrevInit,
        PrevMinute,
        ProcBufferLength,
        ProcRMI,
        SystemDelta,
        UsageCount;

   int  DiskCount,
        DiskDelta,
        MountCount;
   int  DiskNameLen [DISK_DEVICE_MAX];
   char  DiskAlert [DISK_DEVICE_MAX][DISK_NAME_SIZE+1],
         DiskName [DISK_DEVICE_MAX][DISK_NAME_SIZE+1],
         DiskUsage [DISK_DEVICE_MAX][DISK_USAGE_SIZE+1],
         DiskVolName [DISK_DEVICE_MAX][DISK_VOLNAM_SIZE+1];
   unsigned long  DiskErrCnt [DISK_DEVICE_MAX],
                  DiskFreeBlocks [DISK_DEVICE_MAX],
                  DiskOptCnt [DISK_DEVICE_MAX],
                  DiskOperCount [DISK_DEVICE_MAX],
                  DiskPrevOperCount [DISK_DEVICE_MAX],
                  DiskPrevOptCnt [DISK_DEVICE_MAX];

   int  ProcSanityCheck;
   char  ProcName [PROC_MAX][PROC_NAME_SIZE+1];
   unsigned long  ProcBioCnt [PROC_MAX],
                  ProcCpuTim [PROC_MAX],
                  ProcDioCnt [PROC_MAX],
                  ProcEpid [PROC_MAX],
                  ProcPageFlts [PROC_MAX];
   void (*ProcSpiFunction)();
}
SYSTEM_DATA;

#define ACCESS_GENERAL     0
#define ACCESS_TOP_PROCESS 1
#define ACCESS_PER_PROCESS 2

/* used for multiple per-WebSocket client and CGI client (for convenience) */
typedef struct STRUCT_CLIENT_DATA
{
   BOOL  AccessAllowed [3];

   int  NodeInfoTimeStamp,
        PerProcBufferLength,
        PerProcDatum,
        PerProcPid,
        PrevPerProcPid,
        ProcRMI;

   unsigned long  PrevJpiBufIo,
                  PrevJpiDirIo,
                  PrevJpiPageFlts,
                  PrevCpuTime,
                  PrevCpuSekTime;

   char  InputBuffer [256],
         PerProcBuffer [2048+256];

   struct  WsLibStruct  *WsLibPtr;

} CLIENT_DATA;

#pragma __member_alignment __restore

/******************/
/* global storage */
/******************/

BOOL  Debug,
      ApacheHttpd,
      OsuHttpd,
      UsingCGI,
      WasdHttpd;

int  CgiNodeInfoTimeStamp,
     IsCgiPlus,
     PerProcAlert = PER_PROC_ALERT,
     PrevPrio,
     RefreshInterval,
     UsageCount,
     WebSockConnectedCount;

unsigned int  CollectDataSeconds,
              ExitTimeStamp,
              SpiEfn,
              TimeStamp;

__int64  CurrentBinTime;
unsigned short  CurrentNumTime [7];

float  DeltaSeconds = 999999.0;

char  *CgiHttpRefererPtr,
      *CgiHttpUserAgentPtr,
      *CgiRemoteAddrPtr,
      *CgiRemoteUserPtr,
      *CgiScriptNamePtr,
      *CgiQueryStringPtr,
      *CgiPlusEofPtr = NULL,
      *CgiPlusEotPtr = NULL,
      *CgiPlusEscPtr = NULL;

char  NetIntDev [NI_DEVICE_MAX*24],
      NotePad [128],
      RefererHost [128];

CLIENT_DATA  CgiClientData;
SYSTEM_DATA  SystemData;

/**************/
/* prototypes */
/**************/

void AccessControl (CLIENT_DATA*);

void CollectData ();
int CollectDiskSPI ();
void CollectDiskDVI ();
int CollectNetInt ();
int CollectPerProc (CLIENT_DATA*);
void CollectProcSPI1 ();
void CollectProcSPI2 ();
int CollectRMI ();
int CollectSyiStatic ();
int CollectSyiDynamic ();

void MonDeSiForm (char*, int, int);
void MonDeSiFormatData ();
void MonDeSiHtmlData ();
void MonDeSiNode (char*);
void MonDeSiPage (char*, int);
char* MungeProcName (char*);
char* MungeString (char*);
char* MungeTime (__int64 *btptr);
void ReturnPrio ();
void SetDebug (int);
int UrlDecode (char*);

int exe$getspi (__unknown_params);
int sys$getrmi (__unknown_params);

int WebSockAvailable ();
void WebSockDestroy (struct WsLibStruct*);
void WebSockReadAst (struct WsLibStruct*);
void WebSockRequest ();

/*****************************************************************************/
/*
*/

int main ()

{
   static unsigned short  LogValueLen;
   static char  LogValue [32];
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LogValue)-1, LNM$_STRING, &LogValue, &LogValueLen },
      { 0,0,0,0 }
   };
   static $DESCRIPTOR (LogNameProcAlertDsc, "MONDESI_PROC_ALERT");
   static $DESCRIPTOR (LogNameRefreshDsc, "MONDESI_REFRESH");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   int  status,
        NodesAcross,
        ViewWidth;
   char  *cptr, *sptr, *zptr;
   char  IframeId [64],
         NodeList [512];

   /*********/
   /* begin */
   /*********/

   /* ambit attempt to bump the process priority (if INSTALLed with ALTPRI) */
   atexit (&ReturnPrio);
   sys$setpri (0, 0, AMBIT_PRIORITY, &PrevPrio, 0, 0);

   ApacheHttpd = (getenv ("APACHE$SHARED_SOCKET") != NULL ||
                  getenv ("APACHE$PARENT_PID") != NULL ||
                  getenv ("APACHE$SERVER_TAG") != NULL);
   OsuHttpd = (getenv("WWWEXEC_RUNDOWN_STRING") != NULL);
   WasdHttpd = (getenv("HTTP$INPUT") != NULL);

   SetDebug (1);

   IsCgiPlus = ((CgiPlusEofPtr = getenv("CGIPLUSEOF")) != NULL);
   if ((CgiPlusEscPtr = getenv("CGIPLUSESC")) == NULL) CgiPlusEscPtr = "";
   if ((CgiPlusEotPtr = getenv("CGIPLUSEOT")) == NULL) CgiPlusEotPtr = "";

   status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameRefreshDsc, 0, &LnmItems);
   if (VMSnok (status))
      LogValue[0] = '\0';
   else
      LogValue[LogValueLen] = '\0';
   RefreshInterval = atoi(LogValue);
   if (RefreshInterval <= 0 || RefreshInterval > 60)
      RefreshInterval = REFRESH_INTERVAL;

   status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameProcAlertDsc, 0, &LnmItems);
   if (VMSnok (status))
      LogValue[0] = '\0';
   else
      LogValue[LogValueLen] = '\0';
   PerProcAlert = atoi(LogValue);
   if (PerProcAlert <= 0 || PerProcAlert > 100)
      PerProcAlert = PER_PROC_ALERT;

   memset (&SystemData, 0, sizeof(SystemData));
   CollectSyiStatic (&SystemData);
   SystemData.NetIntCount = CollectNetInt (NULL);
   SystemData.DiskCount = CollectDiskSPI (&SystemData);
   SystemData.ProcSpiFunction = &CollectProcSPI1;
   CollectData ();

   do {

      /* with CGIplus this call will block waiting for the next request */
      WsLibCgiVar ("");

      SetDebug (1);

      /********************/
      /* request received */
      /********************/

      UsageCount++;

      CgiHttpRefererPtr = WsLibCgiVar("HTTP_REFERER");
      CgiHttpUserAgentPtr = WsLibCgiVar("HTTP_USER_AGENT");
      CgiQueryStringPtr = WsLibCgiVar("QUERY_STRING");
      CgiRemoteAddrPtr = WsLibCgiVar("REMOTE_ADDR");
      CgiRemoteUserPtr = WsLibCgiVar("REMOTE_USER");
      CgiScriptNamePtr = WsLibCgiVar("SCRIPT_NAME");

      /* derive the URL scheme and host from the HTTP referrer field */
      zptr = (sptr = RefererHost) + sizeof(RefererHost)-1;
      cptr = CgiHttpRefererPtr;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';

      if (isdigit(*CgiQueryStringPtr))
      {
         /********/
         /* data */
         /********/

         if (!SpiEfn)
            if (VMSnok (status = lib$get_ef (&SpiEfn)))
               EXIT_FI_LI (status);

         UsingCGI = (WsLibCgiVarNull ("WEBSOCKET_INPUT") == NULL);

         if (UsingCGI)
            MonDeSiHtmlData ();
         else
         {
            if (!WsLibIsCgiPlus())
            {
               /* WebSocket must be via a persistent process */
               fprintf (stdout,
"Status: 403 Forbidden\r\n\
\r\n\
MonDeSi (using WebSocket) must be accessed via a CGIplus mapping!");
               fflush (stdout);
            }
            else
               WebSockRequest ();
         }
      }
      else
      {
         /********/
         /* page */
         /********/

         NodesAcross = 1;
         NodeList[0] = '\0';
         ViewWidth = 5;

         if (cptr = strstr (CgiQueryStringPtr, "across="))
            NodesAcross = atoi(cptr+7);
         if (NodesAcross < 0 || NodesAcross > 4) NodesAcross = 1;

         if (cptr = strstr (CgiQueryStringPtr, "width="))
            ViewWidth = atoi(cptr+6);
         if (ViewWidth != 5 && ViewWidth != 15) ViewWidth = 5;

         IframeId[0] = '\0';
         if (cptr = strstr (CgiQueryStringPtr, "iframe="))
         {
            zptr = (sptr = IframeId) + sizeof(IframeId)-1;
            for (cptr += 7;
                 *cptr && *cptr != '&' && sptr < zptr;
                 *sptr++ = *cptr++);
            *sptr = '\0';
         }

         if (cptr = strstr (CgiQueryStringPtr, "nodes="))
         {
            zptr = (sptr = NodeList) + sizeof(NodeList)-1;
            for (cptr += 6;
                 *cptr && *cptr != '&' && sptr < zptr;
                 *sptr++ = *cptr++);
            if (sptr >= zptr) EXIT_FI_LI (SS$_BUGCHECK);
            *sptr = '\0';
            UrlDecode (NodeList);
            /* eliminate those pesky W*s <CR><LF>s */
            for (cptr = sptr = NodeList; *cptr; *sptr++ = *cptr++)
            {
               if (*(unsigned short*)cptr == '\r\n') cptr++;
               if (*cptr == ' ')
               {
                  /* bookmarklets substitute spaces for carriage-control */
                  cptr++;
                  *sptr++ = '\n';
               }
            }
            *sptr = '\0';
            /* compress white-space */
            for (cptr = sptr = NodeList; *cptr; *sptr++ = *cptr++)
               while (isspace(*cptr) && isspace(*(cptr+1))) cptr++;
            *sptr = '\0';
         }
         else
            NodeList[0] = '\0';

         if (strstr (CgiQueryStringPtr, "monitor=0"))
            MonDeSiForm (NodeList, NodesAcross, ViewWidth);
         else
         if (IframeId[0])
            MonDeSiNode (IframeId);
         else
         if (NodeList[0])
            MonDeSiPage (NodeList, NodesAcross);
         else
            MonDeSiPage (SystemData.NodeName, 1);
      }

      SetDebug (0);

      if (IsCgiPlus)
      {
         /* send CGIplus end-of-request sentinal */
         fflush (stdout);
         fputs (CgiPlusEofPtr, stdout);
         fflush (stdout);
      }

   } while (IsCgiPlus);

   exit (SS$_NORMAL);
} 

/*****************************************************************************/
/*
Return to previous process priority (in case ambit raise was successful).
*/

void ReturnPrio ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (!PrevPrio) return;

   status = sys$setpri (0, 0, PrevPrio, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$setpri() %%X%08.08X\n", status);
} 

/*****************************************************************************/
/*
Provide the client with the configuration setup form.
*/

void MonDeSiForm
(
char *NodeList,
int NodesAcross,
int ViewWidth
)
{
   char  *cptr, *sptr;
   char  NodeUrl [256];
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   memset (clptr = &CgiClientData, 0, sizeof(CgiClientData));

   AccessControl (clptr);

   if (!clptr->AccessAllowed[ACCESS_GENERAL])
   {
      fputs (ACCESS_NOT_PERMITTED, stdout);
      return;
   }

   if (!*NodeList)
   {
      /* concoct a node list entry */
      sptr = WsLibCgiVar ("REQUEST_SCHEME");
      if (!*sptr)
      {
         sptr = WsLibCgiVar ("SERVER_PORT");
         if (atoi(sptr) == 443)
            sptr = "https:";
         else
            sptr = "http:";
      }
      cptr = WsLibCgiVar ("HTTP_HOST");
      sprintf (NodeList = NodeUrl, "%s//%s%s", sptr, cptr, CgiScriptNamePtr);
   }

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<!DOCTYPE html>\n\
<html>\n\
<head>\n\
<meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">\n\
<meta name=\"copyright\" content=\"%s - GPL licensed\">\n\
\
<link rel=\"stylesheet\" type=\"text/css\" href=\"%s%s\">\n\
<script language=\"JavaScript\">\n\
var WASDwebsocket = %s\n\
var mdsExeVersion = \"%s\";\n\
var mdsConfigNOws = %s;\n\
</script>\n\
<script language=\"JavaScript\" src=\"%s%s\"></script>\n\
<title>%s</title>\n\
</head>\n\
\
<body id=\"mondesi_body\" class=\"mondesi_body\">\n\
<form id=\"config_form\" method=\"GET\" action=\"%s\" \
enctype=\"application/x-www-form-urlencoded\" autocomplete=\"off\">\n\
<div class=\"mondesi_title\">Monitor de Sistema</div>\n\
<div class=\"mondesi_subtitle\">~ System Monitor ~</div>\n\
<p><textarea id=\"nodes\" name=\"nodes\" class=\"config_textarea\" \
rows=\"10\" cols=\"64\" scroll=\"auto\" onchange=\"mdsConfigLink(true)\">\
%s</textarea>\n\
<p><select id=\"across\" name=\"across\" onchange=\"mdsConfigLink(true)\">\n\
<option value=\"1\"%s>1\n\
<option value=\"2\"%s>2\n\
<option value=\"3\"%s>3\n\
<option value=\"4\"%s>4\n\
</select>\
&nbsp;&nbsp;<span id=\"config_across\">node(s) across window</span>\n\
<br><select id=\"view\" name=\"view\" \
onchange=\"mdsConfigLink(true)\">\n\
<option value=\"5\"%s>5&#10003; 15\n\
<option value=\"15\"%s>15&#10003; 5\n\
</select>\
&nbsp;&nbsp;<span id=\"config_view\">minutes</span>\n\
<script language=\"JavaScript\">mdsConfigOptions()</script>\n\
<input type=\"hidden\" name=\"monitor\" value=\"1\">\n\
<p><input id=\"config_monitor\" type=\"submit\" value=\" Monitor \" \
class=\"btn btn2\">\n\
<input id=\"config_open\" type=\"submit\" value=\" Open \" \
class=\"btn btn2\" onclick=\"return openMonDeSi()\">\n\
<input id=\"config_open\" type=\"submit\" value=\" Link \" \
class=\"btn btn2\" onclick=\"return mdsConfigLink()\">\n\
<input id=\"config_reset\" type=\"reset\" value=\" Reset \" \
class=\"btn btn2\" onclick=\"return mdsConfigReset();\">\n\
<input id=\"config_clear\" type=\"reset\" value=\" Clear \" \
class=\"btn btn2\" onclick=\"return mdsConfigClear();\">\n\
\
</form>\n\
\
<div id=\"config_link\">\n\
<a id=\"config_anchor\" \
onclick=\"eval(this.href.substr(11));return false;\"></a>\n\
<div id=\"config_drag\"></div>\n\
</div>\n\
\
</body>\n\
</html>\n",
            SOFTWARECR,
            RefererHost, "/mondesi/-/js/mondesi.css",
            WebSockAvailable() ? "true" : "false",
            SOFTWAREVN,
            atoi(WsLibCgiVar("FORM_NOWS")) ? "true" : "false",
            RefererHost, "/mondesi/-/js/mondesi.js",
            SoftwareID,
            CgiScriptNamePtr,
            /* should be no HTML-forbidden characters in this! */
            NodeList,
            NodesAcross == 1 ? " selected" : "",
            NodesAcross == 2 ? " selected" : "",
            NodesAcross == 3 ? " selected" : "",
            NodesAcross == 4 ? " selected" : "",
            ViewWidth == 5 ? " selected" : "",
            ViewWidth == 15 ? " selected" : "");
} 

/*****************************************************************************/
/*
Provide the page that contains one or more monitor panels (IFRAMEs containing
their own HTML 'page' content).
*/

void MonDeSiPage
(
char *NodeList,
int NodesAcross
)
{
   int  MonitorMode = 1,
        NodeCount = 0;
   char  *aptr, *cptr, *sptr, *zptr;
   char  HostUrl [128];
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MonDeSiPage() |%s|\n", NodeList);

   memset (clptr = &CgiClientData, 0, sizeof(CgiClientData));

   AccessControl (clptr);

   if (!clptr->AccessAllowed[ACCESS_GENERAL])
   {
      fputs (ACCESS_NOT_PERMITTED, stdout);
      return;
   }

   if (cptr = WsLibCgiVarNull("FORM_MONITOR")) MonitorMode = atoi(cptr);

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<!DOCTYPE html>\n\
<html>\n\
<head>\n\
<meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">\n\
<meta name=\"copyright\" content=\"%s - GPL licensed\">\n\
<link rel=\"stylesheet\" type=\"text/css\" href=\"%s%s\">\n\
<script language=\"JavaScript\">\n\
var mdsConfigView = %d;\n\
var mdsExeVersion = \"%s\";\n\
</script>\n\
<script language=\"JavaScript\" src=\"%s%s\"></script>\n\
<title>%s</title>\n\
</head>\n\
<body id=\"mondesi_body\" class=\"mondesi_body%s\" onload=\"mdsMainPage()\">\n",
            SOFTWARECR,
            RefererHost, "/mondesi/-/js/mondesi.css",
            atoi(WsLibCgiVar("FORM_VIEW")),
            SOFTWAREVN,
            RefererHost, "/mondesi/-/js/mondesi.js",
            SoftwareID,
            MonitorMode == 2 ? "2" : "");

   cptr = NodeList;
   while (*cptr)
   {
      while (*cptr && isspace(*cptr)) cptr++;
      zptr = (sptr = HostUrl) + sizeof(HostUrl)-1;
      /* parse to the end of the scheme */
      while (*cptr && !isspace(*cptr) && *cptr != '/' && sptr < zptr)
         *sptr++ = *cptr++;
      if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      /* if no domain name */
      if (!strchr (HostUrl, '.'))
      {
         /* same host, just use the script component */
         sptr = HostUrl;
         for (aptr = CgiScriptNamePtr; *aptr && sptr < zptr; *sptr++ = *aptr++);
      }
      else
      /* if provides a script component */
      if (*cptr && !isspace(*cptr))
      {
         if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
         while (*cptr && !isspace(*cptr) && sptr < zptr) *sptr++ = *cptr++;
      }
      else
      {
         /* append a script component */
         for (aptr = CgiScriptNamePtr; *aptr && sptr < zptr; *sptr++ = *aptr++);
      }
      *sptr = '\0';
      while (*cptr && !isspace(*cptr)) cptr++;
      while (*cptr && isspace(*cptr)) cptr++;

      if (NodesAcross > 1)
      {
         if (!(NodeCount % NodesAcross))
         {
            if (!NodeCount)
               fputs ("<table class=\"nodes_across\">\n<tr><td>\n", stdout);
            else
               fputs ("</td></tr>\n<tr><td>\n", stdout);
         }
         else
            fputs ("</td><td>\n", stdout);
      }
      NodeCount++,

      fprintf (stdout,
"<iframe id=\"node%d\" class=\"node_iframe%s\" \
frameborder=\"0\" scrolling=\"no\" \
src=\"%s?iframe=node%d&nows=%d&view=%d&monitor=%d\"></iframe>\n",
               NodeCount,
               MonitorMode == 2 ? "2" : "",
               HostUrl,
               NodeCount,
               atoi(WsLibCgiVar("FORM_NOWS")) ? 1 : 0,
               atoi(WsLibCgiVar("FORM_VIEW")),
               MonitorMode);
   }

   if (NodesAcross > 1) fputs ("</td></tr>\n</table>\n", stdout);

   fprintf (stdout,
"<form id=\"form_config\" method=\"GET\" action=\"%s\" \
enctype=\"application/x-www-form-urlencoded\" autocomplete=\"off\">\n\
<input type=\"hidden\" name=\"across\" value=\"%d\">\n\
<input type=\"hidden\" name=\"expand\" value=\"%d\">\n\
<input type=\"hidden\" name=\"nodes\" value=\"%s\">\n\
<input type=\"hidden\" name=\"nows\" value=\"%d\">\n\
<input type=\"hidden\" name=\"monitor\" value=\"0\">\n\
<input id=\"page_configure\" type=\"submit\" \
class=\"btn btn2\" value=\" Configure \">\n\
</form>\n\
</body>\n\
</html>\n",
            CgiScriptNamePtr,
            NodesAcross,
            atoi(WsLibCgiVar("FORM_EXPAND")) ? 1 : 0,
            strchr(NodeList,'.') ? NodeList : "",
            atoi(WsLibCgiVar("FORM_NOWS")) ? 1 : 0);
} 

/*****************************************************************************/
/*
Provide the in-frame node display.
*/

void MonDeSiNode (char *IframeId)

{
   int  cnt,
        WebSocketOk;
   char  *cptr, *sptr, *zptr;
   char  DivStyle [256];
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   memset (clptr = &CgiClientData, 0, sizeof(CgiClientData));

   AccessControl (clptr);

   if (!clptr->AccessAllowed[ACCESS_GENERAL])
   {
      fputs (ACCESS_NOT_PERMITTED, stdout);
      return;
   }

   WebSocketOk = WebSockAvailable() && !atoi(WsLibCgiVar("FORM_NOWS"));

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<!DOCTYPE html>\n\
<html>\n\
<head>\n\
<meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">\n\
<meta name=\"copyright\" content=\"%s - GPL licensed\">\n\
\
<link rel=\"stylesheet\" type=\"text/css\" href=\"%s%s\">\n",
            SOFTWARECR,
            RefererHost, "/mondesi/-/js/mondesi.css");

   fprintf (stdout,
"<script language=\"JavaScript\">\n\
\
var WASDwebSocket = %s;\n\
var mdsCluster = %s;\n\
var mdsConfigView = %d;\n\
var mdsCpuCount = %d;\n\
var mdsDiskCount = %d;\n\
var mdsExeVersion = \"%s\";\n\
var mdsIframeId = \"%s\";\n\
var mdsNodeName = \"%s\";\n\
var mdsRefreshInterval = %d;\n\
var mdsRefererHost = \"%s\"\n\
var mdsScriptName = \"%s\";\n\
var mdsTopProc = %s;\n\
\
function mdsCheckJsLoad() {\n\
   if (typeof mdsMonitorNode == \'undefined\')\n\
      document.getElementById(\'node_notes\').innerHTML = \
\'JavaScript not loaded!  Check mapping.\';\n\
}\n\
setTimeout(\'mdsCheckJsLoad()\',5000);\n\
\
</script>\n\
<script language=\"JavaScript\" src=\"%s%s\"></script>\n\
<script language=\"JavaScript\" src=\"%s%s\"></script>\n\
\
<title>%s</title>\n\
</head>\n\
<body id=\"node_body\" class=\"node_body\" onload=\"mdsMonitorNode();\">\n",
            WebSocketOk ? "true" : "false",
            SystemData.ClusterMember ? "true" : "false",
            atoi(WsLibCgiVar("FORM_VIEW")),
            SystemData.ActiveCpuCount,
            SystemData.DiskCount,
            SOFTWAREVN,
            IframeId,
            SystemData.NodeName,
            RefreshInterval,
            RefererHost,
            CgiScriptNamePtr,
            clptr->AccessAllowed[ACCESS_TOP_PROCESS] ? "true" : "false",
            RefererHost, "/mondesi/-/js/mondesi.js",
            RefererHost, "/mondesi/-/js/wz_jsgraphics.js",
            SoftwareID);

   fprintf (stdout,
"<table id=\"node_block\">\n\
<tr><td id=\"node_notes\"></td>\
<td id=\"node_bars\" colspan=\"2\" rowspan=\"2\"></td></tr>\n\
<tr><td id=\"node_control\"></td></tr>\n\
<tr><td id=\"node_process\" colspan=\"3\"></td></tr>\n\
<tr><td id=\"per_proc_pid\" colspan=\"3\"></td></tr>\n\
<tr><td id=\"per_proc_hist\" colspan=\"3\"></td></tr>\n\
<tr><td id=\"node_hist\" colspan=\"3\"></td></tr>\n\
</table>\n");

   fputs ("</body>\n</html>\n", stdout);
} 

/*****************************************************************************/
/*
For an HTTP request.  At the defined interval gather dynamic system and
performance data and for appropriate (changed) data call the relevant
JavaScript (RPC) function in the parent panel (window) to process and graph
associated data.  Do the best we can to disable proxy caching ensuring a timely
flow of response data.
*/

void MonDeSiHtmlData ()

{
   int  status,
        HostUrlLength,
        TimeStamp;
   char  *cptr, *sptr, *zptr;
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MonDeSiHtmlData()\n");

   memset (clptr = &CgiClientData, 0, sizeof(CgiClientData));

   AccessControl (clptr);

   if (!clptr->AccessAllowed[ACCESS_GENERAL])
   {
      fputs (ACCESS_NOT_PERMITTED, stdout);
      return;
   }

   if (clptr->AccessAllowed[ACCESS_TOP_PROCESS])
   {
      clptr->ProcRMI = atoi(CgiQueryStringPtr);
      if (clptr->AccessAllowed[ACCESS_PER_PROCESS])
         if (CgiQueryStringPtr[0] && CgiQueryStringPtr[1])
            clptr->PerProcPid = strtol(CgiQueryStringPtr+2,NULL,16);
   }

   /* derive the host URL from the HTTP referrer field */
   cptr = CgiHttpRefererPtr;
   while (*cptr && *(unsigned short*)cptr != '//') cptr++;
   if (*cptr) cptr += 2;
   while (*cptr && *cptr != '/') cptr++;
   HostUrlLength = cptr - CgiHttpRefererPtr;

   /* onload= restarts if the long-poll page ever completes loading! */
   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Script-Control: X-content-encoding-gzip=0\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<html>\n\
<head>\n\
<meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">\n\
<meta name=\"copyright\" content=\"%s - GPL licensed\">\n\
<script language=\"JavaScript\">parent.mdsProcButton(%s,%s);</script>\
</head>\n\
<body onload=\"parent.mdsMonitorNode();\">\n",
            SOFTWARECR,
            /* [proc] button depending on access control */
            clptr->AccessAllowed[ACCESS_TOP_PROCESS] ? "true" : "false",
            clptr->AccessAllowed[ACCESS_PER_PROCESS] ? "true" : "false");
   fflush (stdout);

   for(;;) sys$hiber();
   /* should never ... */
} 

/*****************************************************************************/
/*
Used for both HTTP and WebSocket requests.
Writes the data update JavaScript into the buffer.
This buffered JavaScript is subsequently written to the client(s).
*/

void MonDeSiFormatData ()

{
   static $DESCRIPTOR (DateTimeFaoDsc, "!17%D\0");
   static $DESCRIPTOR (UpTimeFaoDsc, "!10%D\0");

   int  cnt, status,
        HostUrlLength;
   __int64  UpBinTime;
   char  *cptr, *sptr, *zptr,
         *TimeStampPtr;
   char  CpuActive [32],
         DateTime [32],
         MemSize [32],
         Stupid_njobcnt [16],
         UpTime [32],
         WasdUsing [128];
   $DESCRIPTOR (DateTimeDsc, DateTime);
   $DESCRIPTOR (UpTimeDsc, UpTime);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MonDeSiFormatData()\n");

   if (!SystemData.PrevInit)
   {
      SystemData.PrevInit = TRUE;

      /* initialise the previous value storage */
      SystemData.PrevBufferedIO = -1;
      SystemData.PrevBusyCpu = -1;
      SystemData.PrevClusterMsgSent = -1;
      SystemData.PrevClusterMsgRcvd = -1;
      SystemData.PrevClusterKbyteMap = -1;
      SystemData.PrevClusterKbyteSent = -1;
      SystemData.PrevClusterKbyteRcvd = -1;
      SystemData.PrevClusterNodes = 0;
      SystemData.PrevDirectIO = -1;
      SystemData.PrevNetIntBlocksRx = -1;
      SystemData.PrevNetIntBlocksTx = -1;
      SystemData.PrevNetIntBytesRx = -1;
      SystemData.PrevNetIntBytesTx = -1;
      SystemData.PrevFcpRead = -1;
      SystemData.PrevFcpWrite = -1;
      SystemData.PrevLockActivity = -1;
      SystemData.PrevLockWait = -1;
      SystemData.PrevMemInUse = -1;
      SystemData.PrevMinute = -1;
      SystemData.PrevMscpRead = -1;
      SystemData.PrevMscpWrite = -1;
      SystemData.PrevPageFaults = -1;
      SystemData.PrevPageFilePercent = -1;
      SystemData.PrevProcessCount = -1;
      SystemData.PrevProcessIntCount = -1;
      SystemData.PrevUserCpu = -1;

      if (SystemData.DiskCount)
         for (cnt = 0; cnt < DISK_DEVICE_MAX; cnt++)
            SystemData.DiskPrevOperCount[cnt] = -1;
   }

   zptr = (sptr = SystemData.OutputBuffer) +
          sizeof(SystemData.OutputBuffer)-256;

   if (SystemData.UsageCount != UsageCount ||
       SystemData.PrevMinute != CurrentNumTime[4] ||
       SystemData.ProcessCount != SystemData.PrevProcessCount ||
       SystemData.ClusterNodes != SystemData.PrevClusterNodes ||
       SystemData.PrevWebSockConnectedCount != WebSockConnectedCount ||
       NotePad[0])
   {
      SystemData.PrevMinute = CurrentNumTime[4];
      SystemData.UsageCount = UsageCount;
      SystemData.NodeInfoTimeStamp = TimeStamp;
      SystemData.PrevWebSockConnectedCount = WebSockConnectedCount;
      SystemData.SystemDelta++;
      SystemData.DiskDelta++;

      UpBinTime = SystemData.BootBinTime - CurrentBinTime;
      sys$fao (&DateTimeFaoDsc, 0, &DateTimeDsc, &CurrentBinTime);
      sys$fao (&UpTimeFaoDsc, 0, &UpTimeDsc, &UpBinTime);

      if (SystemData.ActiveCpuCount == SystemData.AvailCpuCount)
         sprintf (CpuActive, "%d", SystemData.ActiveCpuCount);
      else
         sprintf (CpuActive, "%d/%d", SystemData.ActiveCpuCount,
                                      SystemData.AvailCpuCount);

      if (SystemData.MemSize >= 1024)
         sprintf (MemSize, "%.1f&nbsp;GB", (float)SystemData.MemSize / 1024.0);
      else
         sprintf (MemSize, "%d&nbsp;MB", SystemData.MemSize);

      sptr += sprintf (sptr,
"if(window.mdsNodeNotes)the=window;else the=parent;\
%s\
the.mdsCpuCount=%d;\
the.mdsNodeNotes(\"<span class=\'node_name\'>%s</span>\
<span class=\'node_time\'>%s</span>\
<span class=\'node_text\'>\
%s %s&nbsp;CPU%s %s %s&nbsp;%s&nbsp <nobr>Up %s</nobr>",
                       /* if script restarts process monitor must be reset */ 
                       (signed int)SystemData.PrevProcessCount == -1 ?
                          "the.mdsExecStartup();" : "",
                       SystemData.ActiveCpuCount,
                       SystemData.NodeName,
                       DateTime,
                       SystemData.HwName,
                       CpuActive,
                       SystemData.AvailCpuCount > 1 ? "s" : "",
                       MemSize,
                       SystemData.ArchName,
                       SystemData.VmsVersion,
                       UpTime);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (SystemData.ClusterMember)
      {
         sptr += sprintf (sptr,
"&nbsp Cluster nodes:%d votes:%d expect:%d quorum:%d",
                          SystemData.ClusterNodes,
                          SystemData.ClusterVotes,
                          SystemData.ClusterEVotes,
                          SystemData.ClusterQuorum);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
         SystemData.PrevClusterNodes = SystemData.ClusterNodes;
      }

      if ((int)SystemData.ProcessNetCount >= 0)
         sprintf (Stupid_njobcnt, "%d", SystemData.ProcessNetCount);
      else
         strcpy (Stupid_njobcnt, "?");

      sptr += sprintf (sptr,
"&nbsp; Processes:%d interactive:%d batch:%d network:%s",
                       SystemData.ProcessCount,
                       SystemData.ProcessIntCount,
                       SystemData.ProcessBatCount,
                       Stupid_njobcnt);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (SystemData.DiskCount)
      {
         sptr += sprintf (sptr, "&nbsp; Disks:%d/%d",
                          SystemData.MountCount, SystemData.DiskCount);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
      }

      if ((int)SystemData.PageFilePercent >= 50)
      {
         sptr += sprintf (sptr, "&nbsp; <span%s>PGFL:%d%%</span>",
                          (int)SystemData.PageFilePercent >= 85 ?
                             " style=\'font-weight:bold;color:red;\'" : "",
                          SystemData.PageFilePercent);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
      }

      if (WasdHttpd)
      {
         if (WebSockConnectedCount)
            sprintf (WasdUsing, "%s/WebSocket:%d",
                     UsingCGI ? "/CGI" : "", WebSockConnectedCount);
         else
            strcpy (WasdUsing, "/CGI");
      }
      else
         WasdUsing[0] = '\0';

      sptr += sprintf (sptr,
"&nbsp; <span style=\'color:gray\'>(v%s/%s%s%s%s%s)</span>",
                       SOFTWAREVN,
                       ApacheHttpd ? "Apache" : "",
                       OsuHttpd ? "OSU" : "",
                       WasdHttpd ? "WASD" : "",
                       WasdUsing,
                       !(ApacheHttpd || OsuHttpd || WasdHttpd) ?
                          "?httpd?" : "");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (NotePad[0])
      {
         sptr += sprintf (sptr, "&nbsp; <span style=\'color:red;\'>%s</span>",
                          NotePad[0] == '$' ? NotePad+1 : NotePad);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
         if (NotePad[0] != '$') NotePad[0] = '\0';
      }

      sptr += sprintf (sptr, "</span>\");");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.NodeInfoLength = sptr - SystemData.OutputBuffer;

      /* ensure memory in use and page file is refreshed */
      SystemData.PrevMemInUse = SystemData.PrevPageFilePercent = -1;
   }
   else
   {
      /* leave the node information alone (persists for up to a minute) */
      sptr += SystemData.NodeInfoLength;
   }

   /* only after enough samples to make it representative */
   if (CollectDataSeconds <= 2)
   {
      SystemData.MonValLength = 0;
      return;
   }

   sptr += sprintf (sptr, "the.mdsTimeStamp=%u;", TimeStamp);
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
   TimeStampPtr = sptr;

   if (SystemData.SystemDelta ||
       SystemData.CpuBusyPercent != SystemData.PrevBusyCpu ||
       SystemData.CpuUserPercent != SystemData.PrevUserCpu)
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_cpu\",%u,%u,%u);",
                       SystemData.ActiveCpuCount*100,
                       SystemData.CpuBusyPercent,
                       SystemData.CpuBusyPercent - SystemData.CpuUserPercent);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevBusyCpu = SystemData.CpuBusyPercent;
      SystemData.PrevUserCpu = SystemData.CpuUserPercent;
   }

   if (SystemData.SystemDelta ||
       SystemData.MemInUse != SystemData.PrevMemInUse)
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_mem\",%llu,%llu);",
                       (__int64)(SystemData.MemSize) << 20,
                       (__int64)(SystemData.MemInUse) << 20);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevMemInUse = SystemData.MemInUse;
   }

   if (SystemData.SystemDelta ||
       SystemData.PageFilePercent != SystemData.PrevPageFilePercent)
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_pgfl\",%llu,%llu);",
                       (__int64)SystemData.PageFilePage *
                       (__int64)SystemData.PageSize,
                       ((__int64)SystemData.PageFilePage *
                        (__int64)SystemData.PageSize) -
                       ((__int64)SystemData.PageFileFree *
                        (__int64)SystemData.PageSize));
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevPageFilePercent = SystemData.PageFilePercent;
   }

   if (SystemData.SystemDelta ||
       SystemData.PageFaults != SystemData.PrevPageFaults)
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_flts\",-1,%u,%u);",
                       SystemData.PageFaultsSoft + SystemData.PageFaultsHard,
                       SystemData.PageFaultsHard);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevPageFaults = SystemData.PageFaults;
   }

   if (SystemData.BufferedIO != SystemData.PrevBufferedIO)
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_bio\",-1,%u);",
                       SystemData.BufferedIO);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevBufferedIO = SystemData.BufferedIO;
   }

   if (SystemData.SystemDelta ||
       SystemData.DirectIO != SystemData.PrevDirectIO)
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_dio\",-1,%u);",
                       SystemData.DirectIO);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevDirectIO = SystemData.DirectIO;
   }

   if (SystemData.SystemDelta ||
       SystemData.LockActivity != SystemData.PrevLockActivity ||
       SystemData.LockWait != SystemData.PrevLockWait)
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_lock\",-1,%u,%u);",
                       SystemData.LockActivity,
                       SystemData.LockWait);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevLockActivity = SystemData.LockActivity;
      SystemData.PrevLockWait = SystemData.LockWait;
   }

   if (SystemData.SystemDelta ||
       (SystemData.NetIntCount &&
        (SystemData.NetIntBytesRx != SystemData.PrevNetIntBytesRx ||
         SystemData.NetIntBytesTx != SystemData.PrevNetIntBytesTx)))
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_ni\",-1,%llu,%llu);",
                       SystemData.NetIntBytesRx + SystemData.NetIntBytesTx,
                       SystemData.NetIntBytesRx);

      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevNetIntBytesRx = SystemData.NetIntBytesRx;
      SystemData.PrevNetIntBytesTx = SystemData.NetIntBytesTx;

      if (SystemData.NetIntErrorsSoft || SystemData.NetIntErrorsHard)
      {
         sptr += sprintf (sptr,
"the.mdsDataTitle(\"node_ni\",\"NI\",\"\",\"ERR: %d/%d\");",
                          (int)SystemData.NetIntErrorsSoft,
                          (int)SystemData.NetIntErrorsHard);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
      }
   }

   if (SystemData.SystemDelta ||
       SystemData.FcpRead != SystemData.PrevFcpRead ||
       SystemData.FcpWrite != SystemData.PrevFcpWrite)
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_disk\",-1,%u,%u);",
                       SystemData.FcpRead + SystemData.FcpWrite,
                       SystemData.FcpWrite);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevFcpRead = SystemData.FcpRead;
      SystemData.PrevFcpWrite = SystemData.FcpWrite;
   }

   if (SystemData.SystemDelta ||
       SystemData.ProcessCount != SystemData.PrevProcessCount ||
       SystemData.ProcessIntCount != SystemData.PrevProcessIntCount)
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_prc\",-1,%u,%u);",
                       SystemData.ProcessCount,
                       SystemData.ProcessIntCount);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevProcessCount = SystemData.ProcessCount;
      SystemData.PrevProcessIntCount = SystemData.ProcessIntCount;
   }

   if (SystemData.SystemDelta ||
       (SystemData.ClusterMember &&
        (SystemData.ClusterMsgRcvd != SystemData.PrevClusterMsgRcvd ||
         SystemData.ClusterMsgSent != SystemData.PrevClusterMsgSent)))
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_scsmg\",-1,%u,%u);",
                       SystemData.ClusterMsgRcvd + SystemData.ClusterMsgSent,
                       SystemData.ClusterMsgSent);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevClusterMsgRcvd = SystemData.ClusterMsgRcvd;
      SystemData.PrevClusterMsgSent = SystemData.ClusterMsgSent;
   }

   if (SystemData.ClusterMember &&
       (SystemData.ClusterKbyteMap != SystemData.PrevClusterKbyteMap))
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_scskb\",-1,%u,%u);",
                       SystemData.ClusterKbyteMap,
                       0);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevClusterKbyteMap = SystemData.ClusterKbyteMap;
   }

   if (SystemData.PrevMinute != CurrentNumTime[4] ||
       (SystemData.ClusterMember &&
        (SystemData.MscpRead != SystemData.PrevMscpRead ||
         SystemData.MscpWrite != SystemData.PrevMscpWrite)))
   {
      sptr += sprintf (sptr,
                       "the.mdsDatum(\"node_mscp\",-1,%u,%u);",
                       SystemData.MscpRead + SystemData.MscpWrite,
                       SystemData.MscpWrite);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      SystemData.PrevMscpRead = SystemData.MscpRead;
      SystemData.PrevMscpWrite = SystemData.MscpWrite;
   }

   if (SystemData.DiskCount)
   {
      if (SystemData.PrevMinute != CurrentNumTime[4] ||
          SystemData.DiskDelta)
      {
         SystemData.DiskDelta = 0;

         sptr += sprintf (sptr, "the.mdsDiskMount=%d;", SystemData.MountCount);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         for (cnt = 0; cnt < SystemData.DiskCount; cnt++)
         {
            sptr += sprintf (sptr,
"the.mdsDataTitle(\"node_disk_%d\",\"%s %s\",\"%s\",\"%s\");",
                             cnt,
                             SystemData.DiskName[cnt],
                             SystemData.DiskVolName[cnt],
                             SystemData.DiskUsage[cnt],
                             SystemData.DiskAlert[cnt]);
            if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
         }
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
      }

      for (cnt = 0; cnt < SystemData.DiskCount; cnt++)
      {
         if (SystemData.DiskOperCount[cnt] != SystemData.DiskPrevOperCount[cnt])
         {
            sptr += sprintf (sptr,
                             "the.mdsDatum(\"node_disk_%d\",-1,%u,%u);",
                             cnt,
                             SystemData.DiskOperCount[cnt],
                             0);
            if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

            SystemData.DiskPrevOperCount[cnt] = SystemData.DiskOperCount[cnt];
         }
      }
   }

   /* if no data JavaScript follows the timestamp */
   if (sptr == TimeStampPtr) *(sptr = TimeStampPtr) = '\0';

   SystemData.MonValLength = sptr - (SystemData.OutputBuffer+
                                     SystemData.NodeInfoLength);

   SystemData.SystemDelta = 0;
}

/*****************************************************************************/
/*
Indepdendent, timer AST driven update of the system performance data.
This will support multiple clients more efficiently.
The updated data are then read by other indepdendent client threads.
*/

void CollectData ()

{
   static const long cvtf_mode = LIB$K_DELTA_SECONDS_F;
   static __int64  PrevBinTime,
                   SecondsDelta;

   int  status;
   unsigned long  fltf;  /* generic 32 bits for the VAX float */
   __int64  DiffBinTime;
   struct WsLibStruct  *wsctx, *wsptr;
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectData()\n");

   /* one to sixty seconds, just multiply by interval */
   if (!SecondsDelta) SecondsDelta = RefreshInterval * -10000000;

   sys$gettim (&CurrentBinTime);
   sys$numtim (&CurrentNumTime, &CurrentBinTime);
   TimeStamp = decc$fix_time (&CurrentBinTime);

   if (ExitTimeStamp && TimeStamp > ExitTimeStamp) exit (SS$_NORMAL); 

   /* calculate a floating point delta time */
   if (!PrevBinTime) PrevBinTime = CurrentBinTime;
   DiffBinTime = PrevBinTime - CurrentBinTime;
   PrevBinTime = CurrentBinTime;
#ifndef __ia64
   /* Alpha 7.3-1 does not have lib$cvts_..() */
   lib$cvtf_from_internal_time (&cvtf_mode, &fltf, &DiffBinTime);
   /* convert VAX to IEEE float */
   status = cvt$convert_float (&fltf, CVT$K_VAX_F,
                               &DeltaSeconds, CVT$K_IEEE_S,
                               CVT$M_ROUND_TO_NEAREST);
   if (VMSnok(status)) EXIT_FI_LI (status);
#else
   lib$cvts_from_internal_time (&cvtf_mode, &DeltaSeconds, &DiffBinTime);
#endif
   if (Debug) fprintf (stdout, "DeltaSeconds: %f\n", DeltaSeconds);

   if (UsingCGI) clptr = &CgiClientData;

   /* determine whether process data should be collected (expensive!) */
   if (UsingCGI)
      SystemData.ProcRMI = clptr->ProcRMI;
   else
   {
      /* WebSocket access */
      SystemData.ProcRMI = 0;
      wsctx = NULL;
      while (wsptr = WsLibNext(&wsctx))
      {
         clptr = WsLibGetUserData (wsptr);
         if (clptr->ProcRMI) SystemData.ProcRMI = clptr->ProcRMI;
      }
   }

   CollectSyiDynamic ();
   CollectRMI ();
   CollectNetInt ();
   CollectDiskSPI ();
   (*SystemData.ProcSpiFunction)();

   MonDeSiFormatData ();

   if (UsingCGI)
   {
      /* CGI(plus) access */
      if (SystemData.MonValLength ||
          CgiNodeInfoTimeStamp != SystemData.NodeInfoTimeStamp)
      {
         fputs ("<script language=\"JavaScript\">", stdout);

         if (CgiNodeInfoTimeStamp != SystemData.NodeInfoTimeStamp)
         {
            fputs (SystemData.OutputBuffer, stdout);
            CgiNodeInfoTimeStamp = SystemData.NodeInfoTimeStamp;
         }
         else
         if (SystemData.MonValLength)
            fputs (SystemData.OutputBuffer+SystemData.NodeInfoLength, stdout);

         if (clptr->ProcRMI)
         {
            if (clptr->AccessAllowed[ACCESS_TOP_PROCESS])
            {
               if (SystemData.ProcBufferLength)
               {
                  if (clptr->AccessAllowed[ACCESS_PER_PROCESS])
                  {
                     strcpy (SystemData.ProcBuffer+SystemData.ProcBufferLength,
                             "the.mdsMonProcPid();");
                     SystemData.ProcBufferLength += 20;
                  }
                  fputs (SystemData.ProcBuffer, stdout);
               }

               if (clptr->PerProcPid)
                  if (clptr->AccessAllowed[ACCESS_PER_PROCESS])
                  {
                     CollectPerProc (clptr);
                     if (clptr->PerProcBufferLength)
                        fputs (clptr->PerProcBuffer, stdout);
                  }
             }
         }

         fputs ("</script>", stdout);
         fflush (stdout);
      }
   }

   /* WebSocket access */
   wsctx = NULL;
   while (wsptr = WsLibNext(&wsctx))
   {
      clptr = WsLibGetUserData (wsptr);

      if (clptr->NodeInfoTimeStamp != SystemData.NodeInfoTimeStamp)
      {
         WsLibWrite (wsptr,
                     SystemData.OutputBuffer,
                     SystemData.NodeInfoLength+SystemData.MonValLength,
                     WSLIB_ASYNCH);
         clptr->NodeInfoTimeStamp = SystemData.NodeInfoTimeStamp;
      }
      else
      if (SystemData.MonValLength)
         WsLibWrite (wsptr,
                     SystemData.OutputBuffer+SystemData.NodeInfoLength,
                     SystemData.MonValLength,
                     WSLIB_ASYNCH);

      if (clptr->ProcRMI)
      {
         if (clptr->AccessAllowed[ACCESS_TOP_PROCESS])
         {
            if (SystemData.ProcBufferLength)
            {
               if (clptr->AccessAllowed[ACCESS_PER_PROCESS])
               {
                  strcpy (SystemData.ProcBuffer+SystemData.ProcBufferLength,
                          "the.mdsMonProcPid();");
                  SystemData.ProcBufferLength += 20;
               }
               WsLibWrite (wsptr,
                           SystemData.ProcBuffer,
                           SystemData.ProcBufferLength,
                           WSLIB_ASYNCH);
            }

            if (clptr->PerProcPid)
            {
               if (clptr->AccessAllowed[ACCESS_PER_PROCESS])
               {
                  CollectPerProc (clptr);
                  if (clptr->PerProcBufferLength)
                     WsLibWrite (wsptr,
                                 clptr->PerProcBuffer,
                                 clptr->PerProcBufferLength,
                                 WSLIB_ASYNCH);
               }
            }
         }
      }
   }

   SystemData.PrevMinute = CurrentNumTime[4];

   status = sys$setimr (0, &SecondsDelta, CollectData, NULL, 0);

   if (VMSnok(status)) EXIT_FI_LI (status);

   CollectDataSeconds += RefreshInterval;
}

/*****************************************************************************/
/*
Per-boot system data.
*/ 

int CollectSyiStatic ()

{
   static unsigned long  ActiveCpuCnt,
                         AvailCpuCount,
                         MemSize,
                         PageSize;
   static __int64  BootBinTime;
   static unsigned short  HwNameLength;
   static unsigned char  ClusterMember;
   static char  ArchName [15+1],
                HwName [1+60],
                NodeName [15+1],
                VmsVersion [8+1];

   /* initialize structure for SYI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList[] =
   {
       { sizeof(AvailCpuCount), SYI$_AVAILCPU_CNT, &AvailCpuCount, 0 },
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(ArchName), SYI$_ARCH_NAME, &ArchName, 0 },
       { sizeof(BootBinTime), SYI$_BOOTTIME, &BootBinTime, 0 },
       { sizeof(ClusterMember), SYI$_CLUSTER_MEMBER, &ClusterMember, 0 },
       { sizeof(HwName), SYI$_HW_NAME, &HwName, &HwNameLength },
       { sizeof(MemSize), SYI$_MEMSIZE, &MemSize, 0 },
       { sizeof(NodeName), SYI$_SCSNODE, &NodeName, 0 },
       { sizeof(PageSize), SYI$_PAGE_SIZE, &PageSize, 0 },
       { sizeof(VmsVersion), SYI$_VERSION, &VmsVersion, 0 },
       {0,0,0,0}
   };

   int  status;
   char  *cptr, *sptr, *zptr;
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectSyiStatic()\n");

   status = sys$getsyiw (
            0,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            &ItemList,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getsyi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   if (VMSnok (status)) return (status);

   zptr = (sptr = SystemData.ArchName) + sizeof(SystemData.ArchName)-1;
   for (cptr = ArchName;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   HwName[HwNameLength] = '\0';
   zptr = (sptr = SystemData.HwName) + sizeof(SystemData.HwName)-1;
   for (cptr = HwName; *cptr && sptr < zptr; *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = SystemData.NodeName) + sizeof(SystemData.NodeName)-1;
   for (cptr = NodeName;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = SystemData.VmsVersion) + sizeof(SystemData.VmsVersion)-1;
   for (cptr = VmsVersion;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   if (!(SystemData.ClusterMember = ClusterMember))
      if (cptr = getenv("MONDESI_CLUSTER_MEMBER"))
         SystemData.ClusterMember = atoi(cptr);

   /* convert to MBytes */
   SystemData.MemSize = MemSize >> 7;
   SystemData.PageSize = PageSize;

   SystemData.BootBinTime = BootBinTime;

   SystemData.AvailCpuCount = AvailCpuCount;
   SystemData.ActiveCpuCount = ActiveCpuCnt;

   if (Debug)
      fprintf (stdout, "|%s|%s|%s|%s| %d %d\n",
               SystemData.ArchName, SystemData.HwName,
               SystemData.NodeName, SystemData.VmsVersion,
               SystemData.AvailCpuCount, ClusterMember);

   return (status);
} 

/*****************************************************************************/
/*
Dynamic system data.
*/ 

int CollectSyiDynamic ()

{
   static unsigned long  ActiveCpuCnt,
                         PageFileFree,
                         PageFilePage;
   static unsigned short  ClusterEVotes,
                          ClusterNodes,
                          ClusterQuorum,
                          ClusterVotes;

   /* initialize structure for SYI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ClusterItemList[] =
   {
       { sizeof(ClusterNodes), SYI$_CLUSTER_NODES, &ClusterNodes, 0 },
       { sizeof(ClusterQuorum), SYI$_CLUSTER_QUORUM, &ClusterQuorum, 0 },
       { sizeof(ClusterVotes), SYI$_CLUSTER_VOTES, &ClusterVotes, 0 },
       { sizeof(ClusterEVotes), SYI$_CLUSTER_EVOTES, &ClusterEVotes, 0 },
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(PageFileFree), SYI$_PAGEFILE_FREE, &PageFileFree, 0 },
       { sizeof(PageFilePage), SYI$_PAGEFILE_PAGE, &PageFilePage, 0 },
       {0,0,0,0}
   }, NonItemList[] =
   {
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(PageFileFree), SYI$_PAGEFILE_FREE, &PageFileFree, 0 },
       { sizeof(PageFilePage), SYI$_PAGEFILE_PAGE, &PageFilePage, 0 },
       {0,0,0,0}
   };

   int  status;
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectSyiDynamic()\n");

   /* collect System Information */
   status = sys$getsyiw (
            0,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            SystemData.ClusterMember ?
               (void*)&ClusterItemList :
               (void*)&NonItemList,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getsyi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status))
   {
      sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
      return (status);
   }

   if (SystemData.ClusterMember)
   {
      SystemData.ClusterEVotes = ClusterEVotes;
      SystemData.ClusterNodes = ClusterNodes;
      SystemData.ClusterQuorum = ClusterQuorum;
      SystemData.ClusterVotes = ClusterVotes;
   }

   SystemData.ActiveCpuCount = ActiveCpuCnt;

   SystemData.PageFilePercent = 100 - (PageFileFree * 100 / PageFilePage);
   SystemData.PageFileFree = PageFileFree;
   SystemData.PageFilePage = PageFilePage;

   return (status);
} 

/*****************************************************************************/
/*
Collect System Performance Information.

Uses the system service $GETRMI() available for VMS V7.3 and later.

Cannot use RMI$_CPU* or RMI$_PROC*CNT on VMS V7.3-1 (and V7.3?) because they
don't seem to exist!  Fall back to handcrafting CPU from RMI$_MODES and
sys$gl_*jobcnt (many thanks to Graham Burley for the suggestion).  This
approach maintains two item lists and if the V7.3-2 (and later) fails with a
SS$$_BADPARAM switches to the other list.  Subsequent alternate processing is
performed by testing which list is in use.
*/ 

int CollectRMI ()

{
/* required for VMS V7.3-1 and earlier */
#define INTSTK_MODE  0
#define MPSYNCH_MODE 1
#define KERNEL_MODE  2
#define EXEC_MODE    3
#define SUPER_MODE   4
#define USER_MODE    5
/* 6 is reserved */
#define IDLE_MODE    7

#define MODE_COUNT IDLE_MODE+1
#define CPU_MAX             32

   /* structure for a single CPU's mode ticks */
   static struct SingleCPU {
       unsigned char  CPUid;
       unsigned long  Count [MODE_COUNT];
   };

   /* a structure all CPU's mode ticks for a system with up to 32 CPUs */
   static struct {
       unsigned long  NumberOfCPUs;
       struct SingleCPU  CPU [CPU_MAX];
   } ModeTicks;

   static unsigned long  BufferedIO,
                         ClusterMsgSent,
                         ClusterMsgRcvd,
                         ClusterKbyteMap,
                         ClusterKbyteSent,
                         ClusterKbyteRcvd,
                         DirectIO,
                         FcpRead,
                         FcpWrite,
                         LockConvert,
                         LockDequeue,
                         LockEnqueue,
                         LockWait,
                         MscpRead,
                         MscpWrite,
                         PageFaults,
                         PageReadIO,
                         PageWriteIO,
                         ProcessCount,
                         ProcessBatCount,
                         ProcessIntCount,
                         ProcessNetCount,
                         ScsDataSize;

   static unsigned long  PrevBufferedIO,
                         PrevClusterMsgSent,
                         PrevClusterMsgRcvd,
                         PrevClusterKbyteMap,
                         PrevClusterKbyteSent,
                         PrevClusterKbyteRcvd,
                         PrevDirectIO,
                         PrevFcpRead,
                         PrevFcpWrite,
                         PrevLockActivity,
                         PrevLockWait,
                         PrevMscpRead,
                         PrevMscpWrite,
                         PrevPageFaults,
                         PrevPageReadIO,
                         PrevPageWriteIO,
                         PrevMscpIO;

   static unsigned __int64  CpuExec,
                            CpuIntStk,
                            CpuKernel,
                            CpuMpSynch,
                            CpuSuper,
                            CpuUser;

   static unsigned long  MscpData [35];

   static unsigned __int64  PrevCpuTicksBusy,
                            PrevCpuTicksUser;

   static unsigned long  *ScsDataPtr;

#ifdef __ia64
   static unsigned __int64  FreeList,
                            ModList;
#else
   static unsigned long  FreeList,
                         ModList;
#endif

   /* initialize structure for RMI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList1[] =
   {
       /* for 7.3-2(?) and later which has the RMI$_CPU_... items */
       { sizeof(BufferedIO), RMI$_BUFIO, &BufferedIO, 0 },
       { sizeof(CpuExec), RMI$_CPUEXEC, &CpuExec, 0 },
       { sizeof(CpuIntStk), RMI$_CPUINTSTK, &CpuIntStk, 0 },
       { sizeof(CpuKernel), RMI$_CPUKERNEL, &CpuKernel, 0 },
       { sizeof(CpuMpSynch), RMI$_CPUMPSYNCH, &CpuMpSynch, 0 },
       { sizeof(CpuSuper), RMI$_CPUSUPER, &CpuSuper, 0 },
       { sizeof(CpuUser), RMI$_CPUUSER, &CpuUser, 0 },
       { sizeof(LockDequeue), RMI$_DEQ, &LockDequeue, 0 },
       { sizeof(DirectIO), RMI$_DIRIO, &DirectIO, 0 },
       { sizeof(LockConvert), RMI$_ENQCVT, &LockConvert, 0 },
       { sizeof(LockEnqueue), RMI$_ENQNEW, &LockEnqueue, 0 },
       { sizeof(LockWait), RMI$_ENQWAIT, &LockWait, 0 },
       { sizeof(FcpRead), RMI$_FCPREAD, &FcpRead, 0 }, 
       { sizeof(FcpWrite), RMI$_FCPWRITE, &FcpWrite, 0 }, 
       { sizeof(FreeList), RMI$_FRLIST, &FreeList, 0 }, 
       { sizeof(ModList), RMI$_MODLIST, &ModList, 0 },
       { sizeof(MscpData), RMI$_MSCP_EVERYTHING, &MscpData, 0 },
       { sizeof(PageFaults), RMI$_FAULTS, &PageFaults, 0 },
       { sizeof(PageReadIO), RMI$_PREADIO, &PageReadIO, 0 },
       { sizeof(PageWriteIO), RMI$_PWRITIO, &PageWriteIO, 0 },
       { sizeof(ProcessCount), RMI$_PROCS, &ProcessCount, 0 },
       { sizeof(ProcessIntCount), RMI$_PROCINTCNT, &ProcessIntCount, 0 },
       { sizeof(ProcessBatCount), RMI$_PROCBATCNT, &ProcessBatCount, 0 },
       { sizeof(ProcessNetCount), RMI$_PROCNETCNT, &ProcessNetCount, 0 },
       {0,0,0,0}
   }, ItemList2[] =
   {
       /* for 7.3-1(?) and earlier which do not have RMI$_CPU_... items */
       { sizeof(BufferedIO), RMI$_BUFIO, &BufferedIO, 0 },
       { sizeof(ModeTicks), RMI$_MODES, &ModeTicks, 0 },
       { sizeof(LockDequeue), RMI$_DEQ, &LockDequeue, 0 },
       { sizeof(DirectIO), RMI$_DIRIO, &DirectIO, 0 },
       { sizeof(LockConvert), RMI$_ENQCVT, &LockConvert, 0 },
       { sizeof(LockEnqueue), RMI$_ENQNEW, &LockEnqueue, 0 },
       { sizeof(LockWait), RMI$_ENQWAIT, &LockWait, 0 },
       { sizeof(FcpRead), RMI$_FCPREAD, &FcpRead, 0 }, 
       { sizeof(FcpWrite), RMI$_FCPWRITE, &FcpWrite, 0 }, 
       { sizeof(FreeList), RMI$_FRLIST, &FreeList, 0 }, 
       { sizeof(ModList), RMI$_MODLIST, &ModList, 0 },
       { sizeof(PageFaults), RMI$_FAULTS, &PageFaults, 0 },
       { sizeof(PageReadIO), RMI$_PREADIO, &PageReadIO, 0 },
       { sizeof(PageWriteIO), RMI$_PWRITIO, &PageWriteIO, 0 },
       { sizeof(ProcessCount), RMI$_PROCS, &ProcessCount, 0 },
       {0,0,0,0}
   }, ItemList3[] =
   {
       { 0, RMI$_SCS, 0, 0 },
       { sizeof(MscpRead), RMI$_MSCP_READ, &MscpRead, 0 },
       { sizeof(MscpWrite), RMI$_MSCP_WRITE, &MscpWrite, 0 },
       {0,0,0,0}
   }, *ItemListPtr = ItemList1;

   int  cidx, midx, status, tlong;
   unsigned long  datsize;
   unsigned __int64  CpuTicksBusy;
   unsigned long  *datptr;
   float  tfloat;
   IOSBLK  IOsb,
           SynchIOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectRMI()\n");

   /* collect System Performance Information */
   status = sys$getrmi (
            SpiEfn,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            ItemListPtr,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getrmi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   /* fall back to using the pre VMS V7.3-2 items */ 
   if (status == SS$_BADPARAM || IOsb.iosb$w_status == SS$_BADPARAM)
   {
      if (ItemListPtr == ItemList1)
      {
         ItemListPtr = ItemList2;
         return (status);
      }
   }

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status))
   {
      sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
      return (status);
   }

/*
   MGD 16-NOV-2013  Hmmm.  The non-USER RMI$_CPU.. items seem to work for
   Alpha (8.3 and 8.4) but for Itanium (V8.3 and V8.4) return zero values.
   So, look for this and just revert to using RMI$_MODES if noticed.
*/

   if (ItemListPtr == ItemList1)
   {
      if (!(CpuExec + CpuIntStk + CpuKernel + CpuMpSynch + CpuSuper))
      {
         ItemListPtr = ItemList2;
         return (status);
      }
   }

   if (SystemData.ClusterMember)
   {
      /* collect cluster information */
      datsize = ((SystemData.ClusterNodes * 14) + 1) * sizeof(unsigned long);
      if (Debug) fprintf (stdout, "%d %d\n", SystemData.ClusterNodes, datsize);
      if (ScsDataSize < datsize)
      {
         if (ScsDataPtr) free (ScsDataPtr);
         ScsDataPtr = (unsigned long*)calloc (1, ScsDataSize = datsize);
         if (!ScsDataPtr) EXIT_FI_LI (vaxc$errno);
      }
      ItemList3[0].BufferPtr = ScsDataPtr;
      ItemList3[0].BufferLength = ScsDataSize;

      /* does not appear to be supported by $GETRMI */
      status = exe$getspi (
               SpiEfn,  /* efn */
               0,  /* csiaddr */
               0,  /* nodename */
               &ItemList3,  /* item list */
               &IOsb,  /* iosb */
               0,      /* astaddr */
               0 );    /* astprm */

      if (Debug)
         fprintf (stdout, "exe$getspi() %%X%08.08X  IOsb: %%X%08.08X\n",
                  status, IOsb.iosb$w_status);

      if (VMSnok (status))
      {
         sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
         return (status);
      }

      /*** huh? ***/
      ScsDataPtr = ItemList3[0].BufferPtr;

      if (Debug) fprintf (stdout, "%d\n", ScsDataPtr[0]);
      if (ScsDataPtr[0] > SystemData.ClusterNodes)  EXIT_FI_LI (SS$_BUGCHECK);

      for (cidx = 0; cidx < ScsDataPtr[0]; cidx++)
      {
         datptr = &(ScsDataPtr[(cidx*14)+1]);
         if (Debug)
            fprintf (stdout, "%d |%s| %d %d %d %d %d %d %d %d %d %d %d\n",
                     cidx, (char*)(&datptr[0])+1,
                     datptr[2], datptr[3], datptr[4], datptr[5],
                     datptr[6], datptr[7], datptr[8], datptr[9],
                     datptr[10], datptr[11], datptr[12]);
         ClusterMsgSent = datptr[5];
         ClusterMsgRcvd = datptr[6];
         ClusterKbyteSent = datptr[8];
         ClusterKbyteRcvd = datptr[10];
         ClusterKbyteMap = datptr[11];
      }
  }

/***
   status = sys$synch (SpiEfn, &SynchIOsb);

   if (Debug)
      fprintf (stdout, "sys$synch() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, SynchIOsb.iosb$w_status);

   if (VMSnok (status)) return (status);
***/

   /*******/
   /* CPU */
   /*******/

   if (ItemListPtr == ItemList1)
   {
      /* VMS 7.3-2 and later */
      CpuTicksBusy = CpuExec + CpuIntStk + CpuKernel +
                     CpuMpSynch + CpuSuper + CpuUser;
   }
   else
   {
     /* VMS V7.3-1 and earlier */
     int  cidx, midx;
     for (cidx = 1; cidx < ModeTicks.NumberOfCPUs; cidx++)
        for (midx = INTSTK_MODE; midx <= IDLE_MODE; midx++)
            ModeTicks.CPU[0].Count[midx] += ModeTicks.CPU[cidx].Count[midx];

     CpuTicksBusy = CpuUser = 0;
     CpuTicksBusy = ModeTicks.CPU[0].Count[INTSTK_MODE];
     /* interrupt item is actually the sum of interrupt and idle! */
     CpuTicksBusy -= ModeTicks.CPU[0].Count[IDLE_MODE];
     CpuTicksBusy += ModeTicks.CPU[0].Count[KERNEL_MODE] +
                     ModeTicks.CPU[0].Count[MPSYNCH_MODE] +
                     ModeTicks.CPU[0].Count[EXEC_MODE] +
                     ModeTicks.CPU[0].Count[SUPER_MODE] +
                     ModeTicks.CPU[0].Count[USER_MODE];
     CpuUser = ModeTicks.CPU[0].Count[USER_MODE];

     ProcessIntCount = sys$gl_ijobcnt;
     ProcessBatCount = sys$gl_bjobcnt;
     ProcessNetCount = sys$gl_njobcnt;
     /* of course sys$g*_njobcnt doesn't seem to account network jobs! */
     ProcessNetCount = -1;
   }

   if (PrevCpuTicksBusy)
   {
      SystemData.CpuTicksBusy = CpuTicksBusy - PrevCpuTicksBusy;
      tfloat = (float)SystemData.CpuTicksBusy;
      tfloat /= DeltaSeconds;
      SystemData.CpuBusyPercent = (unsigned int)tfloat;
      if (SystemData.CpuBusyPercent > 100 * SystemData.ActiveCpuCount)
         SystemData.CpuBusyPercent = 100 * SystemData.ActiveCpuCount;
   }
   else
      SystemData.CpuBusyPercent = 0;

   PrevCpuTicksBusy = CpuTicksBusy;

   if (PrevCpuTicksUser)
   {
      SystemData.CpuTicksUser = CpuUser - PrevCpuTicksUser;
      tfloat = (float)SystemData.CpuTicksUser;
      tfloat /= DeltaSeconds;
      SystemData.CpuUserPercent = (unsigned int)tfloat;
      if (SystemData.CpuUserPercent > 100 * SystemData.ActiveCpuCount)
         SystemData.CpuUserPercent = 100 * SystemData.ActiveCpuCount;
   }
   else
      SystemData.CpuUserPercent = 0;

   PrevCpuTicksUser = CpuUser;

   if (Debug)
      fprintf (stdout, "CPU: %d%% USER: %d%%\n",
               SystemData.CpuBusyPercent, SystemData.CpuUserPercent);

   /***************/
   /* page faults */
   /***************/

   if (PrevPageFaults && PageFaults > PrevPageFaults)
      SystemData.PageFaults = PageFaults - PrevPageFaults;
   else
      SystemData.PageFaults = 0;
   PrevPageFaults = PageFaults;

   if (PrevPageReadIO && PageReadIO > PrevPageReadIO)
      SystemData.PageReadIO = PageReadIO - PrevPageReadIO;
   else
      SystemData.PageReadIO = 0;
   PrevPageReadIO = PageReadIO;

   if (PrevPageWriteIO && PageWriteIO > PrevPageWriteIO)
      SystemData.PageWriteIO = PageWriteIO - PrevPageWriteIO;
   else
      SystemData.PageWriteIO = 0;
   PrevPageWriteIO = PageWriteIO;

   SystemData.PageFaultsHard = SystemData.PageReadIO + SystemData.PageWriteIO;
   tfloat = (float)SystemData.PageFaultsHard / DeltaSeconds;
   SystemData.PageFaultsHard = (unsigned int)tfloat;

   SystemData.PageFaultsSoft = SystemData.PageFaults - SystemData.PageFaultsHard;
   tfloat = (float)SystemData.PageFaultsSoft / DeltaSeconds;
   SystemData.PageFaultsSoft = (unsigned int)tfloat;

   /********/
   /* MSCP */
   /********/

   if (SystemData.ClusterMember)
   {
      if (ItemListPtr == ItemList1)
      {
         MscpRead = MscpData[23];
         MscpWrite = MscpData[24];
      }

      if (PrevMscpRead && MscpRead > PrevMscpRead)
         SystemData.MscpRead = MscpRead - PrevMscpRead;
      else
         SystemData.MscpRead = 0;
      PrevMscpRead = MscpRead;
      tfloat = (float)SystemData.MscpRead / DeltaSeconds;
      SystemData.MscpRead = (unsigned int)tfloat;

      if (PrevMscpWrite > MscpWrite > PrevMscpWrite)
         SystemData.MscpWrite = MscpWrite - PrevMscpWrite;
      else
         SystemData.MscpWrite = 0;
      PrevMscpWrite = MscpWrite;
      tfloat = (float)SystemData.MscpWrite / DeltaSeconds;
      SystemData.MscpWrite = (unsigned int)tfloat;
   }

   /********/
   /* lock */
   /********/

   if (PrevLockActivity &&
       LockEnqueue + LockDequeue + LockWait > PrevLockActivity)
      SystemData.LockActivity = LockEnqueue + LockDequeue + LockWait -
                             PrevLockActivity;
   else
      SystemData.LockActivity = 0;
   PrevLockActivity = LockEnqueue + LockDequeue + LockWait;
   tfloat = (float)SystemData.LockActivity / DeltaSeconds;
   SystemData.LockActivity = (unsigned int)tfloat;
   tfloat = (float)SystemData.LockWait / DeltaSeconds;
   SystemData.LockWait = (unsigned int)tfloat;

   if (PrevLockWait && LockWait > PrevLockWait)
      SystemData.LockWait = LockWait - PrevLockWait;
   else
      SystemData.LockWait = 0;
   PrevLockWait = LockWait;

   /********/
   /* disk */
   /********/

   if (PrevFcpRead && FcpRead > PrevFcpRead)
      SystemData.FcpRead = FcpRead - PrevFcpRead;
   else
      SystemData.FcpRead = 0;
   PrevFcpRead = FcpRead;
   tfloat = (float)SystemData.FcpRead / DeltaSeconds;
   SystemData.FcpRead = (unsigned int)tfloat;

   if (PrevFcpWrite && FcpWrite > PrevFcpWrite)
      SystemData.FcpWrite = FcpWrite - PrevFcpWrite;
   else
      SystemData.FcpWrite = 0;
   PrevFcpWrite = FcpWrite;
   tfloat = (float)SystemData.FcpWrite / DeltaSeconds;
   SystemData.FcpWrite = (unsigned int)tfloat;

   /*********/
   /* other */
   /*********/

   if (PrevDirectIO && DirectIO > PrevDirectIO)
      SystemData.DirectIO = DirectIO - PrevDirectIO;
   else
      SystemData.DirectIO = 0;
   PrevDirectIO = DirectIO;
   tfloat = (float)SystemData.DirectIO / DeltaSeconds;
   SystemData.DirectIO = (unsigned int)tfloat;

   if (PrevBufferedIO && BufferedIO > PrevBufferedIO)
      SystemData.BufferedIO = BufferedIO - PrevBufferedIO;
   else
      SystemData.BufferedIO = 0;
   PrevBufferedIO = BufferedIO;
   tfloat = (float)SystemData.BufferedIO / DeltaSeconds;
   SystemData.BufferedIO = (unsigned int)tfloat;

   SystemData.MemInUse = SystemData.MemSize - (FreeList >> 7) - (ModList >> 7);

   if (SystemData.MemInUse > SystemData.MemInUseMax)
      SystemData.MemInUseMax = SystemData.MemInUse;

   SystemData.ProcessCount = ProcessCount;
   SystemData.ProcessBatCount = ProcessBatCount;
   SystemData.ProcessIntCount = ProcessIntCount;
   SystemData.ProcessNetCount = ProcessNetCount;

   if (SystemData.ClusterMember)
   {
      /*******/
      /* SCS */
      /*******/

      if (PrevClusterMsgSent && ClusterMsgSent > PrevClusterMsgSent)
         SystemData.ClusterMsgSent = ClusterMsgSent - PrevClusterMsgSent;
      else
         SystemData.ClusterMsgSent = 0;
      PrevClusterMsgSent = ClusterMsgSent;
      tfloat = (float)SystemData.ClusterMsgSent / DeltaSeconds;
      SystemData.ClusterMsgSent = (unsigned int)tfloat;

      if (PrevClusterMsgRcvd && ClusterMsgRcvd > PrevClusterMsgRcvd)
         SystemData.ClusterMsgRcvd = ClusterMsgRcvd - PrevClusterMsgRcvd;
      else
         SystemData.ClusterMsgRcvd = 0;
      PrevClusterMsgRcvd = ClusterMsgRcvd;
      tfloat = (float)SystemData.ClusterMsgRcvd / DeltaSeconds;
      SystemData.ClusterMsgRcvd = (unsigned int)tfloat;

      if (PrevClusterKbyteMap && ClusterKbyteMap > PrevClusterKbyteMap)
         SystemData.ClusterKbyteMap = ClusterKbyteMap - PrevClusterKbyteMap;
      else
         SystemData.ClusterKbyteMap = 0;
      PrevClusterKbyteMap = ClusterKbyteMap;
      tfloat = (float)SystemData.ClusterKbyteMap / DeltaSeconds;
      SystemData.ClusterKbyteMap = (unsigned int)tfloat;

      if (PrevClusterKbyteSent && ClusterKbyteSent > PrevClusterKbyteSent)
         SystemData.ClusterKbyteSent = ClusterKbyteSent - PrevClusterKbyteSent;
      else
         SystemData.ClusterKbyteSent = 0;
      PrevClusterKbyteSent = ClusterKbyteSent;
      tfloat = (float)SystemData.ClusterKbyteSent / DeltaSeconds;
      SystemData.ClusterKbyteSent = (unsigned int)tfloat;

      if (PrevClusterKbyteRcvd && ClusterKbyteRcvd > PrevClusterKbyteRcvd)
         SystemData.ClusterKbyteRcvd = ClusterKbyteRcvd - PrevClusterKbyteRcvd;
      else
         SystemData.ClusterKbyteRcvd = 0;
      PrevClusterKbyteRcvd = ClusterKbyteRcvd;
      tfloat = (float)SystemData.ClusterKbyteRcvd / DeltaSeconds;
      SystemData.ClusterKbyteRcvd = (unsigned int)tfloat;
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
This function uses undocumented sense-mode functionality to load counters from
the network interface (commonly ethernet device).  Multiple devices are
supported using a multi-valued logical name MONDESI_NI and the frame-rate
and byte-rate and soft/hard error-rate statistics are accumlated to represent
total system data.

On Alpha and Itanium uses native 64 bit (quadword) accumulators.

Essential concepts have come from "EMON: moniteur ethernet (V2.1-2) Gerard
Guillaume" and found in DECUSLIB freeware, and from some peeks at the T4 source
code.  The implementation is mine.
*/

int CollectNetInt ()

{
   static int  NetIntCount;
   static unsigned short  NetIntChan [NI_DEVICE_MAX];
   static unsigned long  DcSCOM = 32,  /* DC$_SCOM */
                         DtNI = 13;    /* DT$_NI */

   static unsigned __int64  PrevBlocksRx,
                            PrevBlocksTx,
                            PrevBytesRx,
                            PrevBytesTx,
                            PrevErrorsHard,
                            PrevErrorsSoft;
   static long  DeviceCount = -1,
                DviUnit,
                LnmIndex = -1,
                LnmAttributes;
   static unsigned short  LogValueLen;
   static char  LogValue [32];
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LnmIndex), LNM$_INDEX, &LnmIndex, 0 },
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { sizeof(LogValue)-1, LNM$_STRING, &LogValue, &LogValueLen },
      { 0,0,0,0 }
   },
   DevScanItems [] =
   {
      { sizeof(DcSCOM), DVS$_DEVCLASS, &DcSCOM, 0 },
      { 0,0,0,0 }
   },
   GetDviItems [] =
   {
      { sizeof(DviUnit), DVI$_UNIT, &DviUnit, 0 },
      { 0,0,0,0 }
   };
   static $DESCRIPTOR (NetIntDsc, "");
   static $DESCRIPTOR (LogNameDsc, "MONDESI_NI");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   int  ecnt, status;
   unsigned short  pword,
                   retlen;
   unsigned long  LnmCount,
                  NetIntErrorsHard,
                  NetIntErrorsSoft;
   unsigned long  DevScanContext [2];
   float  tfloat;
   char  *vptr;
   char  DevScanBuffer [64],
         CountBuffer [512];
   struct STRUCT_IOSBLK  IOsb;
   $DESCRIPTOR (CountDsc, CountBuffer);
   $DESCRIPTOR (DevScanDsc, DevScanBuffer);

#define CASTIT (unsigned __int64)
   unsigned __int64  value,
                     NetIntBlocksRx,
                     NetIntBlocksTx,
                     NetIntBytesRx,
                     NetIntBytesTx;

   /*********/
   /* begin */
   /*********/

   if (LnmIndex < 0)
   {
      /* first call, check for logical name defined network devices */
      for (LnmIndex = 0; LnmIndex < NI_DEVICE_MAX; LnmIndex++)
      {
         status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, 0, &LnmItems);
         if (VMSnok (status) || !(LnmAttributes & LNM$M_EXISTS)) break;
         LogValue[LogValueLen] = '\0';

         NetIntDsc.dsc$a_pointer = LogValue;
         NetIntDsc.dsc$w_length = LogValueLen;

         status = sys$assign (&NetIntDsc, &NetIntChan[LnmIndex], 0, 0);
         if (VMSnok (status))
         {
            sprintf (NotePad, "$[%d] %s %%X%08.08X",
                     __LINE__, LogValue, status);
            NetIntCount = 0;
            break;
         }
         if (NetIntDev[0]) strcat (NetIntDev, ",");
         strcat (NetIntDev, LogValue);
         NetIntCount++;
      }
   }

   if (DeviceCount < 0)
   {
      /* first call */
      DeviceCount = 0;

      /* if logical name defined and found devices */
      if (NetIntCount) return (NetIntCount);

      /* scan system for network devices */
      DevScanContext[0] = DevScanContext[1] = 0;
      while (NetIntCount < NI_DEVICE_MAX)
      {
         status = sys$device_scan (&DevScanDsc, &retlen, 0,
                                   &DevScanItems, &DevScanContext);
         if (VMSnok (status))
         {
            if (status == SS$_NOMOREDEV || status == SS$_NOSUCHDEV) break;
            sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
            NetIntCount = 0;
            break;
         }

         status = sys$getdviw (0, 0, &DevScanDsc, &GetDviItems, &IOsb, 
                               0, 0, 0, 0);
         if (VMSok (status)) status = IOsb.iosb$w_status;
         if (VMSnok (status))
         {
            sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
            NetIntCount = 0;
            break;
         }
         /* if not a template device then continue */
         if (DviUnit != 0) continue;

         status = sys$assign (&DevScanDsc, &NetIntChan[DeviceCount++], 0, 0);
         if (VMSnok (status))
         {
            if (status == SS$_NOPRIV) continue;
            sprintf (NotePad, "$[%d] %s %%X%08.08X",
                     __LINE__, LogValue, status);
            NetIntCount = 0;
            break;
         }

         if (NetIntDev[0]) strcat (NetIntDev, ",");
         DevScanBuffer[retlen] = '\0';
         strcat (NetIntDev, DevScanBuffer+1);
         NetIntCount++;
      }

      /* first call, just return having assigned channels (perhaps) */
      return (NetIntCount);
   }

   /****************/
   /* collect data */
   /****************/

   /* initialise structure and local storage */
   SystemData.NetIntBlocksRx = SystemData.NetIntBlocksTx =
      SystemData.NetIntBytesRx = SystemData.NetIntBytesTx =
      NetIntBlocksRx = NetIntBlocksTx =
      NetIntBytesRx = NetIntBytesTx =
      NetIntErrorsHard = NetIntErrorsSoft = 0;

   /* if no devices then just return from here */
   if (!NetIntCount) return (NetIntCount);

   /* accumulate counts from each device */
   for (ecnt = 0; ecnt < NetIntCount; ecnt++)
   {
      memset (CountBuffer, 0, sizeof(CountBuffer));

      status = sys$qiow (0, NetIntChan[ecnt],
#define IO$M_RD_64COUNT 0x4000
                         IO$_SENSEMODE | IO$M_RD_COUNT |
                         IO$M_RD_64COUNT | IO$M_CTRL,
                         &IOsb, 0, 0,
                         0, &CountDsc, 0, 0, 0, 0);

      if (VMSok (status)) status = IOsb.iosb$w_status;
      if (VMSnok (status))
      {
         sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
         return (NetIntCount);
      }

      vptr = CountBuffer;
      while (vptr < CountBuffer + IOsb.iosb$w_bcnt)
      {
         pword = *(unsigned short*)vptr;
         vptr += 2;

         /* end of list */
         if (!pword) break;

         /* MSB should be set! */
         if (!(pword & 0x8000))
         {
            sprintf (NotePad, "$[%d] %04.04X", __LINE__, pword);
            break;
         }

         value = 0;
         if ((pword & 0x0fff) < 200)
         {
            /* quadword value (item number less than 200) */
            value = *(unsigned __int64*)vptr;
            vptr += 8;
         }
         else
         if ((pword & 0x6000) == 0x6000)
         {
            /* longword value */
            value = *(unsigned long*)vptr;
            vptr += 4;
         } 
         else
         if ((pword & 0x6000) == 0x4000)
         {
            /* word value */
            value = *(unsigned short*)vptr;
            vptr += 2;
         } 
         else
         if ((pword & 0x6000) == 0x2000)
         {
            /* byte value */
            value = *(unsigned char*)vptr;
            vptr++;
         } 

         if (pword & 0x1000)
         {
            /* hw map (?) */
            value = *(unsigned long*)vptr;
            vptr += 2;
            continue;
         } 

         switch (pword & 0x0fff)
         {
            /* these look suspiciously like "LANCP SHOW DEVICE /COUNT"! */

            case 1 :
                 /* seconds since last zeroed */
                 break;

            case 2 :
            case 6 :
                 NetIntBytesRx += value;
                 break;

            case 3 :
            case 7 :
                 NetIntBytesTx += value;
                 break;

            case 4 :
            case 8 :
                 NetIntBlocksRx += value;
                 break;

            case 5 :
            case 9 :
                 NetIntBlocksTx += value;
                 break;

            case 12 :
            case 13 :
            case 14 :
            case 15 :
            case 16 :
            case 17 :
            case 18 :
            case 19 :
            case 30 :
            case 21 :
            case 22 :
            case 23 :
            case 24 :
            case 27 :
            case 28 :
            case 29 :
                 NetIntErrorsSoft += value;
                 break;

            case 25 :
            case 26 :
            case 33 :
            case 34 :
                 NetIntErrorsHard += value;
                 break;

            /* more "traditional" counters */

            case 1000 /* NMA$C_CTLIN_BRC */ :
            case 1002 /* NMA$C_CTLIN_MBY */ :
                 NetIntBytesRx += value;
                 break;

            case 1001 /* NMA$C_CTLIN_BSN */ :
                 NetIntBytesTx += value;
                 break;

            case 1010 /* NMA$C_CTLIN_DBR */ :
            case 1012 /* NMA$C_CTLIN_MBL */ :
                 NetIntBlocksRx += value;
                 break;

            case 1011 /* NMA$C_CTLIN_DBS */ :
                 NetIntBlocksTx += value;
                 break;

            case 1013 /* NMA$C_CTLIN_BID */ :
            case 1014 /* NMA$C_CTLIN_BS1 */ :
            case 1015 /* NMA$C_CTLIN_BSM */ :
            case 1040 /* NMA$C_CTLIN_RBE */ :
            case 1041 /* NMA$C_CTLIN_LBE */ :
            case 1064 /* NMA$C_CTLIN_OVR */ :
            case 1066 /* NMA$C_CTLIN_UBU */ :

                 NetIntErrorsSoft += value;
                 break;

            case 1060 /* NMA$C_CTLIN_SFL */ :
            case 1061 /* NMA$C_CTLIN_CDC */ :
            case 1063 /* NMA$C_CTLIN_UFD */ :
            case 1062 /* NMA$C_CTLIN_RFL */ :
            case 1065 /* NMA$C_CTLIN_SBU */ :

                 NetIntErrorsHard += value;
                 break;

            default :
         }
      }
   }

   if (PrevBytesRx && NetIntBytesRx > PrevBytesRx)
      SystemData.NetIntBytesRx = NetIntBytesRx - PrevBytesRx;
   PrevBytesRx = NetIntBytesRx;

   if (PrevBytesTx && NetIntBytesTx > PrevBytesTx)
      SystemData.NetIntBytesTx = NetIntBytesTx - PrevBytesTx;
   PrevBytesTx = NetIntBytesTx;

   tfloat = (float)SystemData.NetIntBytesRx / DeltaSeconds;
   SystemData.NetIntBytesRx = CASTIT tfloat;

   tfloat = (float)SystemData.NetIntBytesTx / DeltaSeconds;
   SystemData.NetIntBytesTx = CASTIT tfloat;

   if (PrevBlocksRx && NetIntBlocksRx > PrevBlocksRx)
      SystemData.NetIntBlocksRx = NetIntBlocksRx - PrevBlocksRx;
   PrevBlocksRx = NetIntBlocksRx;

   if (PrevBlocksTx && NetIntBlocksTx > PrevBlocksTx)
      SystemData.NetIntBlocksTx = NetIntBlocksTx - PrevBlocksTx;
   PrevBlocksTx = NetIntBlocksTx;

   tfloat = (float)SystemData.NetIntBlocksRx / DeltaSeconds;
   SystemData.NetIntBlocksRx = CASTIT tfloat;

   tfloat = (float)SystemData.NetIntBlocksTx / DeltaSeconds;
   SystemData.NetIntBlocksTx = CASTIT tfloat;

   if (PrevErrorsSoft && NetIntErrorsSoft > PrevErrorsSoft)
      SystemData.NetIntErrorsSoft += NetIntErrorsSoft - PrevErrorsSoft;
   PrevErrorsSoft = NetIntErrorsSoft;

   if (PrevErrorsHard && NetIntErrorsHard > PrevErrorsHard)
      SystemData.NetIntErrorsHard += NetIntErrorsHard - PrevErrorsHard;
   PrevErrorsHard = NetIntErrorsHard;

   return (NetIntCount);
}

/*****************************************************************************/
/*
Collect disk System Performance Information.
Uses the system service EXE$GETSPI() as the equivalent SYS$GETRMI() doesn't
seem to implement the functionality.
*/ 

int CollectDiskSPI ()

{
#define GETSPI$_DISKS 4203

typedef struct _getspi$disk_offsets {
    short int getspi$w_disk_alloclass;
    int getspi$l_disk_devname;
    short int getspi$w_disk_unitnum;
    char getspi$b_disk_flags;
    int getspi$q_disk_nodename [2];
    int getspi$q_disk_volnamel [2];
    int getspi$l_disk_volnameh;
    int getspi$l_disk_optcnt;
    int getspi$l_disk_qcount;
    } GETSPI$DISK_OFFSETS;

#define DISK_DATA_SIZE sizeof(unsigned long) + \
                       (sizeof(GETSPI$DISK_OFFSETS)*DISK_OFFSETS_MAX)

   static char  DiskData [DISK_DATA_SIZE];

   /* initialize structure for SPI items */
   static struct
   {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList1[] =
   {
       { sizeof(DiskData), GETSPI$_DISKS, &DiskData, 0 },
       {0,0,0,0}
   };

   static int  DiskCount = 0;
   static long  LnmIndex = -1,
                LnmAttributes;
   static unsigned short  LogValueLen;
   static $DESCRIPTOR (LogNameDsc, "MONDESI_DISK");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LnmIndex), LNM$_INDEX, &LnmIndex, 0 },
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { DISK_NAME_SIZE, LNM$_STRING, 0, &LogValueLen },
      { 0,0,0,0 }
   };

   int  dcnt, ncnt, status;
   float  tfloat;
   char  *cptr, *sptr;
   char  DiskName [128];
   struct STRUCT_IOSBLK  IOsb;
   GETSPI$DISK_OFFSETS  *doptr;
   SYSTEM_DATA  DummyData;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectDiskSPI()\n");

   if (LnmIndex < 0)
   {
      for (LnmIndex = 0; LnmIndex < DISK_DEVICE_MAX; LnmIndex++)
      {
         LnmItems[2].BufferPtr = SystemData.DiskName[LnmIndex];
         status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, 0, &LnmItems);
         if (VMSnok (status) || !(LnmAttributes & LNM$M_EXISTS)) break;
         SystemData.DiskName[LnmIndex][LogValueLen] = '\0';
         SystemData.DiskNameLen[LnmIndex] = LogValueLen;
      }
      if ((SystemData.DiskCount = LnmIndex) == DISK_DEVICE_MAX)
         SystemData.DiskCount--;

      return (SystemData.DiskCount);
   }

   /* if no disks being monitored */
   if (!SystemData.DiskCount) return (0);

   /* collect System Performance Information */
   status = exe$getspi (SpiEfn, 0, 0, &ItemList1, &IOsb, 0, 0);

   if (Debug)
      fprintf (stdout, "exe$getspi() %%X%08.08X  IOsb: %%X%08.08X disk: %d\n",
               status, IOsb.iosb$w_status, *(unsigned long*)&DiskData);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status)) return (0);

   DiskCount = *(unsigned long*)&DiskData;
   if (DiskCount == DISK_OFFSETS_MAX) EXIT_FI_LI (SS$_BUGCHECK);

   doptr = (GETSPI$DISK_OFFSETS*)(DiskData + sizeof(unsigned long));

   /* if the volume name is not updated the data is not fresh */
   SystemData.MountCount = 0;
   for (ncnt = 0; ncnt < SystemData.DiskCount; ncnt++)
      SystemData.DiskVolName[ncnt][0] = '\0';

   /* transfer from SPI into system data structure */
   for (dcnt = 0; dcnt < DiskCount; dcnt++, doptr++)
   {
      if (doptr->getspi$w_disk_alloclass)
         sprintf (DiskName, "$%d$%*.*s%d:",
                  doptr->getspi$w_disk_alloclass,
                  (int)(((char*)(&doptr->getspi$l_disk_devname))[0]),
                  (int)(((char*)(&doptr->getspi$l_disk_devname))[0]),
                  (char*)&doptr->getspi$l_disk_devname+1,
                  doptr->getspi$w_disk_unitnum);
      else
      if (((char*)(&doptr->getspi$q_disk_nodename))[0])
         sprintf (DiskName, "%*.*s$%*.*s%d:",
                  (int)(((char*)(&doptr->getspi$q_disk_nodename))[0]),
                  (int)(((char*)(&doptr->getspi$q_disk_nodename))[0]),
                  (char*)doptr->getspi$q_disk_nodename+1,
                  (int)(((char*)(&doptr->getspi$l_disk_devname))[0]),
                  (int)(((char*)(&doptr->getspi$l_disk_devname))[0]),
                  (char*)&doptr->getspi$l_disk_devname+1,
                  doptr->getspi$w_disk_unitnum);
      else
         sprintf (DiskName, "%*.*s%d:",
                  (int)(((char*)(&doptr->getspi$l_disk_devname))[0]),
                  (int)(((char*)(&doptr->getspi$l_disk_devname))[0]),
                  (char*)&doptr->getspi$l_disk_devname+1,
                  doptr->getspi$w_disk_unitnum);

      /* locate the particular device in the system data */
      for (ncnt = 0; ncnt < SystemData.DiskCount; ncnt++)
      {
         /* a fractionally more efficient test */
         if (*(__int64*)SystemData.DiskName[ncnt] != *(__int64*)DiskName) continue;
         if (!strcmp (SystemData.DiskName[ncnt], DiskName)) break;
      }

      /* if it's not a required disk device */
      if (ncnt >= SystemData.DiskCount) continue;

      /* volume name terminating on first space */
      sptr = SystemData.DiskVolName[ncnt];
      for (cptr = (char*)doptr->getspi$q_disk_volnamel;
           cptr < (char*)(doptr->getspi$q_disk_volnamel)+12 && *cptr != ' ';
           *sptr++ = *cptr++);
      *sptr = '\0';
      if (SystemData.DiskVolName[ncnt][0]) SystemData.MountCount++;

      SystemData.DiskOptCnt[ncnt] = doptr->getspi$l_disk_optcnt;
      if (SystemData.DiskPrevOptCnt[ncnt] &&
          SystemData.DiskOptCnt[ncnt] > SystemData.DiskPrevOptCnt[ncnt])
      {
         tfloat = (float)(SystemData.DiskOptCnt[ncnt] -
                          SystemData.DiskPrevOptCnt[ncnt]) / DeltaSeconds;
         SystemData.DiskOperCount[ncnt] = (unsigned int)tfloat;
      }
      else
         SystemData.DiskOperCount[ncnt] = 0;
      SystemData.DiskPrevOptCnt[ncnt] = SystemData.DiskOptCnt[ncnt];
   }

   CollectDiskDVI ();

   return (SystemData.DiskCount);
}

/*****************************************************************************/
/*
Collect disk device total and free block (usage).
*/ 

void CollectDiskDVI ()

{
   static unsigned long  DviErrCnt,
                         DviFreeBlocks,
                         DviMaxBlock;
   static $DESCRIPTOR (DevNameDsc, "");

   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   GetDviItems [] =
   {
      { sizeof(DviErrCnt), DVI$_ERRCNT, &DviErrCnt, 0 },
      { sizeof(DviFreeBlocks), DVI$_FREEBLOCKS, &DviFreeBlocks, 0 },
      { sizeof(DviMaxBlock), DVI$_MAXBLOCK, &DviMaxBlock, 0 },
      { 0,0,0,0 }
   };

   int  dcnt, len, status;
   struct STRUCT_IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectDiskDVI()\n");

   for (dcnt = 0; dcnt < SystemData.DiskCount; dcnt++)
   {
      if (!SystemData.DiskVolName[dcnt][0])
      {
         if (SystemData.DiskUsage[dcnt][0]) SystemData.DiskDelta++;
         SystemData.DiskFreeBlocks[dcnt] = 0;
         SystemData.DiskAlert[dcnt][0] = SystemData.DiskUsage[dcnt][0] = '\0';
         continue;
      }

      DevNameDsc.dsc$a_pointer = SystemData.DiskName[dcnt];
      DevNameDsc.dsc$w_length = SystemData.DiskNameLen[dcnt];

      status = sys$getdviw (0, 0, &DevNameDsc, &GetDviItems, &IOsb, 
                            0, 0, 0, 0);
      if (VMSok (status)) status = IOsb.iosb$w_status;
      if (VMSnok (status))
      {
         SystemData.DiskFreeBlocks[dcnt] = 0;
         sprintf (SystemData.DiskUsage[dcnt], "%%X%08.08X", status);
         continue;
      }

      if (!SystemData.DiskUsage[dcnt][0] ||
          SystemData.DiskFreeBlocks[dcnt] != DviFreeBlocks)
      {
         SystemData.DiskDelta++;
         SystemData.DiskFreeBlocks[dcnt] = DviFreeBlocks;
         if (DviFreeBlocks > 2097152) /* 1GB */
            len = sprintf (SystemData.DiskUsage[dcnt], "%dGB %d%%",
                           (int)((float)DviMaxBlock / 2097152.0),
                           (int)((((float)DviMaxBlock) -
                                   (float)DviFreeBlocks) * 100.0 /
                                   (float)DviMaxBlock));
         else
            len = sprintf (SystemData.DiskUsage[dcnt], "%dMB %d%%",
                           (int)((float)DviMaxBlock / 2048.0),
                           (int)((((float)DviMaxBlock) -
                                   (float)DviFreeBlocks) * 100.0 /
                                   (float)DviMaxBlock));
      }

#ifdef TEST_ERRCNT
      {
         static int  ErrCnt;
         DviErrCnt = ErrCnt++ / 30;
      }
#endif

      if (SystemData.DiskErrCnt[dcnt] != DviErrCnt) SystemData.DiskDelta++;

      if (SystemData.DiskErrCnt[dcnt] = DviErrCnt)
         sprintf (SystemData.DiskAlert[dcnt], "ERR: %d", DviErrCnt); 
      else
         SystemData.DiskAlert[dcnt][0] = '\0';
   }
}

/*****************************************************************************/
/*
Collect top-process System Performance Information for younger VMSes.
Uses the system service EXE$GETSPI() as the equivalent SYS$GETRMI() doesn't
seem to implement the functionality.

This function collects process SPI and sorts the top five consumers each of CPU
time, super+exec+kernel time, pages faults, buffered and direct I/O.  These
five are then compiled into an HTML table for display purposes.

This function uses a data structure derived in part from VMS sources and
contains items and alignment (seemingly) not documented elsewhere.  Not being
completely obvious when the new structure started to be used (VMS V8.3?) this
function switches between this and the older (documented and header-file
described) data structure by sanity-checking the data returned.  If any of the
item values appear out-of-range then it's likely the offsets are wrong for this
(older) system.  Switch to using an equivalent function that employs that older
structure.
*/ 

void CollectProcSPI1 ()

{
#define GETSPI$_PROC 4123

/* LOTS of trial-and-error in this structure! */

typedef struct _getspi$proc_class {
    int getspi$l_proc_ipid;
    int getspi$l_proc_uic;
    short int getspi$w_proc_state;
    char getspi$b_proc_pri;
    /* OCTAWORD basealign octaword? */
    char filler_to_octaword_alignment [5];
    char getspi$o_proc_lname [16];
    int getspi$l_proc_gpgcnt;
    int getspi$l_proc_ppgcnt;
    int getspi$l_proc_sts;
    int getspi$l_proc_diocnt;
    int getspi$l_proc_pageflts;
    int getspi$l_proc_cputim;
    int getspi$l_proc_biocnt;
    int getspi$l_proc_epid;
    int getspi$l_proc_efwm;
    int getspi$l_proc_rbstran;
    /* from V8.3 IA64 SPIDEF.SDL */
    int getspi$l_proc_kernel_counter;
    int getspi$l_proc_executive_counter;
    int getspi$l_proc_supervisor_counter;
    int getspi$l_proc_user_counter;
    int getspi$l_proc_reserved1;
    /* use the second reserve to accumulate non-user ticks */
    /** int getspi$l_proc_reserved2; **/
    int getspi$l_proc_SEK;
    } GETSPI$PROC_CLASS;

   static long  BufferNumber,
                StateCefCount,
                StateColpgCount,
                StateComCount,
                StateComoCount,
                StateCurCount,
                StateFpgCount,
                StateHibCount,
                StateHiboCount,
                StateLefCount,
                StateLefoCount,
                StateMwaitCount,
                StatePfwCount,
                StateSuspCount,
                StateSuspoCount;
   static int  ProcDataSize,
               ProcMax;
   static unsigned short  PrevMinute = -1;
   static char  *ProcData1,
                *ProcData2;

   /* initialize structure for SPI items */
   static struct
   {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList1[] =
   {
       { 0, GETSPI$_PROC, 0, 0 },
       { sizeof(StateCefCount),   RMI$_CEF,   &StateCefCount, 0 },
       { sizeof(StateColpgCount), RMI$_COLPG, &StateColpgCount, 0 },
       { sizeof(StateComCount),   RMI$_COM,   &StateComCount, 0 },
       { sizeof(StateComoCount),  RMI$_COMO,  &StateComoCount, 0 },
       { sizeof(StateCurCount),   RMI$_CUR,   &StateCurCount, 0 },
       { sizeof(StateFpgCount),   RMI$_FPG,   &StateFpgCount, 0 },
       { sizeof(StateHibCount),   RMI$_HIB,   &StateHibCount, 0 },
       { sizeof(StateHiboCount),  RMI$_HIBO,  &StateHiboCount, 0 },
       { sizeof(StateLefCount),   RMI$_LEF,   &StateLefCount, 0 },
       { sizeof(StateLefoCount),  RMI$_LEFO,  &StateLefoCount, 0 },
       { sizeof(StateMwaitCount), RMI$_MWAIT, &StateMwaitCount, 0 },
       { sizeof(StatePfwCount),   RMI$_PFW,   &StatePfwCount, 0 },
       { sizeof(StateSuspCount),  RMI$_SUSP,  &StateSuspCount, 0 },
       { sizeof(StateSuspoCount), RMI$_SUSPO, &StateSuspoCount, 0 },
       {0,0,0,0}
   };

   int  cnt, spicnt, sprcnt, status, topcnt,
        PrevProcCount,
        ProcCount,
        RowCount;
   float  tfloat;
   char  *sptr, *zptr;
   struct STRUCT_IOSBLK  IOsb;
   GETSPI$PROC_CLASS  *prvptr, *spiptr, *sprptr;
   GETSPI$PROC_CLASS  *TopCpuTim [PROC_MAX],
                      *TopBioCnt [PROC_MAX],
                      *TopDioCnt [PROC_MAX],
                      *TopPageFlts [PROC_MAX],
                      *TopSekTim [PROC_MAX];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectProcSPI1()\n");

   /* if process not being monitored */
   if (!SystemData.ProcRMI) return;

   /***************/
   /* collect SPI */
   /***************/

   if (SystemData.ProcessCount >= ProcMax)
   {
      /* grow the buffer space within limits */
      while (SystemData.ProcessCount >= ProcMax &&
             ProcMax < PROC_CLASS_MAX) ProcMax += PROC_CLASS_INC;
      /* plus ten gives a little margin for process growth since count */
      ProcDataSize = sizeof(long) + (sizeof(GETSPI$PROC_CLASS) * (ProcMax+10));
      if (ProcData1) free (ProcData1);
      ProcData1 = calloc (1, ProcDataSize);
      if (!ProcData1) EXIT_FI_LI (vaxc$errno);
      if (ProcData2) free (ProcData2);
      ProcData2 = calloc (1, ProcDataSize);
      if (!ProcData2) EXIT_FI_LI (vaxc$errno);
      ItemList1[0].BufferLength = ProcDataSize;
      BufferNumber = 0;
   }

   /* swap back and forth between the two buffers */
   if (BufferNumber++ % 2)
   {
      spiptr = (GETSPI$PROC_CLASS*)ProcData2;
      prvptr = (GETSPI$PROC_CLASS*)ProcData1;
   }
   else
   {
      spiptr = (GETSPI$PROC_CLASS*)ProcData1;
      prvptr = (GETSPI$PROC_CLASS*)ProcData2;
   }
   ItemList1[0].BufferPtr = spiptr;

   /* collect System Performance Information */
   status = exe$getspi (SpiEfn, 0, 0, &ItemList1, &IOsb, 0, 0);

   if (Debug)
      fprintf (stdout, "exe$getspi() %%X%08.08X  IOsb: %%X%08.08X proc: %d\n",
               status, IOsb.iosb$w_status, *(unsigned long*)spiptr);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status))
   {
      SystemData.ProcBufferLength =
         sprintf (SystemData.ProcBuffer,
"the.mdsMonProc(\"<tr><td>EXE$GETSPI() %%X%08.08X</td></tr>\");",
                  status);
      return;
   }

   if (SystemData.ProcSanityCheck == 0)
   {
      /* sanity check the data structure */
      SystemData.ProcSanityCheck = 1;
      ProcCount = *(unsigned long*)spiptr;
      spiptr = (GETSPI$PROC_CLASS*)((char*)spiptr + sizeof(unsigned long));
      for (spicnt = 0; spicnt < ProcCount; spicnt++, spiptr++)
      {
          /* if any seem out-of-range the offsets are probably wrong */
          if (spiptr->getspi$l_proc_epid == 0 ||
              spiptr->getspi$w_proc_state < 0 ||
              spiptr->getspi$w_proc_state > 255 ||
              *spiptr->getspi$o_proc_lname < 0 ||
              *spiptr->getspi$o_proc_lname > 15)
          {
             /* try using the older version of the structure */
             free (ProcData1);
             free (ProcData2);
             SystemData.ProcSpiFunction = &CollectProcSPI2;
             CollectProcSPI2 ();
             return;
          }
       }
   }

   /**********************/
   /* get process counts */
   /**********************/

   /* number of processes read by SPI */
   ProcCount = *(unsigned long*)spiptr;
   if (ProcCount == PROC_CLASS_MAX) EXIT_FI_LI (SS$_BUGCHECK);
   /* now step over the process count */
   spiptr = (GETSPI$PROC_CLASS*)((char*)spiptr + sizeof(unsigned long));

   /* number of process read into the previous SPI */
   PrevProcCount = *(unsigned long*)prvptr;
   /* now step over the process count */
   prvptr = (GETSPI$PROC_CLASS*)((char*)prvptr + sizeof(unsigned long));
   /* if there is no previous data then just skip this initial call */
   if (!PrevProcCount) return;

   /**************************/
   /* loop through processes */
   /**************************/

   for (cnt = 0; cnt < PROC_MAX; cnt++)
       TopCpuTim[cnt] = TopBioCnt[cnt] = TopDioCnt[cnt] =
          TopPageFlts[cnt] = TopSekTim[cnt] = NULL;

   for (spicnt = 0; spicnt < ProcCount; spicnt++, spiptr++)
   {
      if (Debug)
         fprintf (stdout,
"SPI %2d. %8.8X |%*.*s| cpu:%u pgflt:%u dio:%u bio:%u\n",
                  spicnt+1, spiptr->getspi$l_proc_epid,
                  *spiptr->getspi$o_proc_lname,
                  *spiptr->getspi$o_proc_lname,
                  spiptr->getspi$o_proc_lname+1,
                  spiptr->getspi$l_proc_cputim,
                  spiptr->getspi$l_proc_pageflts,
                  spiptr->getspi$l_proc_diocnt,
                  spiptr->getspi$l_proc_biocnt);

      /*************************/
      /* find in previous data */
      /*************************/

      sprptr = prvptr;
      for (sprcnt = 0; sprcnt < PrevProcCount; sprcnt++, sprptr++)
         if (spiptr->getspi$l_proc_epid == sprptr->getspi$l_proc_epid) break;
      /* if process not found in previous data */
      if (sprcnt >= PrevProcCount) continue;

      if (Debug)
         fprintf (stdout,
"PRV %2d. %8.8X |%*.*s| cpu:%u pgflt:%u dio:%u bio:%u\n",
                  sprcnt+1, sprptr->getspi$l_proc_epid,
                  *spiptr->getspi$o_proc_lname,
                  *spiptr->getspi$o_proc_lname,
                  sprptr->getspi$o_proc_lname+1,
                  sprptr->getspi$l_proc_cputim,
                  sprptr->getspi$l_proc_pageflts,
                  sprptr->getspi$l_proc_diocnt,
                  sprptr->getspi$l_proc_biocnt);

      /********************/
      /* calculate deltas */
      /********************/

      sprptr->getspi$l_proc_cputim =
        (int)((float)(spiptr->getspi$l_proc_cputim -
                      sprptr->getspi$l_proc_cputim) / DeltaSeconds);
      sprptr->getspi$l_proc_biocnt =
         (int)((float)(spiptr->getspi$l_proc_biocnt -
                       sprptr->getspi$l_proc_biocnt) / DeltaSeconds);
      sprptr->getspi$l_proc_diocnt =
         (int)((float)(spiptr->getspi$l_proc_diocnt -
                       sprptr->getspi$l_proc_diocnt) / DeltaSeconds);
      sprptr->getspi$l_proc_pageflts =
         (int)((float)(spiptr->getspi$l_proc_pageflts -
                       sprptr->getspi$l_proc_pageflts) / DeltaSeconds);
      sprptr->getspi$l_proc_SEK = spiptr->getspi$l_proc_kernel_counter -
                                    sprptr->getspi$l_proc_kernel_counter;
      sprptr->getspi$l_proc_SEK += spiptr->getspi$l_proc_executive_counter -
                                     sprptr->getspi$l_proc_executive_counter;
      sprptr->getspi$l_proc_SEK += spiptr->getspi$l_proc_supervisor_counter -
                                     sprptr->getspi$l_proc_supervisor_counter;
      sprptr->getspi$l_proc_SEK =
         (int)((float)sprptr->getspi$l_proc_SEK / DeltaSeconds);

      /**************************/
      /* populate sorted values */
      /**************************/

      if (sprptr->getspi$l_proc_cputim)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopCpuTim[topcnt])
            {
               if (sprptr->getspi$l_proc_cputim >
                   TopCpuTim[topcnt]->getspi$l_proc_cputim)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopCpuTim[cnt] = TopCpuTim[cnt-1];
                  TopCpuTim[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopCpuTim[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_biocnt)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopBioCnt[topcnt])
            {
               if (sprptr->getspi$l_proc_biocnt >
                   TopBioCnt[topcnt]->getspi$l_proc_biocnt)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopBioCnt[cnt] = TopBioCnt[cnt-1];
                  TopBioCnt[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopBioCnt[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_diocnt)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopDioCnt[topcnt])
            {
               if (sprptr->getspi$l_proc_diocnt >
                   TopDioCnt[topcnt]->getspi$l_proc_diocnt)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopDioCnt[cnt] = TopDioCnt[cnt-1];
                  TopDioCnt[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopDioCnt[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_pageflts)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopPageFlts[topcnt])
            {
               if (sprptr->getspi$l_proc_pageflts >
                   TopPageFlts[topcnt]->getspi$l_proc_pageflts)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopPageFlts[cnt] = TopPageFlts[cnt-1];
                  TopPageFlts[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopPageFlts[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_SEK)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopSekTim[topcnt])
            {
               if (sprptr->getspi$l_proc_SEK >
                   TopSekTim[topcnt]->getspi$l_proc_SEK)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopSekTim[cnt] = TopSekTim[cnt-1];
                  TopSekTim[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopSekTim[topcnt] = sprptr;
               break;
            }
         }

      /********************/
      /* end process loop */
      /********************/
   }

   /********************/
   /* write html table */
   /********************/

   zptr = (sptr = SystemData.ProcBuffer) + sizeof(SystemData.ProcBuffer)-256;

   sptr += sprintf (sptr,
"the.mdsMonProc(\"<b>COM:</b> %d  <b>COMO:</b> %d  <b>CUR:</b> %d  \
<b>HIB:</b> %d  <b>HIBO:</b> %d  <b>LEF:</b> %d  <b>LEFO:</b> %d  \
<b>MWAIT:</b> %d  <b>PFW:</b> %d  <b>SUSP:</b> %d  <b>SUSPO:</b> %d  \
SWAPPED: %d  WAITING: %d\",\"",
                    StateComCount, StateComoCount, StateCurCount,
                    StateHibCount, StateHiboCount, StateLefCount, 
                    StateLefoCount, StateMwaitCount, 
                    StatePfwCount, StateSuspCount, StateSuspoCount,
                    StateComoCount + StateHiboCount +
                       StateLefoCount + StateSuspoCount,
                    StateCefCount + StateColpgCount +
                       StateFpgCount + StateMwaitCount + StatePfwCount);
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   for (cnt = RowCount = 0; cnt < PROC_MAX; cnt++)
   {
      if (!(TopCpuTim[cnt] || TopSekTim[cnt] ||
            TopPageFlts[cnt] || TopBioCnt[cnt] || TopDioCnt[cnt]))
         continue;
      RowCount++;

      sptr += sprintf (sptr, "<tr>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopCpuTim[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %1.2f</td>",
                          TopCpuTim[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopCpuTim[cnt]->getspi$o_proc_lname),
                          (float)TopCpuTim[cnt]->getspi$l_proc_cputim/100.0);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopSekTim[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %1.2f</td>",
                          TopSekTim[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopSekTim[cnt]->getspi$o_proc_lname),
                          (float)TopSekTim[cnt]->getspi$l_proc_SEK/100.0);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopPageFlts[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopPageFlts[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopPageFlts[cnt]->getspi$o_proc_lname),
                          TopPageFlts[cnt]->getspi$l_proc_pageflts);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopBioCnt[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopBioCnt[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopBioCnt[cnt]->getspi$o_proc_lname),
                          TopBioCnt[cnt]->getspi$l_proc_biocnt);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopDioCnt[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopDioCnt[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopDioCnt[cnt]->getspi$o_proc_lname),
                          TopDioCnt[cnt]->getspi$l_proc_diocnt);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sptr += sprintf (sptr, "</tr>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
   }

   sptr += sprintf (sptr, "\");");
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   if (RowCount)
      SystemData.ProcBufferLength = sptr - SystemData.ProcBuffer;
   else
      SystemData.ProcBuffer[SystemData.ProcBufferLength = 0] = '\0';

   if (Debug)
      fprintf (stdout, "\n%d\n%s\n",
               SystemData.ProcBufferLength, SystemData.ProcBuffer);
}

/*****************************************************************************/
/*
Collect top-process System Performance Information for older VMSes.
Functionality is as per CollectProcSPI1().
*/ 

void CollectProcSPI2 ()

{
#define GETSPI$_PROC 4123

/* structure according to the book! */

typedef struct _getspi$proc_class {
    int getspi$l_proc_ipid;
    int getspi$l_proc_uic;
    short int getspi$w_proc_state;
    char getspi$b_proc_pri;
    char getspi$o_proc_lname [16];
    int getspi$l_proc_gpgcnt;
    int getspi$l_proc_ppgcnt;
    int getspi$l_proc_sts;
    int getspi$l_proc_diocnt;
    int getspi$l_proc_pageflts;
    int getspi$l_proc_cputim;
    int getspi$l_proc_biocnt;
    int getspi$l_proc_epid;
    int getspi$l_proc_efwm;
    int getspi$l_proc_rbstran;
    } GETSPI$PROC_CLASS;

   static long  BufferNumber,
                StateCefCount,
                StateColpgCount,
                StateComCount,
                StateComoCount,
                StateCurCount,
                StateFpgCount,
                StateHibCount,
                StateHiboCount,
                StateLefCount,
                StateLefoCount,
                StateMwaitCount,
                StatePfwCount,
                StateSuspCount,
                StateSuspoCount;
   static int  ProcDataSize,
               ProcMax;
   static unsigned short  PrevMinute = -1;
   static char  *ProcData1,
                *ProcData2;

   /* initialize structure for SPI items */
   static struct
   {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList1[] =
   {
       { 0, GETSPI$_PROC, 0, 0 },
       { sizeof(StateCefCount),   RMI$_CEF,   &StateCefCount, 0 },
       { sizeof(StateColpgCount), RMI$_COLPG, &StateColpgCount, 0 },
       { sizeof(StateComCount),   RMI$_COM,   &StateComCount, 0 },
       { sizeof(StateComoCount),  RMI$_COMO,  &StateComoCount, 0 },
       { sizeof(StateCurCount),   RMI$_CUR,   &StateCurCount, 0 },
       { sizeof(StateFpgCount),   RMI$_FPG,   &StateFpgCount, 0 },
       { sizeof(StateHibCount),   RMI$_HIB,   &StateHibCount, 0 },
       { sizeof(StateHiboCount),  RMI$_HIBO,  &StateHiboCount, 0 },
       { sizeof(StateLefCount),   RMI$_LEF,   &StateLefCount, 0 },
       { sizeof(StateLefoCount),  RMI$_LEFO,  &StateLefoCount, 0 },
       { sizeof(StateMwaitCount), RMI$_MWAIT, &StateMwaitCount, 0 },
       { sizeof(StatePfwCount),   RMI$_PFW,   &StatePfwCount, 0 },
       { sizeof(StateSuspCount),  RMI$_SUSP,  &StateSuspCount, 0 },
       { sizeof(StateSuspoCount), RMI$_SUSPO, &StateSuspoCount, 0 },
       {0,0,0,0}
   };

   int  cnt, spicnt, sprcnt, status, topcnt,
        PrevProcCount,
        ProcCount,
        RowCount;
   float  tfloat;
   char  *sptr, *zptr;
   struct STRUCT_IOSBLK  IOsb;
   GETSPI$PROC_CLASS  *prvptr, *spiptr, *sprptr;
   GETSPI$PROC_CLASS  *TopCpuTim [PROC_MAX],
                      *TopBioCnt [PROC_MAX],
                      *TopDioCnt [PROC_MAX],
                      *TopPageFlts [PROC_MAX];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectProcSPI2()\n");

   SystemData.ProcBuffer[SystemData.ProcBufferLength = 0] = '\0';

   /* if process not being monitored */
   if (!SystemData.ProcRMI) return;

   /***************/
   /* collect SPI */
   /***************/

   if (SystemData.ProcessCount >= ProcMax)
   {
      /* grow the buffer space within limits */
      while (SystemData.ProcessCount >= ProcMax &&
             ProcMax < PROC_CLASS_MAX) ProcMax += PROC_CLASS_INC;
      /* plus ten gives a little margin for process growth since count */
      ProcDataSize = sizeof(long) + (sizeof(GETSPI$PROC_CLASS) * (ProcMax+10));
      if (ProcData1) free (ProcData1);
      ProcData1 = calloc (1, ProcDataSize);
      if (!ProcData1) EXIT_FI_LI (vaxc$errno);
      if (ProcData2) free (ProcData2);
      ProcData2 = calloc (1, ProcDataSize);
      if (!ProcData2) EXIT_FI_LI (vaxc$errno);
      ItemList1[0].BufferLength = ProcDataSize;
      BufferNumber = 0;
   }

   /* swap back and forth between the two buffers */
   if (BufferNumber++ % 2)
   {
      spiptr = (GETSPI$PROC_CLASS*)ProcData2;
      prvptr = (GETSPI$PROC_CLASS*)ProcData1;
   }
   else
   {
      spiptr = (GETSPI$PROC_CLASS*)ProcData1;
      prvptr = (GETSPI$PROC_CLASS*)ProcData2;
   }
   ItemList1[0].BufferPtr = spiptr;

   /* collect System Performance Information */
   status = exe$getspi (SpiEfn, 0, 0, &ItemList1, &IOsb, 0, 0);

   if (Debug)
      fprintf (stdout, "exe$getspi() %%X%08.08X  IOsb: %%X%08.08X proc: %d\n",
               status, IOsb.iosb$w_status, *(unsigned long*)spiptr);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status))
   {
      SystemData.ProcBufferLength =
         sprintf (SystemData.ProcBuffer,
"the.mdsMonProc(\"<tr><td>EXE$GETSPI() %%X%08.08X</td></tr>\");",
                  status);
      return;
   }

   if (SystemData.ProcSanityCheck == 1)
   {
      /* sanity check the data structure */
      SystemData.ProcSanityCheck = 2;
      ProcCount = *(unsigned long*)ItemList1[0].BufferPtr;
      spiptr = (GETSPI$PROC_CLASS*)((char*)spiptr + sizeof(unsigned long));
      for (spicnt = 0; spicnt < ProcCount; spicnt++, spiptr++)
      {
         /* if any seem out-of-range the offsets are probably wrong - hmmm! */
         if (spiptr->getspi$l_proc_epid == 0 ||
             spiptr->getspi$w_proc_state < 0 ||
             spiptr->getspi$w_proc_state > 255 ||
             *spiptr->getspi$o_proc_lname < 0 ||
             *spiptr->getspi$o_proc_lname > 15)
            EXIT_FI_LI (SS$_BUGCHECK);
      }
   }

   /**********************/
   /* get process counts */
   /**********************/

   /* number of processes read by SPI */
   ProcCount = *(unsigned long*)ItemList1[0].BufferPtr;
   if (ProcCount == PROC_CLASS_MAX) EXIT_FI_LI (SS$_BUGCHECK);
   /* now step over the process count */
   spiptr = (GETSPI$PROC_CLASS*)((char*)spiptr + sizeof(unsigned long));

   /* number of process read into the previous SPI */
   PrevProcCount = *(unsigned long*)prvptr;
   /* now step over the process count */
   prvptr = (GETSPI$PROC_CLASS*)((char*)prvptr + sizeof(unsigned long));
   /* if there is no previous data then just skip this initial call */
   if (!PrevProcCount) return;

   /**************************/
   /* loop through processes */
   /**************************/

   for (cnt = 0; cnt < PROC_MAX; cnt++)
       TopCpuTim[cnt] = TopBioCnt[cnt] = TopDioCnt[cnt] =
          TopPageFlts[cnt] = NULL;

   for (spicnt = 0; spicnt < ProcCount; spicnt++, spiptr++)
   {
      if (Debug)
         fprintf (stdout,
"SPI %2d. %8.8X |%*.*s| cpu:%u pgflt:%u dio:%u bio:%u\n",
                  spicnt+1, spiptr->getspi$l_proc_epid,
                  *spiptr->getspi$o_proc_lname,
                  *spiptr->getspi$o_proc_lname,
                  spiptr->getspi$o_proc_lname+1,
                  spiptr->getspi$l_proc_cputim,
                  spiptr->getspi$l_proc_pageflts,
                  spiptr->getspi$l_proc_diocnt,
                  spiptr->getspi$l_proc_biocnt);

      /*************************/
      /* find in previous data */
      /*************************/

      sprptr = prvptr;
      for (sprcnt = 0; sprcnt < PrevProcCount; sprcnt++, sprptr++)
         if (spiptr->getspi$l_proc_epid == sprptr->getspi$l_proc_epid) break;
      /* if process not found in previous data */
      if (sprcnt >= PrevProcCount) continue;

      if (Debug)
         fprintf (stdout,
"PRV %2d. %8.8X |%*.*s| cpu:%u pgflt:%u dio:%u bio:%u\n",
                  sprcnt+1, sprptr->getspi$l_proc_epid,
                  *spiptr->getspi$o_proc_lname,
                  *spiptr->getspi$o_proc_lname,
                  sprptr->getspi$o_proc_lname+1,
                  sprptr->getspi$l_proc_cputim,
                  sprptr->getspi$l_proc_pageflts,
                  sprptr->getspi$l_proc_diocnt,
                  sprptr->getspi$l_proc_biocnt);

      /********************/
      /* calculate deltas */
      /********************/

      sprptr->getspi$l_proc_cputim =
        (int)((float)(spiptr->getspi$l_proc_cputim -
                      sprptr->getspi$l_proc_cputim) / DeltaSeconds);
      sprptr->getspi$l_proc_biocnt =
         (int)((float)(spiptr->getspi$l_proc_biocnt -
                       sprptr->getspi$l_proc_biocnt) / DeltaSeconds);
      sprptr->getspi$l_proc_diocnt =
         (int)((float)(spiptr->getspi$l_proc_diocnt -
                       sprptr->getspi$l_proc_diocnt) / DeltaSeconds);
      sprptr->getspi$l_proc_pageflts =
         (int)((float)(spiptr->getspi$l_proc_pageflts -
                       sprptr->getspi$l_proc_pageflts) / DeltaSeconds);

      /**************************/
      /* populate sorted values */
      /**************************/

      if (sprptr->getspi$l_proc_cputim)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopCpuTim[topcnt])
            {
               if (sprptr->getspi$l_proc_cputim >
                   TopCpuTim[topcnt]->getspi$l_proc_cputim)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopCpuTim[cnt] = TopCpuTim[cnt-1];
                  TopCpuTim[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopCpuTim[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_biocnt)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopBioCnt[topcnt])
            {
               if (sprptr->getspi$l_proc_biocnt >
                   TopBioCnt[topcnt]->getspi$l_proc_biocnt)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopBioCnt[cnt] = TopBioCnt[cnt-1];
                  TopBioCnt[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopBioCnt[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_diocnt)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopDioCnt[topcnt])
            {
               if (sprptr->getspi$l_proc_diocnt >
                   TopDioCnt[topcnt]->getspi$l_proc_diocnt)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopDioCnt[cnt] = TopDioCnt[cnt-1];
                  TopDioCnt[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopDioCnt[topcnt] = sprptr;
               break;
            }
         }

      if (sprptr->getspi$l_proc_pageflts)
         for (topcnt = 0; topcnt < PROC_MAX; topcnt++)
         {
            if (TopPageFlts[topcnt])
            {
               if (sprptr->getspi$l_proc_pageflts >
                   TopPageFlts[topcnt]->getspi$l_proc_pageflts)
               {
                  for (cnt = PROC_MAX-1; cnt > topcnt; cnt--)
                     TopPageFlts[cnt] = TopPageFlts[cnt-1];
                  TopPageFlts[topcnt] = sprptr;
                  break;
               }
            }
            else
            {
               TopPageFlts[topcnt] = sprptr;
               break;
            }
         }

      /********************/
      /* end process loop */
      /********************/
   }

   /********************/
   /* write html table */
   /********************/

   zptr = (sptr = SystemData.ProcBuffer) + sizeof(SystemData.ProcBuffer)-256;

   sptr += sprintf (sptr,
"the.mdsMonProc(\"<b>COM:</b> %d  <b>COMO:</b> %d  <b>CUR:</b> %d  \
<b>HIB:</b> %d  <b>HIBO:</b> %d  <b>LEF:</b> %d  <b>LEFO:</b> %d  \
<b>MWAIT:</b> %d  <b>PFW:</b> %d  <b>SUSP:</b> %d  <b>SUSPO:</b> %d  \
SWAPPED: %d  WAITING: %d\",\"",
                    StateComCount, StateComoCount, StateCurCount,
                    StateHibCount, StateHiboCount, StateLefCount, 
                    StateLefoCount, StateMwaitCount, 
                    StatePfwCount, StateSuspCount, StateSuspoCount,
                    StateComoCount + StateHiboCount +
                       StateLefoCount + StateSuspoCount,
                    StateCefCount + StateColpgCount +
                       StateFpgCount + StateMwaitCount + StatePfwCount);
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   for (cnt = RowCount = 0; cnt < PROC_MAX; cnt++)
   {
      if (!(TopCpuTim[cnt] || TopPageFlts[cnt] ||
            TopBioCnt[cnt] || TopDioCnt[cnt]))
         continue;
      RowCount++;

      sptr += sprintf (sptr, "<tr>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopCpuTim[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %1.2f</td><td>(n/a)</td>",
                          TopCpuTim[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopCpuTim[cnt]->getspi$o_proc_lname),
                          (float)TopCpuTim[cnt]->getspi$l_proc_cputim/100.0);
      else
         sptr += sprintf (sptr, "<td></td><td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopPageFlts[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopPageFlts[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopPageFlts[cnt]->getspi$o_proc_lname),
                          TopPageFlts[cnt]->getspi$l_proc_pageflts);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopBioCnt[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopBioCnt[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopBioCnt[cnt]->getspi$o_proc_lname),
                          TopBioCnt[cnt]->getspi$l_proc_biocnt);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (TopDioCnt[cnt])
         sptr += sprintf (sptr, "<td>%8.8X %s %4u</td>",
                          TopDioCnt[cnt]->getspi$l_proc_epid,
                          MungeProcName(TopDioCnt[cnt]->getspi$o_proc_lname),
                          TopDioCnt[cnt]->getspi$l_proc_diocnt);
      else
         sptr += sprintf (sptr, "<td></td>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sptr += sprintf (sptr, "</tr>");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
   }

   sptr += sprintf (sptr, "\");");
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   if (RowCount)
      SystemData.ProcBufferLength = sptr - SystemData.ProcBuffer;
   else
      SystemData.ProcBuffer[SystemData.ProcBufferLength = 0] = '\0';

   if (Debug)
      fprintf (stdout, "\n%d\n%s\n",
               SystemData.ProcBufferLength, SystemData.ProcBuffer);
}

/*****************************************************************************/
/*
Collect and format per-process data.
*/ 

int CollectPerProc (CLIENT_DATA *clptr)

{
/* e.g. U-S-E-K- U+S-E-K- U^S-E-K- */
#define ASTFLG(bit,idx) { if (JpiAstEn & bit) if (JpiAstAct & bit) \
                          AstFlags[idx] = '^'; else AstFlags[idx] = '+'; \
                          else AstFlags[idx] = '-'; }

#define ALERT_THRESHOLD(lm,cnt) (((cnt)*100)/(lm) <= PerProcAlert) ? \
                                " style=\"color:red;\"" : ""

   static long  JpiAstAct,
                JpiAstCnt,
                JpiAstEn,
                JpiAstLm,
                JpiBioCnt,
                JpiBioLm,
                JpiBufIo,
                JpiBytCnt,
                JpiBytLm,
                JpiCpuId,
                JpiCpuLim,
                JpiCpuTim,
                JpiCreator,
                JpiDioCnt,
                JpiDioLm,
                JpiDirIo,
                JpiEnqCnt,
                JpiEnqLm,
                JpiExecTim,
                JpiFilCnt,
                JpiFilLm,
                JpiControlFlags = JPI$M_IGNORE_TARGET_STATUS,
                JpiGpgCnt,
                JpiJobPrcCnt,
                JpiJobType,
                JpiKtCount,
                JpiKrnlTim,
                JpiMode,
                JpiMultiThread,
                JpiOwner,
                JpiPageFlts,
                JpiPagFilCnt,
                JpiPgFlQuota,
                JpiPid,
                JpiPpgCnt,
                JpiPrcCnt,
                JpiPrcLm,
                JpiPri,
                JpiPrib,
                JpiState,
                JpiSts,
                JpiSts2,
                JpiSuprTim,
                JpiTqCnt,
                JpiTqLm,
                JpiUic,
                JpiUserTim,
                JpiVolumes,
                JpiWsAuth,
                JpiWsAuthExt,
                JpiWsExtent,
                JpiWsPeak,
                JpiWsQuota,
                JpiWsSize;

   static __int64  JpiFrePteCnt,
                   JpiLoginTim,
                   JpiVirtPeak;

   static unsigned short  JpiImagNameLen,
                          JpiTtAccPorNamLen,
                          JpiTtPhyDevNamLen;

   static char  JpiImagName [127+1],
                JpiPrcNam [15+1],
                JpiTtAccPorNam [80+1],
                JpiTtPhyDevNam [32+1],
                JpiTerminal [16+1],
                JpiUserName [12+1];

   /* initialize structure for SPI items */
   static struct
   {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemListJpi[] =
   {
       { sizeof(JpiControlFlags), JPI$_GETJPI_CONTROL_FLAGS,
         &JpiControlFlags, 0 },
       { sizeof(JpiAstAct), JPI$_ASTACT, &JpiAstAct, 0 },
       { sizeof(JpiAstCnt), JPI$_ASTCNT, &JpiAstCnt, 0 },
       { sizeof(JpiAstEn),  JPI$_ASTEN,  &JpiAstEn, 0 },
       { sizeof(JpiAstLm),  JPI$_ASTLM,  &JpiAstLm, 0 },
       { sizeof(JpiBioCnt), JPI$_BIOCNT, &JpiBioCnt, 0 },
       { sizeof(JpiBioLm),  JPI$_BIOLM,  &JpiBioLm, 0 },
       { sizeof(JpiBufIo),  JPI$_BUFIO,  &JpiBufIo, 0 },
       { sizeof(JpiBytCnt), JPI$_BYTCNT, &JpiBytCnt, 0 },
       { sizeof(JpiBytLm),  JPI$_BYTLM,  &JpiBytLm, 0 },
       { sizeof(JpiCpuId),  JPI$_CPU_ID,  &JpiCpuId, 0 },
       { sizeof(JpiCpuLim), JPI$_CPULIM,  &JpiCpuLim, 0 },
       { sizeof(JpiCpuTim), JPI$_CPUTIM, &JpiCpuTim, 0 },
       { sizeof(JpiCreator), JPI$_CREATOR, &JpiCreator, 0 },
       { sizeof(JpiDioCnt), JPI$_DIOCNT,  &JpiDioCnt, 0 },
       { sizeof(JpiDioLm),  JPI$_DIOLM,  &JpiDioLm, 0 },
       { sizeof(JpiDirIo),  JPI$_DIRIO,  &JpiDirIo, 0 },
       { sizeof(JpiEnqCnt), JPI$_ENQCNT, &JpiEnqCnt, 0 },
       { sizeof(JpiEnqLm),  JPI$_ENQLM,  &JpiEnqLm, 0 },
       { sizeof(JpiExecTim),   JPI$_EXECTIM,   &JpiExecTim, 0 },
       { sizeof(JpiFilCnt), JPI$_FILCNT, &JpiFilCnt, 0 },
       { sizeof(JpiFilLm),  JPI$_FILLM,  &JpiFilLm, 0 },
       { sizeof(JpiFrePteCnt), JPI$_FREPTECNT, &JpiFrePteCnt, 0 },
       { sizeof(JpiGpgCnt),    JPI$_GPGCNT,    &JpiGpgCnt, 0 },
       { sizeof(JpiJobPrcCnt), JPI$_JOBPRCCNT, &JpiJobPrcCnt, 0 },
       { sizeof(JpiJobType),   JPI$_JOBTYPE,   &JpiJobType, 0 },
       { sizeof(JpiKtCount),   JPI$_KT_COUNT,  &JpiKtCount, 0 },
       { sizeof(JpiKrnlTim),   JPI$_KRNLTIM,   &JpiKrnlTim, 0 },
       { sizeof(JpiMode),      JPI$_MODE,      &JpiMode, 0 },
       { sizeof(JpiMultiThread), JPI$_MULTITHREAD, &JpiMultiThread, 0 },
       { sizeof(JpiOwner),     JPI$_OWNER,     &JpiOwner, 0 },
       { sizeof(JpiPageFlts),  JPI$_PAGEFLTS,  &JpiPageFlts, 0 },
       { sizeof(JpiPagFilCnt), JPI$_PAGFILCNT, &JpiPagFilCnt, 0 },
       { sizeof(JpiPgFlQuota), JPI$_PGFLQUOTA, &JpiPgFlQuota, 0 },
       { sizeof(JpiPid),    JPI$_PID,    &JpiPid, 0 },
       { sizeof(JpiPpgCnt), JPI$_PPGCNT, &JpiPpgCnt, 0 },
       { sizeof(JpiPrcCnt), JPI$_PRCCNT, &JpiPrcCnt, 0 },
       { sizeof(JpiPrcLm),  JPI$_PRCLM,  &JpiPrcLm, 0 },
       { sizeof(JpiPri),    JPI$_PRI,    &JpiPri, 0 },
       { sizeof(JpiPrib),   JPI$_PRIB,   &JpiPrib, 0 },
       { sizeof(JpiState),  JPI$_STATE,  &JpiState, 0 },
       { sizeof(JpiSts),    JPI$_STS,    &JpiSts, 0 },
       { sizeof(JpiSts2),   JPI$_STS2,   &JpiSts2, 0 },
       { sizeof(JpiSuprTim),   JPI$_SUPRTIM,  &JpiSuprTim, 0 },
       { sizeof(JpiTqCnt),  JPI$_TQCNT,  &JpiTqCnt, 0 },
       { sizeof(JpiTqLm),   JPI$_TQLM,   &JpiTqLm, 0 },
       { sizeof(JpiUic),    JPI$_UIC,    &JpiUic, 0 },
       { sizeof(JpiUserTim),   JPI$_USERTIM,   &JpiUserTim, 0 },
       { sizeof(JpiVirtPeak),  JPI$_VIRTPEAK,  &JpiVirtPeak, 0 },
       { sizeof(JpiVolumes),   JPI$_VOLUMES,   &JpiVolumes, 0 },
       { sizeof(JpiWsAuth),    JPI$_WSAUTH,    &JpiWsAuth, 0 },
       { sizeof(JpiWsAuthExt), JPI$_WSAUTHEXT, &JpiWsAuthExt, 0 },
       { sizeof(JpiWsExtent),  JPI$_WSEXTENT,  &JpiWsExtent, 0 },
       { sizeof(JpiWsPeak),    JPI$_WSPEAK,    &JpiWsPeak, 0 },
       { sizeof(JpiWsQuota),   JPI$_WSQUOTA,   &JpiWsQuota, 0 },
       { sizeof(JpiWsSize),    JPI$_WSSIZE,    &JpiWsSize, 0 },
       { sizeof(JpiLoginTim),  JPI$_LOGINTIM,  &JpiLoginTim, 0 },
       { sizeof(JpiImagName),  JPI$_IMAGNAME,  &JpiImagName, &JpiImagNameLen },
       { sizeof(JpiPrcNam),    JPI$_PRCNAM,    &JpiPrcNam, 0 },
       { sizeof(JpiTerminal),  JPI$_TERMINAL,    &JpiTerminal, 0 },
       { sizeof(JpiTtAccPorNam), JPI$_TT_ACCPORNAM,
         &JpiTtAccPorNam, &JpiTtAccPorNamLen },
       { sizeof(JpiTtPhyDevNam), JPI$_TT_PHYDEVNAM,
         &JpiTtPhyDevNam, &JpiTtPhyDevNamLen },
       { sizeof(JpiUserName),  JPI$_USERNAME,  &JpiUserName, 0 },
       {0,0,0,0}
   };

   static char  AlertMsg [128],
                UicString [64];
   static $DESCRIPTOR (AlertMsgDsc, AlertMsg);
   static $DESCRIPTOR (UicStringDsc, UicString);
   static $DESCRIPTOR (UicFaoDsc, "!%U\0");

   int  status;
   char  *cptr, *sptr, *zptr,
         *mode, *state, *type;
   unsigned long  pid,
                  BufIo,
                  CpuTime,
                  CpuSekTime,
                  CpuPercent,
                  CpuSekPercent,
                  DirIo,
                  PageFlts;
   __int64  ConBinTime,
            CpuBinTime,
            CpuExecTime,
            CpuKrnlTime,
            CpuLimTime,
            CpuSuprTime,
            CpuUserTime;
   unsigned short  slen;
   char  AstFlags [16],
         CpuHash [16],
         CpuStats [64],
         PcbFlags [64*8];
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectPerProc()\n");

   clptr->PerProcBuffer[clptr->PerProcBufferLength = 0] = '\0';

   if ((pid = clptr->PerProcPid) != clptr->PrevPerProcPid)
   {
      clptr->PrevPerProcPid = clptr->PerProcPid;
      clptr->PrevCpuTime = clptr->PrevCpuSekTime = clptr->PrevJpiBufIo =
         clptr->PrevJpiDirIo = clptr->PrevJpiPageFlts =
         clptr->PerProcDatum = 0;
   }

   /****************/
   /* collect data */
   /****************/

   status = sys$getjpiw (
            SpiEfn,  /* efn */
            &pid,  /* pidaddr */
            0,  /* prcnam */
            &ItemListJpi,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getjpi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status))
   {
      sys$getmsg (status, &slen, &AlertMsgDsc, 0, 0); 
      AlertMsg[slen] = '\0';
      zptr = (sptr = clptr->PerProcBuffer) + sizeof(clptr->PerProcBuffer)-256;
      sptr += sprintf (sptr,
"the.mdsMonPerProc(\'<tr><td>%08.08X %%X%08.08X %s</td></tr>\');",
                       pid, status, AlertMsg);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
      clptr->PerProcBufferLength = sptr - clptr->PerProcBuffer;
      return (status);
   }

   switch (JpiMode)
   {
      case  0 : mode = "OTH"; break;
      case  1 : mode = "NET"; break;
      case  2 : mode = "BAT"; break;
      case  3 : mode = "INT"; break;
      default : mode = "?";
   }

   switch (JpiJobType)
   {
      case 0 : type = "DET"; break;
      case 1 : type = "NET"; break;
      case 2 : type = "BAT"; break;
      case 3 : type = "LOC"; break;
      case 4 : type = "DIA"; break;
      case 5 : type = "REM"; break;
      default : type = "?";
   }

   switch (JpiState)
   {
      case  1 : state = "COLPG"; break;
      case  2 : state = "MWAIT"; break;
      case  3 : state = "CEF"; break;
      case  4 : state = "PFW"; break;
      case  5 : state = "LEF"; break;
      case  6 : state = "LEFO"; break;
      case  7 : state = "HIB"; break;
      case  8 : state = "HIBO"; break;
      case  9 : state = "SUSP"; break;
      case 10 : state = "SUSPO"; break;
      case 11 : state = "FPG"; break;
      case 12 : state = "COM"; break;
      case 13 : state = "COMO"; break;
      case 14 : state = "CUR"; break;
      default : state = "?";
   }

   /* AST flags */
   AstFlags[0] = 'U'; ASTFLG(1,1)
   AstFlags[2] = 'S'; ASTFLG(2,3)
   AstFlags[4] = 'E'; ASTFLG(4,5)
   AstFlags[6] = 'K'; ASTFLG(8,7)
   AstFlags[8] = '\0';

   PcbFlags[0] = '\0';
   if (JpiSts & 0x1) strcat (PcbFlags, " RES");
   if (JpiSts & 0x2) strcat (PcbFlags, " DELPEN");
   if (JpiSts & 0x4) strcat (PcbFlags, " FORCPEN");
   if (JpiSts & 0x8) strcat (PcbFlags, " INQUAN");
   if (JpiSts & 0x10) strcat (PcbFlags, " PSWAPM");
   if (JpiSts & 0x20) strcat (PcbFlags, " RESPEN");
   if (JpiSts & 0x40) strcat (PcbFlags, " SSFEXC");
   if (JpiSts & 0x80) strcat (PcbFlags, " SSFEXCE");
   if (JpiSts & 0x100) strcat (PcbFlags, " SSFEXCS");
   if (JpiSts & 0x200) strcat (PcbFlags, " SSFEXCU");
   if (JpiSts & 0x400) strcat (PcbFlags, " SSRWAIT");
   if (JpiSts & 0x800) strcat (PcbFlags, " SUSPEN");
   if (JpiSts & 0x2000) strcat (PcbFlags, " WALL");
   if (JpiSts & 0x4000) strcat (PcbFlags, " BATCH");
   if (JpiSts & 0x8000) strcat (PcbFlags, " NOACNT");
   if (JpiSts & 0x10000) strcat (PcbFlags, " NOSUSPEND");
   if (JpiSts & 0x20000) strcat (PcbFlags, " ASTPEN");
   if (JpiSts & 0x40000) strcat (PcbFlags, " PHDRES");
   if (JpiSts & 0x80000) strcat (PcbFlags, " HIBER");
   if (JpiSts & 0x100000) strcat (PcbFlags, " LOGIN");
   if (JpiSts & 0x200000) strcat (PcbFlags, " NETWRK");
   if (JpiSts & 0x400000) strcat (PcbFlags, " PWRAST");
   if (JpiSts & 0x800000) strcat (PcbFlags, " NODELET");
   if (JpiSts & 0x1000000) strcat (PcbFlags, " DISAWS");
   if (JpiSts & 0x2000000) strcat (PcbFlags, " INTER");
   if (JpiSts & 0x4000000) strcat (PcbFlags, " RECOVER");
   if (JpiSts & 0x8000000) strcat (PcbFlags, " 0x8000000");
   if (JpiSts & 0x1000000) strcat (PcbFlags, " HARDAFF");
   if (JpiSts & 0x2000000) strcat (PcbFlags, " ERDACT");
   if (JpiSts & 0x4000000) strcat (PcbFlags, " SOFTSUSP");
   if (JpiSts & 0x8000000) strcat (PcbFlags, " PREEMPTED");
   if (JpiSts2 & 0x1) strcat (PcbFlags, " QRESCHED");
   if (JpiSts2 & 0x2) strcat (PcbFlags, " DEPLOCK");
   if (JpiSts2 & 0x3) strcat (PcbFlags, " FREDLOCK");
   if (JpiSts2 & 0x4) strcat (PcbFlags, " PHDLOCK");
   if (JpiSts2 & 0x10) strcat (PcbFlags, " TCB");
   if (JpiSts2 & 0x20) strcat (PcbFlags, " TBSPEND");
   if (JpiSts2 & 0x40) strcat (PcbFlags, " SSLOGENAB");
   if (JpiSts2 & 0x80) strcat (PcbFlags, " SSLOGPERM");
   if (JpiSts2 & 0x100) strcat (PcbFlags, " BRKRUNLOAD");
   if (JpiSts2 & 0x8000) strcat (PcbFlags, " CLASSSCHEDPERM");
   if (JpiSts2 & 0x10000) strcat (PcbFlags, " TERMNOTIFY");
   if (JpiSts2 & 0x20000) strcat (PcbFlags, " BYTLMLOAN");
   if (JpiSts2 & 0x40000) strcat (PcbFlags, " DISPREEMPT");
   if (JpiSts2 & 0x80000) strcat (PcbFlags, " NOUNSHELV");
   if (JpiSts2 & 0x100000) strcat (PcbFlags, " SHELVRES");
   if (JpiSts2 & 0x200000) strcat (PcbFlags, " CLASSCHED");
   if (JpiSts2 & 0x400000) strcat (PcbFlags, " CLASSUPP");
   if (JpiSts2 & 0x800000) strcat (PcbFlags, " INTBSSTATE");
   if (JpiSts2 & 0x1000000) strcat (PcbFlags, " WINDFALL");
   if (JpiSts2 & 0x2000000) strcat (PcbFlags, " NOTIFY");
   if (JpiSts2 & 0x3C000000) strcat (PcbFlags, " SINTHREAD");
   if (JpiSts2 & 0x40000000) strcat (PcbFlags, " RWAST");
   if (JpiSts2 & 0x80000000) strcat (PcbFlags, " SOFTSINTHR");

   /* terminate at first trailing space */
   JpiUserName[sizeof(JpiUserName)-1] = '\0';
   for (sptr = JpiUserName; *sptr && *sptr != ' '; sptr++);
   *sptr = '\0';

   /* terminate and trim trailing spaces */
   JpiPrcNam[sizeof(JpiPrcNam)-1] = '\0';
   for (sptr = JpiPrcNam; *sptr; sptr++);
   *sptr-- = '\0';
   while (sptr > JpiPrcNam && *sptr == ' ') sptr--;
   *++sptr = '\0';

   /* terminate based on returned length */
   JpiImagName[JpiImagNameLen] = '\0';
   if (!JpiImagName[0]) strcpy (JpiImagName, "(DCL)");
   JpiTtAccPorNam[JpiTtAccPorNamLen] = '\0';
   if (!JpiTerminal[0]) strcpy (JpiTerminal, "(none)");

   sys$fao (&UicFaoDsc, 0, &UicStringDsc, JpiUic);

   ConBinTime = JpiLoginTim - CurrentBinTime;

   /* make the 10mS ticks delta times (MungeTime() will handle the zeros) */
   CpuBinTime = JpiCpuTim * -100000;
   CpuLimTime = JpiCpuLim * -100000;
   CpuExecTime = JpiExecTim * -100000;
   CpuKrnlTime = JpiKrnlTim * -100000;
   CpuSuprTime = JpiSuprTim * -100000;
   CpuUserTime = JpiUserTim * -100000;

   if (JpiCpuId >= 0)
      sprintf (CpuHash, " CPU#%d", JpiCpuId);
   else
      CpuHash[0] = '\0';

   if (JpiCpuLim)
      sprintf (CpuStats, "%s / %s %u%%",
               MungeTime(&CpuLimTime),
               MungeTime(&CpuBinTime),
               (JpiCpuLim*100)/(JpiCpuTim));
   else
      sprintf (CpuStats, "%s", MungeTime(&CpuBinTime));

   zptr = (sptr = clptr->PerProcBuffer) + sizeof(clptr->PerProcBuffer)-256;

   /***************/
   /* build table */
   /***************/

   sptr += sprintf (sptr, "the.mdsMonPerProc(\'");
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   sptr += sprintf (sptr,
"<tr><td><b onclick=\"mdsPidPrompt(event)\">PID:</b> %8.8X \
%s:<span onclick=\"mdsPidClick(event)\">%08.08X</span></td>\
<td><b>USER:</b> %s %s</td><td><b>PRCN:</b> %s</td>",
                    JpiPid,
                    JpiOwner ? "O" : "C", JpiOwner ? JpiOwner : JpiCreator,
                    JpiUserName, UicString, MungeString(JpiPrcNam));
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   sptr += sprintf (sptr,
"<td rowspan=\"2\" colspan=\"2\" style=\"white-space:pre-wrap;\">\
<b>IMAG:</b> %s</td></tr>",
                    MungeString(JpiImagName));
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   sptr += sprintf (sptr,
"<tr><td><b>PRI:</b> %u/%u</td>\
<td><b>LGIN:</b> %s</td><td><b>CNCT:</b> %s</td></tr>",
                    JpiPrib, JpiPri,
                    /** HP C V7.3-009 on OpenVMS Alpha V8.3 PTRMISMATCH? **/
                    MungeTime(&JpiLoginTim),
                    MungeTime(&ConBinTime));
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   sptr += sprintf (sptr,
"<tr><td><b>STA:</b> %s%s %s %s:%s%s</td>\
<td></td><td></td><td colspan=\"2\"><b>TERM:</b> %s%s%s%s</td>",
                    state, CpuHash, AstFlags, type, mode, PcbFlags,
                    JpiTerminal, JpiTtAccPorNam[0] ? " (" : "",
                                 JpiTtAccPorNam,
                                 JpiTtAccPorNam[0] ? ")" : "");
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   sptr += sprintf (sptr,
"<tr><td colspan=\"2\"><b>CPU:</b> %s  \
<b>U</b> %s  <b>S</b> %s  <b>E</b> %s  <b>K</b> %s</td>",
                    CpuStats,
                    MungeTime(&CpuUserTime),
                    MungeTime(&CpuSuprTime),
                    MungeTime(&CpuExecTime),
                    MungeTime(&CpuKrnlTime));
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   sptr += sprintf (sptr,
"<td%s><b>WSET:</b> %.4g / %.4g %.4g MB</td>\
<td><b>PGS:</b> %.4g + %.4g = %.3g MB</td>\
<td%s><b>PGFL:</b> %.5g / %.5g MB %u%%</td></tr>",
                    ALERT_THRESHOLD(JpiWsExtent,
                                    JpiWsExtent-JpiPpgCnt+JpiGpgCnt),
                    (float)JpiWsExtent * 0.0005,
                    /* current usage is private plus global pages */
                    ((float)JpiPpgCnt + (float)JpiGpgCnt) * 0.0005,
                    (float)JpiWsPeak * 0.0005,
                    /* private and global pages */
                    (float)JpiPpgCnt * 0.0005,
                    (float)JpiGpgCnt * 0.0005,
                    ((float)JpiPpgCnt + (float)JpiGpgCnt) * 0.0005,
                    /* page file usage */
                    ALERT_THRESHOLD(JpiPgFlQuota,JpiPagFilCnt),
                    (float)JpiPgFlQuota * 0.0005,
                    (float)JpiPagFilCnt * 0.0005,
                    (int)(((float)JpiPagFilCnt*100.0)/(float)JpiPgFlQuota));
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   sptr += sprintf (sptr,
"<tr><td%s><b>AST:</b> %u / %u %u%%</td><td%s><b>BIO:</b> %u / %u %u%%</td>\
<td%s><b>DIO:</b> %u / %u %u%%</td><td%s><b>ENQ:</b> %u / %u %u%%</td>\
<td><b>VIRT:</b> %.6g / %.6g MB</td></tr>",
                    ALERT_THRESHOLD(JpiAstLm,JpiAstCnt),
                    JpiAstLm, JpiAstCnt, (JpiAstCnt*100)/(JpiAstLm),
                    ALERT_THRESHOLD(JpiBioLm,JpiBioCnt),
                    JpiBioLm, JpiBioCnt, (JpiBioCnt*100)/(JpiBioLm),
                    ALERT_THRESHOLD(JpiDioLm,JpiDioCnt),
                    JpiDioLm, JpiDioCnt, (JpiDioCnt*100)/(JpiDioLm),
                    ALERT_THRESHOLD(JpiEnqLm,JpiEnqCnt),
                    JpiEnqLm, JpiEnqCnt, (JpiEnqCnt*100)/(JpiEnqLm),
                    /* possible is minimum physical plus maxium page file */
                    ((float)JpiWsQuota + (float)JpiPgFlQuota) * 0.0005, 
                    (float)JpiVirtPeak*0.0005);
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   /* careful of MULTITHREAD div zero with percentage! */
   sptr += sprintf (sptr,
"<tr><td%s><b>FIL:</b> %u / %u %u%%</td><td%s><b>BYT:</b> %u / %u %u%%</td>\
<td%s><b>PRC:</b> %u / %u %u%%</td><td%s><b>TQ:</b> %u / %u %u%%</td>\
<td><b>KTH:</b> %u / %u</td></tr>",
                    ALERT_THRESHOLD(JpiFilLm,JpiFilCnt),
                    JpiFilLm, JpiFilCnt, (JpiFilCnt*100)/(JpiFilLm),
                    ALERT_THRESHOLD(JpiBytLm/1000,JpiBytCnt/1000),
                    JpiBytLm, JpiBytCnt,
                       (int)(((float)JpiBytCnt*100.0)/(float)JpiBytLm),
                    ALERT_THRESHOLD(JpiPrcLm,JpiPrcLm-JpiPrcCnt),
                    JpiPrcLm, JpiPrcCnt, ((JpiPrcLm-JpiPrcCnt)*100)/JpiPrcLm,
                    ALERT_THRESHOLD(JpiTqLm,JpiTqCnt),
                    JpiTqLm, JpiTqCnt, (JpiTqCnt*100)/(JpiTqLm),
                    JpiMultiThread, JpiKtCount);
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   sptr += sprintf (sptr,
"<tr class=\"per_proc_last\"><td><b>FLT:</b> %u</td><td><b>BIO:</b> %u</td>\
<td><b>DIO:</b> %u</td><td><b>VOL:</b> %u</td>\
<td><input id=\"per_proc_hist_btn\" class=\"btn btn3\" \
type=\"submit\" onclick=\"return mdsPerProcHistClick(event)\" \
value=\"history\"></td></tr>",
                    JpiPageFlts, JpiBufIo, JpiDirIo, JpiVolumes);
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   sptr += sprintf (sptr, "\');");
   if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

   /******************/
   /* update history */
   /******************/

   CpuTime = JpiUserTim + JpiSuprTim + JpiExecTim + JpiKrnlTim;
   CpuPercent = CpuTime - clptr->PrevCpuTime;
   clptr->PrevCpuTime = CpuTime;

   CpuSekTime = JpiSuprTim + JpiExecTim + JpiKrnlTim;
   CpuSekPercent = CpuSekTime - clptr->PrevCpuSekTime;
   clptr->PrevCpuSekTime = CpuSekTime;

   /* only after we've collected at least one previously */
   if (clptr->PerProcDatum++)
   {
      CpuPercent = (uint)((float)CpuPercent / DeltaSeconds);
      if (CpuPercent > 100 * SystemData.ActiveCpuCount)
         CpuPercent = 100 * SystemData.ActiveCpuCount;

      CpuSekPercent = (uint)((float)CpuSekPercent / DeltaSeconds);
      if (CpuSekPercent > 100 * SystemData.ActiveCpuCount)
         CpuSekPercent = 100 * SystemData.ActiveCpuCount;

      sptr += sprintf (sptr,
"the.mdsDatum(\"proc_cpu\",%u,%u,%u);\
the.mdsDatum(\"proc_wset\",%u,%u,%u);\
the.mdsDatum(\"proc_pgs\",-1,%u,%u);",
                       SystemData.ActiveCpuCount*100,
                       CpuPercent,
                       CpuSekPercent,
                       /* wset */
                       JpiWsExtent * 512,
                       /* current usage is private plus global pages */
                       JpiPpgCnt * 512 + JpiGpgCnt * 512,
                       /* pgs */
                       JpiGpgCnt * 512,
                       JpiPpgCnt * 512,
                       JpiGpgCnt * 512);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      sptr += sprintf (sptr, "the.mdsDatum(\"proc_pgfl\",%llu,%llu);",
                       (__int64)JpiPgFlQuota * 512,
                       ((__int64)(JpiPgFlQuota - JpiPagFilCnt)) * 512);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      BufIo = (uint)((float)(JpiBufIo - clptr->PrevJpiBufIo) / DeltaSeconds);
      DirIo = (uint)((float)(JpiDirIo - clptr->PrevJpiDirIo) / DeltaSeconds);
      PageFlts = (uint)((float)(JpiPageFlts -
                                clptr->PrevJpiPageFlts) / DeltaSeconds);
      sptr += sprintf (sptr,
"the.mdsDatum(\"proc_bio\",-1,%u);\
the.mdsDatum(\"proc_dio\",-1,%u);\
the.mdsDatum(\"proc_flts\",-1,%u);",
                       BufIo, DirIo, PageFlts);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
   }

   clptr->PrevJpiBufIo = JpiBufIo;
   clptr->PrevJpiDirIo = JpiDirIo;
   clptr->PrevJpiPageFlts = JpiPageFlts;

   clptr->PerProcBufferLength = sptr - clptr->PerProcBuffer;

   if (Debug) fputs (clptr->PerProcBuffer, stdout);
}

/*****************************************************************************/
/*
General access and then access to the top-process and then again per-process
displays is controlled by three strings read from a multi-valued logical name. 
The control strings sets the client access booleans.  For WebSocket Must be
called while CGI variables are valid.
*/

void AccessControl (CLIENT_DATA *clptr)

{
   static $DESCRIPTOR (LogNameDsc, "MONDESI_ACCESS");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   static long  LnmAttributes,
                LnmIndex;
   static unsigned short  LogValueLen;
   static char  LogValue[3][255+1];

   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LnmIndex), LNM$_INDEX, &LnmIndex, 0 },
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { sizeof(LogValue[0])-1, LNM$_STRING, 0, &LogValueLen },
      { 0,0,0,0 }
   };

   int  status,
        AddrHit, AddrMatch, NameHit, NameMatch;
   unsigned int  idx, IPaddr, IPnet, mask;
   char  ch;
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AccessControl() |%s|%s|\n",
               CgiRemoteUserPtr, CgiRemoteAddrPtr);

   /* initialise and attempt to read up to three values */
   for (LnmIndex = 0; LnmIndex <= 2; LnmIndex++)
      LogValue[LnmIndex][0] = '\0';
   for (LnmIndex = 0; LnmIndex <= 2; LnmIndex++)
   {
      LnmItems[2].BufferPtr = LogValue[LnmIndex];
      status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, 0, &LnmItems);
      if (VMSnok (status) || !(LnmAttributes & LNM$M_EXISTS)) break;
      LogValue[LnmIndex][LogValueLen] = '\0';
   }

   if (Debug)
      fprintf (stdout, "|%s|%s|%s|\n",
               LogValue[0], LogValue[1], LogValue[2]);

   if (!LogValue[0][0])
   {
      /* logical name does not exist, therefore general access allowed */
      clptr->AccessAllowed[ACCESS_GENERAL] = TRUE;
      clptr->AccessAllowed[ACCESS_TOP_PROCESS] =
         clptr->AccessAllowed[ACCESS_PER_PROCESS] = FALSE;
      return;
   }

   /* first general access, then top-process access, then per-process access */
   for (idx = 0; idx <= 2; idx++)
   {
      AddrHit = AddrMatch = NameHit = NameMatch = FALSE;
      cptr = LogValue[idx];

      while (*cptr)
      {
         while (*cptr && *cptr != '_' && *cptr != '*' &&
                         *cptr != '$' && *cptr != '#') cptr++;
         if (!*cptr) break;

         if (*cptr == '_')
         {
            /* none allowed access */
            AddrHit = AddrMatch = NameHit = NameMatch = FALSE;
            break;
         }

         if (*cptr == '*')
         {
            /* all allowed access */
            AddrHit = AddrMatch = NameHit = NameMatch = TRUE;
            break;
         }

         if ((*(USHORTPTR)cptr) == '$*')
         {
            /* allowed if authenticated */
            cptr += 2;
            NameHit = TRUE;
            if (CgiRemoteUserPtr[0]) NameMatch = TRUE;
            continue;
         }

         if (*cptr == '$')
         {
            /* allowed if matching authentication */
            cptr++;
            NameHit = TRUE;
            for (zptr = cptr; *zptr && *zptr != ','; zptr++);
            if (ch = *zptr) *zptr = '\0';
            sptr = CgiRemoteUserPtr;
            while (*cptr && *sptr && toupper(*cptr) == toupper(*sptr))
            {
               cptr++;
               sptr++;
            }
            if (!*cptr && !*sptr) NameMatch = TRUE;
            *zptr = ch;
            continue;
         }

         if (*cptr == '#')
         {
            /* access allowed if matching client IP(v4) address */
            AddrHit = TRUE;
            IPaddr = inet_addr (CgiRemoteAddrPtr);
            if (IPaddr == 0xffffffff) return;

            cptr++;
            for (zptr = cptr; *zptr && *zptr != ','; zptr++);
            if (ch = *zptr) *zptr = '\0';
            for (sptr = cptr; *sptr && *sptr != '/'; sptr++);
            if (*sptr)
            {
               /* subnet mask */
               *sptr = '\0';
               /* inet_aton() is not available on VMS V7.2 */
               IPnet = inet_addr (cptr);
               *sptr = '/';
               *zptr = ch;
               if (IPnet == 0xffffffff) return;
               mask = atoi(sptr+1);
               mask = 0xffffffff >> (32 - mask); 
               if (IPaddr && IPnet && (IPnet == (IPaddr & mask)))
               {
                  /* both addr are valid and masked client addr matches */
                  AddrMatch = TRUE;
                  continue;
               }
            }
            if (!strcmp (cptr, CgiRemoteAddrPtr))
            {
               /* matched address */
               *zptr = ch;
               AddrMatch = TRUE;
               continue;
            }
            if (*(cptr = zptr) = ch) cptr++;
            continue;
         }
      }

      if (AddrHit && !NameHit)
         clptr->AccessAllowed[idx] = AddrMatch;
      else
      if (!AddrHit && NameHit)
         clptr->AccessAllowed[idx] = NameMatch;
      else
      if (AddrHit && NameHit)
         clptr->AccessAllowed[idx] = AddrMatch && NameMatch;
   }

   if (Debug)
      fprintf (stdout, "|%d|%d|%d|\n",
               clptr->AccessAllowed[0], clptr->AccessAllowed[1],
               clptr->AccessAllowed[2]);
}

/*****************************************************************************/
/*
HTML-escape any forbidden characters in a null-terminated string.
Static buffer so can only be called the once per print call.
Up to 255 character resultant string length.
*/

char* MungeString (char *cptr)

{
   static char  StringBuffer [255+6];

   char  *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   zptr = (sptr = StringBuffer) + sizeof(StringBuffer)-6;
   while (*cptr && sptr < zptr)
   {
      switch (*cptr)
      {
         case '<' :
            *(ULONGPTR)sptr = '&lt;';
            sptr += 4;
            break;
         case '>' :
            *(ULONGPTR)sptr = '&gt;';
            sptr += 4;
            break;
         case '&' :
            *(ULONGPTR)sptr = '&amp';
            sptr += 4;
            *sptr++ = ';';
            break;
         default :
            *sptr++ = *cptr;
      }
      cptr++;
   }
   *sptr = '\0';
   return (StringBuffer);
}

/*****************************************************************************/
/*
HTML-escape any forbidden characters in the process name.
Process name is a maximum fifteen character counted string.
Static buffer so can only be called the once per print call.
*/

char* MungeProcName (char *ProcName)

{
   /* maximum fifteen ampersand's worth */
   static char  NameBuffer [15*5+1];

   int  cnt, filcnt;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   sptr = NameBuffer;
   cnt = *ProcName;
   filcnt = 15 - cnt;
   cptr = ProcName + 1;
   while (cnt--)
   {
      switch (*cptr)
      {
         case '<' :
            *(ULONGPTR)sptr = '&lt;';
            sptr += 4;
            break;
         case '>' :
            *(ULONGPTR)sptr = '&gt;';
            sptr += 4;
            break;
         case '&' :
            *(ULONGPTR)sptr = '&amp';
            sptr += 4;
            *sptr++ = ';';
            break;
         default :
            *sptr++ = *cptr;
      }
      cptr++;
   }
   while (filcnt--) *sptr++ = ' ';
   *sptr = '\0';

   return (NameBuffer);
}

/*****************************************************************************/
/*
Make a delta binary time into an null-terminated ASCII string, absorbing
non-significant leading characters.  Up to four concurrent times are stored in
static storage.
*/

char* MungeTime (__int64 *btptr)

{
#define MAX_TIME_BUF 4
   static int  TimeBufCount;
   static char  TimeBuf [MAX_TIME_BUF][32];
   static $DESCRIPTOR (TimeBufDsc, TimeBuf[0]);
   static $DESCRIPTOR (TimeFaoDsc, "!%D\0");

   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MungeTime()\n");

   TimeBufDsc.dsc$a_pointer = TimeBuf[TimeBufCount++ % MAX_TIME_BUF];
   if (*btptr == 0)
      strcpy (TimeBufDsc.dsc$a_pointer, "0.00");
   else
   {
      sys$fao (&TimeFaoDsc, 0, &TimeBufDsc, btptr);
      for (cptr = sptr = TimeBufDsc.dsc$a_pointer;
           (*cptr == ' ' || *cptr == '0' || *cptr == ':') && cptr - sptr <= 11;
           cptr++);
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';
   }
   return (TimeBufDsc.dsc$a_pointer);
}

/*****************************************************************************/
/*
Sets up a WebSocket client which then runs independently and asynchronously in
a CGIplus environment.
*/

void WebSockRequest ()

{
   int  len;
   char  ProcButton [32];
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   /* allocate a data structure for this connected client */
   clptr = calloc (1, sizeof(CLIENT_DATA));
   if (!clptr) EXIT_FI_LI (vaxc$errno);

   AccessControl (clptr);

   if (!clptr->AccessAllowed[ACCESS_GENERAL])
   {
      fputs (ACCESS_NOT_PERMITTED, stdout);
      free (clptr);
      return;
   }

   /* allocate a WebSocket library structure */
   if (!(clptr->WsLibPtr = WsLibCreate (clptr, WebSockDestroy)))
   {
      /* failed, commonly on some WebSocket protocol issue */
      free (clptr);
      return;
   }

   WebSockConnectedCount++;

   if (clptr->AccessAllowed[ACCESS_TOP_PROCESS])
   {
      clptr->ProcRMI = atoi(CgiQueryStringPtr);
      if (clptr->AccessAllowed[ACCESS_PER_PROCESS])
         if (CgiQueryStringPtr[0] && CgiQueryStringPtr[1])
            clptr->PerProcPid = strtol(CgiQueryStringPtr+2,NULL,16);
   }

   /* open the WebSocket IPC to the server (and from there to the client) */
   WsLibOpen (clptr->WsLibPtr);

   /* [proc] button depending on access control */
   len = sprintf (ProcButton, "mdsProcButton(%s,%s);",
                  clptr->AccessAllowed[ACCESS_TOP_PROCESS] ? "true" : "false",
                  clptr->AccessAllowed[ACCESS_PER_PROCESS] ? "true" : "false");
   WsLibWrite (clptr->WsLibPtr, ProcButton, len, WSLIB_ASYNCH);

   WsLibRead (clptr->WsLibPtr,
              clptr->InputBuffer,
              sizeof(clptr->InputBuffer)-1,
              WebSockReadAst);
}

/*****************************************************************************/
/*
WebSocket has closed, request concluded.
*/

void WebSockDestroy (struct WsLibStruct *wsptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WebSockDestroy()\n");

   free (WsLibGetUserData (wsptr));

   if (WebSockConnectedCount) WebSockConnectedCount--;
}

/*****************************************************************************/
/*
Enable/disable collection and display of specific data.
*/

void WebSockReadAst (struct WsLibStruct *wsptr)

{
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "WebSockReadAst() %%X%08.08X\n",
               WsLibReadStatus (wsptr));

   if (VMSnok (WsLibReadStatus (wsptr)))
   {
      WsLibClose (wsptr, 0, NULL);
      return;
   }

   clptr = WsLibGetUserData (wsptr);

   /* ensure it's null-terminated */
   clptr->InputBuffer[WsLibReadCount(wsptr)] = '\0';

   if (!strncmp (clptr->InputBuffer, "perproc=", 8))
      clptr->PerProcPid = strtol(clptr->InputBuffer+8, NULL, 16);
   else
   if (!strncmp (clptr->InputBuffer, "topproc=", 8))
      clptr->ProcRMI = atoi(clptr->InputBuffer+8);
   else
      sprintf (NotePad, "$[%d] |%16.16s| %04.04x%04.04x%04.04x",
               __LINE__, clptr->InputBuffer,
               ntohl(*((unsigned long*)clptr->InputBuffer)),
               ntohl(*((unsigned long*)(clptr->InputBuffer+4))),
               ntohl(*((unsigned long*)(clptr->InputBuffer+8))));

   /* read next */
   WsLibRead (wsptr,
              clptr->InputBuffer,
              sizeof(clptr->InputBuffer)-1,
              WebSockReadAst);
}

/*****************************************************************************/
/*
If WASD and the version >= 10.1 then WebSocket is potentially available.  The
multi-valued logical name MONDESI_WEBSOCKET allows this to be overridden by
defining the first value to be zero, or one.  Further, additional values can
represent partial strings from the user-agent field.  A match returns the
complement of the first value.  So, 0,"Firefox" would disable WebSocket for all
but user-agents containing the string "Firefox".  If the leading character is
'!' then negates the match, so 0,"!Windows NT","Firefox" would disable
WebSocket for all but user-agents containing the string "Firefox" unless it
also contained the string "Windows NT".  Could be used with the first value
defined as 1, in which case it would function to exclude usage by specified
WebSocket agent strings.  With CGIplus-enabled WebSocket, as it is directly
accessing a CGI variable, can only be used before the request setup has
completed and the asynchronous WebSocket I/O has begun.  Bit of a kludge but
allows the flexibility currently required. 
*/

int WebSockAvailable ()

{
   static long  LnmIndex = -1,
                LnmAttributes;
   static unsigned short  LogValueLen;
   static char  LogValue [255+1];
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LnmIndex), LNM$_INDEX, &LnmIndex, 0 },
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { sizeof(LogValue)-1, LNM$_STRING, &LogValue, &LogValueLen },
      { 0,0,0,0 }
   };
   static $DESCRIPTOR (LogNameDsc, "MONDESI_WEBSOCKET");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   int  cnt, status,
        WebSocketOk = 0;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WebSockAvailable()\n");

   if (cptr = WsLibCgiVarNull("SERVER_SOFTWARE"))
      if (cptr = strstr (cptr, "WASD/"))  /* e.g. "WASD/10.2" */
         if (*(cptr+7) == '.')
            if ((atoi(cptr+5) >= 10 && atoi(cptr+8) >= 1) || atoi(cptr+5) >= 11)
               WebSocketOk = 1;

   if (!WebSocketOk) return (WebSocketOk);

   for (LnmIndex = 0; LnmIndex <= 127; LnmIndex++)
   {
      status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, 0, &LnmItems);
      if (VMSnok (status) || !(LnmAttributes & LNM$M_EXISTS)) break;
      LogValue[LogValueLen] = '\0';
      if (!LnmIndex)
         WebSocketOk = atoi(LogValue);
      else
      if (LogValue[0] == '!') {
         if (strstr (CgiHttpUserAgentPtr, LogValue+1))
            return (WebSocketOk);
      }
      else
      if (strstr (CgiHttpUserAgentPtr, LogValue))
         return (!WebSocketOk);
   }

   return (WebSocketOk);
}

/*****************************************************************************/
/*
Decode URL-encoded string.  Resultant string is always the same size or smaller
so it can be done in-situ!  Returns the size of the resultant string.
*/

int UrlDecode (char *String)

{
   unsigned char  ch;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "UrlDecode() |%s|\n", String);

   cptr = sptr = String;
   while (*cptr)
   {
      switch (*cptr)
      {
         case '+' :

            *sptr++ = ' ';
            cptr++;

            continue;

         case '%' :

            cptr++;
            while (*cptr == '\r' || *cptr == '\n') cptr++;
            ch = 0;
            if (*cptr >= '0' && *cptr <= '9')
               ch = (*cptr - (int)'0') << 4;
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               ch = (toupper(*cptr) - (int)'A' + 10) << 4;
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            while (*cptr == '\r' || *cptr == '\n') cptr++;
            if (*cptr >= '0' && *cptr <= '9')
               ch += (*cptr - (int)'0');
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               ch += (toupper(*cptr) - (int)'A' + 10);
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            *sptr++ = ch;

            continue;

         case '\r' :
         case '\n' :

            cptr++;
            continue;

         default :
            
            *sptr++ = *cptr++;

      }
   }
   *sptr = '\0';

   if (Debug) fprintf (stdout, "|%s|\n", String);
   return (sptr-String);
}

/*****************************************************************************/
/*
Control debug characteristics to allow use in a CGIplus environment.
Set <stdio> to record if debug enabled or binary if not.
*/

void SetDebug (int DebugOn)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SetDebug()\n");

   if (!DebugOn)
   {
       Debug = 0;
       return;
   }
   else
   if (Debug)
      return;

   if (getenv ("MONDESI$DBUG")) Debug = 1;

   if (Debug)
      fprintf (stdout,
"Status: 200\n\
Content-Type: text/plain\n\
Script-Control: X-content-encoding-gzip=0\n\
\r\n");

   if (Debug || OsuHttpd)
      stdout = freopen ("SYS$OUTPUT:", "w", stdout, "ctx=rec");
   else
      stdout = freopen ("SYS$OUTPUT:", "w", stdout, "ctx=bin");
}

/****************************************************************************/

